/***************************Copyright-DO-NOT-REMOVE-THIS-LINE**
  *
  * Condor Software Copyright Notice
  * Copyright (C) 1990-2005, Condor Team, Computer Sciences Department,
  * University of Wisconsin-Madison, WI.
  *
  * This source code is covered by the Condor Public License, which can
  * be found in the accompanying LICENSE.TXT file, or online at
  * www.condorproject.org.
  *
  * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
  * AND THE UNIVERSITY OF WISCONSIN-MADISON "AS IS" AND ANY EXPRESS OR
  * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
  * WARRANTIES OF MERCHANTABILITY, OF SATISFACTORY QUALITY, AND FITNESS
  * FOR A PARTICULAR PURPOSE OR USE ARE DISCLAIMED. THE COPYRIGHT
  * HOLDERS AND CONTRIBUTORS AND THE UNIVERSITY OF WISCONSIN-MADISON
  * MAKE NO MAKE NO REPRESENTATION THAT THE SOFTWARE, MODIFICATIONS,
  * ENHANCEMENTS OR DERIVATIVE WORKS THEREOF, WILL NOT INFRINGE ANY
  * PATENT, COPYRIGHT, TRADEMARK, TRADE SECRET OR OTHER PROPRIETARY
  * RIGHT.
  *
  ****************************Copyright-DO-NOT-REMOVE-THIS-LINE**/

#include <sys/utsname.h>
#include "nest_advertiser.h"
#include "nest_storage_manager.h"
#include "nest_collection_manager.h"
#include "nest_io_manager.h"
#include "nest_version.h"

class AdvertisingVisitor : public nestCollectionVisitor
{
public:
	AdvertisingVisitor( nestProtocol *dest ) { this->dest = dest; };
	~AdvertisingVisitor( void ) { };
	NestReplyStatus visit( const nestCollection &collection );
private:
	nestProtocol	*dest;
};

NestReplyStatus AdvertisingVisitor::visit( const nestCollection &collection )
{
    string buffer;
	const string *name = collection.getName( );
	const string *tmp;

	tmp = collection.getPath();
	if ( (!name->length()) || (!tmp->length()) ) {
		return NEST_BAD_COLLECTION;
	}

	buffer = *name + "Path = \"" + *tmp + "\"\n";

	time_t mtime = collection.getMtime( );
	if ( mtime > 0 ) {
		char	datebuf[32];
		NestSnprintf( datebuf, 32, "%d", mtime );
		buffer += *name + "Date = " + datebuf + "\n";
	}

	tmp = collection.getVersion( );
	if ( tmp->length() ) {
		buffer += *name + "Version = \"" + *tmp + "\"\n";
	}

	tmp = collection.getComment( );
	if ( tmp->length() ) {
		buffer += *name + "Comment = \"" + *tmp + "\"\n";
	}

	dest->Send( buffer );

	return NEST_SUCCESS;

};

nestAd::nestAd( nestProtocol *src, int len ) {
    myad  = (char *)NestMalloc( len + 1 );
    mylen = len;
    if ( len != src->Receiven( myad, len ) ) {
        nest_warn( __FILE__, __FUNCTION__, "Unable to read ad file\n" );
        NestFree( myad );
        myad = NULL;
        mylen = 0;
    }
    myad[len] = '\0';
}

nestAd::~nestAd( ) {
    if ( myad ) NestFree( myad );
}

char *
nestAd::toString( ) {
    return myad;
}

int
nestAd::length( ) {
    return mylen;
}

long uptime_start;

// **********************************
// nestAdvertiser methods
// **********************************

// Simple constructor & destructor
nestAdvertiser::nestAdvertiser( void )
{
    uptime_start = (long)time(NULL);
}

nestAdvertiser::~nestAdvertiser( void )
{
}

/*
NestReplyStatus
nestAdvertiser::readAd( nestProtocol *src ) {
        // TODO: read old collections out of file and pass to collection mngr
    return NEST_NOT_YET_IMPLEMENTED;
}
*/

nestAd *
nestAdvertiser::readAd( nestProtocol *src, int len ) {
    return ( len <= 0 ) 
        ? NULL
        : new nestAd( src, len );
}

NestReplyStatus
nestAdvertiser::writeAd(
	nestProtocol			*dest
	)
{
    char buffer[NEST_MESSAGESIZE];
    struct utsname uname_buf;

    long now = (long)time(NULL);
    long uptime = now - uptime_start;


        // first write some meta data
    NestSnprintf( buffer, NEST_MESSAGESIZE, 
            "MyType = \"Storage\"\n" 
            "TargetType = \"Storage\"\n"
            "Subtype = \"%s\"\n",
            storage_manager->type()
            );
    dest->Sendn( buffer, NestStrlen( buffer ) );
    if ( uname( &uname_buf ) < 0 ) {
        nest_perror( "uname" );
    } else {
        NestSnprintf( buffer, NEST_MESSAGESIZE, "Name = \"%s\"\n", 
                uname_buf.nodename );
        dest->Sendn( buffer, NestStrlen( buffer ) );
    }

    
    NestSnprintf( buffer,
                  NEST_MESSAGESIZE,
                  "NestTime = %ld\n"
                  "Uptime   = %ld\n", 
                  now, uptime );
    dest->Sendn( buffer, NestStrlen( buffer ) );

        // now storage
    storage_manager->writeAd( dest );

        // now the supported security models
    NestSnprintf( buffer, NEST_MESSAGESIZE, "AuthenticationModel = \"GSI\"\n");
    dest->Sendn(buffer, NestStrlen(buffer) );

        // now the subnet
    NestSnprintf( buffer, NEST_MESSAGESIZE, "Subnet = \"%s\"\n", "my ip addr");
    dest->Sendn( buffer, NestStrlen( buffer ) );
        

        // now the I/O
    io_manager->writeAd( dest );

        // now the debugging level
    NestSnprintf( buffer, NEST_MESSAGESIZE, "DebugLevel = %ld\n",
            get_nest_debug_level( ) );
    dest->Sendn( buffer, NestStrlen( buffer ) );

        // now the protocols
    for( int sd = protocol_manager->firstProtocolSocket();
            sd != protocol_manager->lastProtocolSocket();
            sd =  protocol_manager->nextProtocolSocket() )
    {
        if ( sd > 0 ) {
            NestSnprintf( buffer, NEST_MESSAGESIZE, "%sPort = %d\n",
						  protocol_manager->getProtocolNameFromSocket( sd ),
						  protocol_manager->getProtocolPortFromSocket( sd ) );
            dest->Sendn( buffer, NestStrlen( buffer ) );
        }
    }

        // now the subject name of the credential
    char *subject;
    long lifetime;
    NestReplyStatus status = protocol_manager->getCredentialInfo( 
		&subject, &lifetime );
    if ( status == NEST_SUCCESS ) {
        NestSnprintf( buffer, NEST_MESSAGESIZE, 
                "Credential_name     = \"%s\"\n"
                "Credential_lifetime = \"%ld\"\n",
                subject,
                lifetime );
        dest->Sendn( buffer, NestStrlen( buffer ) );
    }
        
        // now the local uid
    uid_t uid = NestGetuid();
    NestSnprintf( buffer, NEST_MESSAGESIZE, "uid = %d\n", uid );
    dest->Sendn( buffer, NestStrlen( buffer ) );


        // now the collections
	AdvertisingVisitor		*visitor = new AdvertisingVisitor( dest );

	// ghost storage doesn't have a collection_mgr
    if ( collection_manager != NULL ) {
        collection_manager->VisitAllCollections( *visitor );
    }
	delete visitor;

        // now the version
    NestSnprintf( buffer, NEST_MESSAGESIZE, "Version = \"%s\"\n",
				  NEST_VERSION );
    dest->Sendn( buffer, NestStrlen( buffer ) );

    return NEST_SUCCESS;
}

