/***************************Copyright-DO-NOT-REMOVE-THIS-LINE**
  *
  * Condor Software Copyright Notice
  * Copyright (C) 1990-2005, Condor Team, Computer Sciences Department,
  * University of Wisconsin-Madison, WI.
  *
  * This source code is covered by the Condor Public License, which can
  * be found in the accompanying LICENSE.TXT file, or online at
  * www.condorproject.org.
  *
  * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
  * AND THE UNIVERSITY OF WISCONSIN-MADISON "AS IS" AND ANY EXPRESS OR
  * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
  * WARRANTIES OF MERCHANTABILITY, OF SATISFACTORY QUALITY, AND FITNESS
  * FOR A PARTICULAR PURPOSE OR USE ARE DISCLAIMED. THE COPYRIGHT
  * HOLDERS AND CONTRIBUTORS AND THE UNIVERSITY OF WISCONSIN-MADISON
  * MAKE NO MAKE NO REPRESENTATION THAT THE SOFTWARE, MODIFICATIONS,
  * ENHANCEMENTS OR DERIVATIVE WORKS THEREOF, WILL NOT INFRINGE ANY
  * PATENT, COPYRIGHT, TRADEMARK, TRADE SECRET OR OTHER PROPRIETARY
  * RIGHT.
  *
  ****************************Copyright-DO-NOT-REMOVE-THIS-LINE**/

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <unistd.h>
#include <dirent.h>
#include <sys/stat.h>
#include <sys/time.h>
#include <signal.h>

#include <list>
#include <string>

#include "nest_speak.h"
#include "nest_client_util.h"
#include "nest_version.h"

using namespace std;

#define NEST_LINE_MAX 1024

#ifdef HAS_LIBREADLINE

extern "C" {
    #include <readline/readline.h>
    #include <readline/history.h>
}

#else /* HAS_LIBREADLINE */

char * readline( const char *prompt )
{
    char line[NEST_LINE_MAX];
    printf( "%s", prompt );
    if(!fgets(line,sizeof(line),stdin)) return 0;

        // replace newline with nullterm
    line[strlen(line)-1] = '\0';
    return NestStrdup(line);
}

void add_history( const char *line )
{
}


#endif /* HAS_LIBREADLINE */

static char remote_host[NEST_LINE_MAX];
static char *remote_dir = NULL;

typedef NestReplyStatus dirFunc( const char *source_path, const char *target_path, NestConnection, FILE * );

NestReplyStatus doOpen      ( NestConnection *con, char **argv );
NestReplyStatus doGet       ( NestConnection *con, char **argv );
NestReplyStatus doPut       ( NestConnection *con, char **argv );
NestReplyStatus doRead      ( NestConnection *con, char **argv );
NestReplyStatus doDelete    ( NestConnection *con, char **argv );
NestReplyStatus doRename    ( NestConnection *con, char **argv );
NestReplyStatus do3Get      ( NestConnection *con, char **argv );
NestReplyStatus do3Put      ( NestConnection *con, char **argv );
NestReplyStatus doClose     ( NestConnection *con, char **argv );
NestReplyStatus doChangeDir ( NestConnection *con, char **argv );
NestReplyStatus doVersion   ( NestConnection *con, char **argv );
NestReplyStatus doMakeDir   ( NestConnection *con, char **argv );
NestReplyStatus doRemoveDir ( NestConnection *con, char **argv );
NestReplyStatus doPrintDir  ( NestConnection *con, char **argv );
NestReplyStatus doCopy      ( NestConnection *con, char **argv );
NestReplyStatus doLink      ( NestConnection *con, char **argv );
NestReplyStatus doLStats    ( NestConnection *con, char **argv );
NestReplyStatus doList      ( NestConnection *con, char **argv );
NestReplyStatus doLList     ( NestConnection *con, char **argv );
NestReplyStatus doLChangeDir( NestConnection *con, char **argv );
NestReplyStatus doLPrintDir ( NestConnection *con, char **argv );
NestReplyStatus doUser      ( NestConnection *con, char **argv );
NestReplyStatus doGroup     ( NestConnection *con, char **argv );
NestReplyStatus doLot       ( NestConnection *con, char **argv );
NestReplyStatus doCollection( NestConnection *con, char **argv );
NestReplyStatus doKill      ( NestConnection *con, char **argv );
NestReplyStatus doDebug     ( NestConnection *con, char **argv );
NestReplyStatus doTime      ( NestConnection *con, char **argv );
NestReplyStatus doTimeout   ( NestConnection *con, char **argv );
NestReplyStatus doSync      ( NestConnection *con, char **argv );
NestReplyStatus doFormat    ( NestConnection *con, char **argv );
NestReplyStatus doAvail     ( NestConnection *con, char **argv );
NestReplyStatus doSize      ( NestConnection *con, char **argv );
NestReplyStatus doInfo      ( NestConnection *con, char **argv );
NestReplyStatus doRequestAd ( NestConnection *con, char **argv );
NestReplyStatus doHelp      ( NestConnection *con, char **argv );
NestReplyStatus doQuit      ( NestConnection *con, char **argv );
NestReplyStatus doGetPhysNam( NestConnection *con, char **argv );
NestReplyStatus doGetAcl    ( NestConnection *con, char **argv );
NestReplyStatus doSetAcl    ( NestConnection *con, char **argv );
NestReplyStatus doGetGrpAcl ( NestConnection *con, char **argv );
NestReplyStatus doSetGrpAcl ( NestConnection *con, char **argv );
NestReplyStatus doProxyQuery( NestConnection *con, char **argv );

// Helpers
NestReplyStatus ShowNestLot     ( NestLot *arg );
NestReplyStatus ShowNestLotList ( NestLotList *arg );
NestReplyStatus ShowLotHandle   ( NestLotHandle *arg );

struct command
{
    char *name;
    NestReplyStatus (*action) ( NestConnection *con, char **argv );
    int minargs;
    int maxargs;
    int must_be_connected;
    char *args;
    char *help;
};

struct command command_list[] = {
{ "open",    doOpen,      1, 2, 0, "<hostname> [port]","connect to a server" },
{ "close",   doClose,     0, 0, 1, "","disconnect from the server" },
{ "get",     doGet,       1, 2, 1, "<remote-file> [local-file]",
                                    "retrieve a file", },
{ "retr",    doGet,       1, 2, 1, "<remote-file> [local-file]",
                                    "retrieve a file", },
{ "put",     doPut,       1, 2, 1, "<local-file> [remote-file]",
                                    "send a file", },
{ "send",    doPut,       1, 2, 1, "<local-file> [remote-file]",
                                    "send a file", },
{ "3get",    do3Get,      4, 4, 1, "<file> <otherhost> <otherfile>"
                                        " <otherprotocol>",
                                    "third-party get", },
{ "3put",    do3Put,      4, 4, 1, "<file> <otherhost> <otherfile>"
                                        " <otherprotocol>",
                                    "third-party put", },
{ "read",    doRead,      3, 3, 1, "<file> <offset> <length>",
                                    "read bytes from a file", },
{ "rm",      doDelete,    1, 2, 1, "<file>", "delete a remote file", },
{ "mv",      doRename,    2, 2, 1, "<old name> <new name> ",
                                   "renames a file", },
{ "rename",  doRename,    2, 2, 1, "<old name> <new name> ",
                                   "renames a file", },
{ "pwd",     doPrintDir,  0, 0, 1, "", "shows the remote directory", },
{ "version", doVersion,   0, 0, 0, "", "check version information", },
{ "cd",      doChangeDir, 1, 1, 1, "<dir>", "change the remote directory", },
{ "chdir",   doChangeDir, 1, 1, 1, "<dir>","change the remote directory", },
{ "md",      doMakeDir,   1, 1, 1, "<dir>","make a directory", },
{ "mkdir",   doMakeDir,   1, 1, 1, "<dir>","make a directory", },
{ "rd",      doRemoveDir, 1, 2, 1, "[-r] <dir>","remove a directory", },
{ "rmdir",   doRemoveDir, 1, 2, 1, "[-r] <dir>","remove a directory", },
{ "ln",      doLink,      2, 3, 1, "[-s] <target> <link>",
                                    "create a link, -s for symbolic", },
{ "cp",      doCopy,      2, 2, 1, "oldpath newpath","copy a file"},
{ "lstats",  doLStats,    0, 0, 0, "","shows local statistics", },
{ "ls",      doList,      0, 1, 1, "[-l]","list the remote current dir", },
{ "lls",     doLList,     0, 0, 0, "","list the local current dir", },
{ "lcd",     doLChangeDir,1, 1, 0, "<dir>","move to this dir locally", },
{ "lpwd",    doLPrintDir, 0, 0, 0, "","shows the local current dir", },
{ "shutdown",doKill,      0, 0, 1, "","kill the server", },
{ "time",    doTime,      0, 0, 1, "","shows the time at the server", },
{ "timeout", doTimeout,   1, 1, 0, "<seconds>","abort session after this time",},
{ "debug",   doDebug,     1, 1, 1, "<debug-level>",
                                    "change the debugging level", },
{ "sync",    doSync,      0, 0, 1, "","commit changes", },
{ "commit",  doSync,      0, 0, 1, "","commit changes", },
{ "format",  doFormat,    0, 0, 1, "","format the disk", },
{ "avail",   doAvail,     0, 0, 1, "","shows available space", },
{ "size",    doSize,      1, 1, 1, "<file> ","shows the size of a file", },
{ "info",    doInfo,      1, 1, 1, "<file> ","shows info about a file", },
{ "proxys",  doProxyQuery,0, 0, 1, "", "lists all registered proxys",  },
{ "getpname",doGetPhysNam,1, 1, 1, "<file> ",
  "shows the physical name of a file", },
{ "getacl",  doGetAcl,    1, 2, 1, "<file> [user] ", "\n\
 shows the file access permissions for a given user\n\
 shows the access permissions for all users and groups\
 if no user is specified!", },
{ "setacl",  doSetAcl,    3, 3, 1, "<file> <user> <access> ", "\n\
 sets the file access permissions for a given user\n\
 <access> string can include one or more characters from the following:\n\
 Character   Access Permission\n\
 ---------   -----------------\n\
     r             read\n\
     w             write\n\
     l             list\n\
     d            delete\n\
     i            insert\n\
     a       modify access permissions\n\
 \n\
 To set global permissions, use nest_everybody\n\
 To remove all rights use \"none\"", },
{ "getgacl", doGetGrpAcl, 1, 2, 1, "<file> [group] ", "\n\
 shows the file access permissions for a given group\n\
 shows the access permissions for all users and groups\
 if no group is specified!", },
{ "setgacl", doSetGrpAcl, 3, 3, 1, "<file> <group> <access> ", "\n\
 sets the file access permissions for a given group\n\
 <access> string can include one or more characters from the following:\n\
 Character   Access Permission\n\
 ---------   -----------------\n\
     r           read\n\
     w           write\n\
     l           list\n\
     d           delete\n\
     i           insert\n\
     a           modify access permissions\n\
 \n\
 To remove all rights use \"none\"", },
{ "user",    doUser,      2, 4, 1, "<command> <parameters> ...",
 "\n\
 Commands       Parameters \n\
 --------       ---------- \n\
 \n\
 add            <username> <groupname> [identity] or \n\
 delete         <username>", },

{ "group",   doGroup,     2, 3, 1, "<command> <parameters> ...",
 "\n\
 Commands       Parameters \n\
 --------       ---------- \n\
 \n\
 add            <groupname> or \n\
 delete         <groupname> or \n\
 adduser        <username> <groupname> or \n\
 deleteuser     <username> <groupname>", },

{ "lot",     doLot,       0, 13, 1, "[command] <parameters> ...",
 "\n\
 Default-command is \"user_lots\"\n\
 \n\
 Commands       Parameters \n\
 --------       ---------- \n\
 user           size[bkmg] [duration] [-t <type>] [-u <user>]\n\
                [-U <parent user>] [-G <parent group>] [-L <parent handle>] \n\
 group          groupname size[bkmg] [duration] [-t <type>]\n\
                [-U <parent user>] [-G <parent group>] [-L <parent handle>] \n\
 user_lots      [user]\n\
 group_lots     <groupname> \n\
 stat           <Lot handle> \n\
 update         <Lot handle> <more size[bkmg]> [more duration] \n\
 change_t       <Lot handle> <type> [<more duration>] \n\
 movefile       <filename> <source Lot handle> "
                "<destination Lot handle> <bytes> \n\
 remove|rm      <Lot handle> \n\
 add_user       <Lot handle> <username> \n\
 delete_user    <Lot handle> <username> \n\
 attach         <Lot handle> <file/dir> \n\
 detach         <Lot handle> <file/dir> \n\
 \n\
 The lot type can be one of the following \n\
        G|g     Guaranteed \n\
        B|b     Best-effort \n\
 The default lot type is Guaranteed \n\
 [duration] format = days:hours:minutes:secons or n[.m][dhms]", },

{
 "collection", doCollection, 2, 3, 1, "[command] <parameters> ...",
 "\n\
 Commands       Parameters \n\
 --------       ---------- \n\
 add            <name> <directory>\n\
 delete         <name>\n\
 date           <name> [<seconds-since-epoch>]\n\
 version        <name> \"version string\"\n\
 comment        <name> \"comment string\"", 
},

{ "ad",      doRequestAd, 0, 0, 1, "","get nest classad", },
{ "help",    doHelp,      0, 1, 0, " [command]", "\n\
 shows usage information of a particular command\n\
 displays the list of commands if no argument is specified", },
{ "?",       doHelp,      0, 1, 0, " [command]", "\n\
 shows usage information of a particular command\n\
 displays the list of commands if no argument is specified", },
{ "exit",    doQuit,      0, 0, 0, "","quit the program", },
{ "quit",    doQuit,      0, 0, 0, "","quit the program", },
{ "logout",  doQuit,      0, 0, 0, "","quit the program", },
{ "end",     doQuit,      0, 0, 0, "","quit the program", },
{ 0, 0, 0, 0, 0 },
};


struct statistics
{
    int  reads;
    int  writes;
    nestFileSize bytesread;
    nestFileSize byteswrote;
    long timereads;
    long timewrites;
    int  total_commands;
    int  total_errors;
};

static struct statistics stats;

/*
transferInfo *
NestmakeTransferInfo( ) {
    NestTransferInfo *trans = (NestTransferInfo *)
            NestCalloc( sizeof ( NestTransferInfo ) );
    NestGettimeofday( &trans->starttime );
    memcpy(&trans->lastprogress, &trans->starttime, sizeof(struct timeval));
    return trans;
}
*/

void
Progress( nestFileSize transferred, nestFileSize total, void *args )
{
    NestTransferInfo *transfer = ( NestTransferInfo *) args;
    struct timeval now;
    NestGettimeofday( &now );
    assert( transfer != NULL );

    long millis = NestTimediff( &transfer->starttime, &now, __FUNCTION__ );
    double bw = NestBandwidth( transferred, millis );

        // print the progress info
    fprintf( stdout, "\r%10lld / %10lld (%4.2f MB/s) ",
			 transferred, total, bw );
    fflush(  stdout );

        // update global statistics
        // it is possible that Progress is called but progress is not
        // actually being made
    int progress = transferred - transfer->transferred;
    if ( progress ) {
        int newtime = NestTimediff(&transfer->lastprogress,
								   &now, __FUNCTION__);
        memcpy( &transfer->lastprogress, &now, sizeof( struct timeval ) );
        transfer->transferred += progress;
        if ( transfer->read ) {
            stats.bytesread += progress;
            stats.timereads += newtime;
        } else {
            stats.byteswrote += progress;
            stats.timewrites += newtime;
        }
    }
}

NestReplyStatus
recurseDir( char *source_path, char *target_path,
			dirFunc *func, NestConnection con, FILE *out )
{
    DIR *dir;
    struct dirent *d;
    char new_source_path[NEST_LINE_MAX];
    char new_target_path[NEST_LINE_MAX];
    NestReplyStatus ret, total_ret;

    ret = func(source_path, target_path, con, out);
    if(!NestIsDir(source_path)) return ret;

    dir = opendir(source_path);
    if(!dir) return ErrnoToNest(errno);

    total_ret = NEST_SUCCESS;

    while( (d=readdir(dir)) ) {
        if( !strcmp(d->d_name,".") || !strcmp(d->d_name,"..") ) {
			continue;
        }

        sprintf( new_source_path, "%s/%s", source_path, d->d_name );
        sprintf( new_target_path, "%s/%s", target_path, d->d_name );

        ret = recurseDir( new_source_path, new_target_path, func, con, out );
        if( ret!=NEST_SUCCESS ) total_ret = ret;
    }

    closedir(dir);
    return total_ret;
}

NestReplyStatus
getPwd( NestConnection *con )
{
    NestReplyStatus reply;
    char *newdir;
    reply = NestGetPwd( &newdir, *con );
    if(reply==NEST_SUCCESS) {
        if ( remote_dir != NULL ) {
            NestFree( remote_dir );
        }
        remote_dir = NestStrdup( newdir );
        NestFree(newdir);
    } 
    return reply;
}

NestReplyStatus
doVersion( NestConnection *con, char **args )
{
    NestReplyStatus reply = NEST_SUCCESS;
    char version[64];

    printf( "Client version: %s\n", NEST_VERSION );
	NestGetLibVersion( version, sizeof( version ) );
    printf( "Client lib version: %s\n", version ); 
    if ( *con > 0 ) {
        reply = NestGetVersion( version, sizeof(version), *con );
        if ( reply == NEST_SUCCESS ) {
            printf( "Server version: %s\n", version );
        }

            // maybe add some checking code for compatibility here
    }
    
    return reply;
}

NestReplyStatus
doReceiveNestFile( const char *path, const char *remote, NestConnection con )
{
    NestTransferInfo *trans = NestmakeTransferInfo();
    trans->read = true;
    stats.reads++;
    NestReplyStatus status;
    status = NestReceiveFile( path, remote, con, Progress, (void *)trans);
    if (status == NEST_SUCCESS) {
        fprintf( stdout, "\n" );
    }
    NestFree( trans );
    return status;
}

NestReplyStatus
doSendNestFile( const char *path, const char *remote, NestConnection con )
{
    NestTransferInfo *trans = NestmakeTransferInfo();
    trans->read = false;
    stats.writes++;
    NestReplyStatus status;
    if ( remote == NULL ) remote = path;
    status = NestSendFile( path, remote, con, Progress, (void *)trans );
    if (status == NEST_SUCCESS) {
        fprintf( stdout, "\n" );
    }
    NestFree( trans );
    return status;
}

NestReplyStatus
SendAnything( const char *source_path, const char *target_path,
			  NestConnection con, FILE *out )
{
    NestReplyStatus status;
	
    fprintf( out, "%-60.60s ", source_path );

    if ( NestIsDir( source_path ) ) {
        fprintf(out,"%-8s ","dir");
        fflush(out);
        status = NestMkDir( target_path, con ); 
        if(NEST_ALREADY_EXISTS) status = NEST_SUCCESS;
    } else if(NestIsFile(source_path) || NestIsLink(source_path) ) {
        fprintf(out,"%-8s ","\n");
        fflush(out);
        status = doSendNestFile( (char*)source_path, (char*)target_path, con );
    } else {
        fprintf(out,"%-8s ","???");
        fflush(out);
        status = NEST_SUCCESS;
    }

    if ( status != NEST_SUCCESS ) {
        fprintf( out, "%s\n", NestErrorString(status) );
    }
    else {
        fprintf( out, "\n" );
    }
    return status;
}

NestReplyStatus
doClose( NestConnection *con, char **argv )
{
    NestReplyStatus reply;
    if((*con)<0) return NEST_SUCCESS;
    reply = NestCloseConnection(*con);
    if(reply==NEST_SUCCESS) {
        *con=-1;
    }
    return reply;
}

NestReplyStatus
doOpen( NestConnection *con, char **argv )
{
    NestReplyStatus reply;
    if(*con>=0) doClose(con,argv);
    
    if ( ! argv[2] ) reply = NestOpenConnection(*con,argv[1]);
    else reply = NestOpenConnection( *con, argv[1], atoi( argv[2] ) );

    if(reply==NEST_SUCCESS) {
        strcpy(remote_host,argv[1]);
        getPwd(con);
    } else {
        *con=-1;
    }
    return reply;
}

NestReplyStatus
doRead( NestConnection *con, char **argv )
{
    NestReplyStatus status = NEST_SUCCESS;
    nestFileSize offset = atoll( argv[2] );
    int  length = atoi( argv[3] );
    if ( offset < 0 || length < 0 ) {
        return NEST_BAD_PARAMETERS;
    }
    char *string = (char *)NestMalloc( length + 1 );
    int n;
    status = NestReadBytes( *con, argv[1], offset, string, length, n );
    if ( status == NEST_SUCCESS ) {
        string[n] = '\0';
        printf( "%d bytes read:\n%s\n", n, string );
    }
    return status;
}

void
printProxyList( char *proxy, void *args )
{
    assert( args == NULL );
    printf( "%s\n", proxy );

    /*
        // wanted to print the human name not the IP address
    struct hostent *h = NestGethostbyname( proxy );
    printf( "%s\n", h->h_name );
    NestFree( h );
    */
}

NestReplyStatus
doProxyQuery( NestConnection *con, char **argv )
{
    return NestProxyQuery( *con, printProxyList, NULL );
}

NestReplyStatus
doGet( NestConnection *con, char **argv )
{
    if(!argv[2]) argv[2]=argv[1];
    return doReceiveNestFile( argv[1], argv[2], *con );
}

NestReplyStatus
doPut( NestConnection *con, char **argv )
{
	char *target;
	if(argv[2] && *argv[2]) {
		target = argv[2];
	} else {
		target = ".";
	}
	return recurseDir( argv[1], target, SendAnything, *con, stdout );
}

NestReplyStatus
do3Get( NestConnection *con, char **argv )
{
    return NestThirdPartyGet(argv[1],argv[2],argv[3],argv[4],*con);
}

/*
Grab the last component of a path name.
*/

void getStem( const char *longpath, char *shortpath )
{
	const char *s;

	s = &longpath[strlen(longpath)-2];
	while(s>longpath) {
		if(*s=='/') {
			s++;
			break;
		}
		s--;
	}

	strcpy(shortpath,s);
}

static void addToList( const NestFileInfo *info, void *arg )
{
	list<string> *dlist = (list<string>*) arg;
	string s(info->name);
	if(!strcmp(info->name,".") || !strcmp(info->name,"..")) {
		/* do nothing */
	} else {
		dlist->push_front(s);
	}
}

NestReplyStatus
NestRecursive3Put( char *file,
				   char *rhost,
				   char *rfile,
				   NestConnection con,
				   NestConnection rcon )
{
	NestReplyStatus r;
	nestFileSize size;

	fprintf(stderr,"\t%s -> %s:%s\n",file,rhost,rfile);

	r = NestFilesize(size,file,con);
	if(r==NEST_SUCCESS) {
		if(size>=0) {
			r = NestThirdPartyPut(file,rhost,rfile,"chirp",con);
		} else {
			r = NestMkDir(rfile,rcon);
			if(r==NEST_SUCCESS || r==NEST_ALREADY_EXISTS) {
				r = NestChangeDir(file,con);
				if(r==NEST_SUCCESS) {
					list<string> dlist;
					r = NestListFiles(con,addToList,&dlist);
					if(r==NEST_SUCCESS) {
						while(!dlist.empty()) {
							string sname = dlist.front();
							dlist.pop_front();
							char *name = (char*) sname.c_str();
							char shortname[NEST_MAXPATHLEN];
							char rname[NEST_MAXPATHLEN];
							getStem(name,shortname);
							sprintf(rname,"%s/%s",rfile,shortname);
							r = NestRecursive3Put(name,rhost,rname,con,rcon);
							if(r!=NEST_SUCCESS) break;
						}
					}
					if(NestChangeDir("..",con)!=NEST_SUCCESS) {
						fprintf(stderr,"lost hold of parent directory\n");
						abort();
					}
				}
			}
		}
	}

	return r;
}

NestReplyStatus
NestRecursiveSetacl( NestConnection *con, char *file, char *user, 
        nest_access_rights_t access, bool is_group )
{
	NestReplyStatus r;
	nestFileSize size;

	fprintf(stderr,"\t%s\n",file );

	r = NestFilesize(size,file,*con);
	if(r==NEST_SUCCESS) {
		if(size>=0) {
            r = NestSetAcl( *con, file, user, access, is_group );
		} else {
            r = NestChangeDir(file,*con);
            if(r==NEST_SUCCESS) {
                list<string> dlist;
                r = NestListFiles(*con,addToList,&dlist);
                if(r==NEST_SUCCESS) {
                    while(!dlist.empty()) {
                        string sname = dlist.front();
                        dlist.pop_front();
                        char *name = (char*) sname.c_str();
                        char shortname[NEST_MAXPATHLEN];
                        getStem(name,shortname);
                        r = NestRecursiveSetacl( con, shortname, 
                                user, access, is_group);
                        if(r!=NEST_SUCCESS) break;
                    }
                }
                if(NestChangeDir("..",*con)!=NEST_SUCCESS) {
                    fprintf(stderr,"lost hold of parent directory\n");
                    abort();
                }
            }
        }
    }
    if ( r == NEST_NO_CONNECTION ) {
        *con = -1;
    }
	return r;
}

NestReplyStatus
doSetAcl( NestConnection *con, char **argv )
{
    nest_access_rights_t    access;
    NestReplyStatus            status;

    char *file = argv[1];
    char *user = argv[2];

    // Parse the access string..
    status = NestParseAccessString( &access, argv[3] );
    if ( NEST_SUCCESS != status ) {
        return status;
    }

    // Do it
    return NestRecursiveSetacl( con, file, user, access, false );
}

NestReplyStatus
do3Put( NestConnection *con, char **argv )
{
	char *file, *rhost, *rfile;
	NestReplyStatus r;
	NestConnection rcon;

	file = argv[1];
	rhost = argv[2];
	rfile = argv[3];

	r = NestOpenConnection(rcon,rhost);
	if(r==NEST_SUCCESS) {
		r = NestRecursive3Put(file,rhost,rfile,*con,rcon);
		NestCloseConnection(rcon);
	}

	return r;
}

NestReplyStatus
doDelete( NestConnection *con, char **argv )
{
	if( !strcmp(argv[1],"-r") || !strcmp(argv[1],"-rf")) {
		return NestRmRecursive(argv[2],*con);
	} else {
		return NestRemoveFile( argv[1], *con );
	}
}

NestReplyStatus
doRename( NestConnection *con, char **argv )
{
    return NestRenameFile( argv[1], argv[2],  *con );
}

NestReplyStatus
doGetPhysNam( NestConnection *con, char **argv )
{
    NestReplyStatus reply;
    char *PhysicalName;
    reply = NestGetPhysicalName( &PhysicalName, argv[1], *con );
    if (reply==NEST_SUCCESS) {
        printf( " %s\n", PhysicalName );
        NestFree( PhysicalName );
    } 
    return reply;
}

void
printAccess( const NestACE *ace, void *CallbackData )
{
    (void) CallbackData;

    if (! ace) {
        return;
    }

    // Dump it out
    char    accessString[10];
    NestFormatAccessString( ace->rights, accessString );
    printf( "%15s %c %s\n", ace->name,
            ace->is_user ? 'U' : 'G',
            accessString );
}

NestReplyStatus
doGetAcl( NestConnection *con, char **argv )
{
    if ( argv[2] ) {
        printf( "Access for file '%s', user '%s':\n", argv[1], argv[2] );
    } else {
        printf( "ACL for file '%s':\n", argv[1] );
    }
    return NestGetAcl( *con, argv[1], argv[2], printAccess, NULL, false );
}

NestReplyStatus
doGetGrpAcl( NestConnection *con, char **argv )
{
    if (argv[2]) {
        printf( "Access for file '%s', group '%s':\n", argv[1], argv[2] );
    } else {
        printf( "ACL for file '%s':\n", argv[1] );
        return NestGetAcl( *con, argv[1], argv[2], printAccess, NULL, false );
    }
    return NestGetAcl( *con, argv[1], argv[2], printAccess, NULL, true );
}

NestReplyStatus
doSetGrpAcl( NestConnection *con, char **argv )
{
    nest_access_rights_t    access;
    NestReplyStatus            status;

    // Parse the access string..
    status = NestParseAccessString( &access, argv[3] );
    if ( NEST_SUCCESS != status ) {
        return status;
    }

    // Do it
    return NestSetAcl( *con, argv[1], argv[2], access, true );
}

NestReplyStatus
doSize( NestConnection *con, char **argv )
{
    NestReplyStatus reply;
    nestFileSize size;
    reply = NestFilesize( size, argv[1], *con );
    if( reply==NEST_SUCCESS ) {
        printf( "%s: %lld bytes\n", argv[1], size );
    }
    return reply;
}

NestReplyStatus
doInfo( NestConnection *con, char **argv )
{
    NestReplyStatus reply;
    NestFileInfo	info;
    reply = NestFileinfo( *con, argv[1], &info );

    if( reply==NEST_SUCCESS ) {
		const char	*ftype;
		if ( NEST_FILE == info.type ) {
			ftype = "FILE";
		} else if ( NEST_DIR == info.type ) {
			ftype = "DIRECTORY";
		} else if ( NEST_LINK == info.type ) {
			ftype = "LINK";
		} else {
			ftype = "UNKNOWN";
		}
        printf( "%s %ld bytes, type %s\n",
                info.name,
                info.size,
				ftype );
    }
    return reply;
}

NestReplyStatus
doMakeDir( NestConnection *con, char **argv )
{
    return NestMkDir( argv[1], *con );
}

NestReplyStatus
doCopy( NestConnection *con, char **argv )
{
    return NestCopy(argv[1], argv[2], *con);
}

NestReplyStatus
doLink( NestConnection *con, char **argv )
{
    if(!strcmp(argv[1],"-s")) {
        return NestLink( argv[2], argv[3], 1, *con );
    } else {
        return NestLink( argv[1], argv[2], 0, *con );
    }
}

void
printFileInfo( const NestFileInfo *f_info, void *args )
{

    char    *ls_args = (char *) args;

    if (!f_info) {
        return;
    }

    if (!ls_args) {
        printf( "%s%c\n",
				f_info->name,
				(f_info->type == NEST_DIR ? '/' : ' ' ) );
    } else if (!strcmp(ls_args, "-l")) {
		char	buf[32];
        printf( "%s%c %-22s %s\n",
                f_info->name,
				(f_info->type == NEST_DIR ? '/' : ' ' ),
                NestFormatFileSize( f_info->size, buf, sizeof( buf ), true ),
                f_info->LotList );
    }

    return;
}

NestReplyStatus
doList( NestConnection *con, char **argv )
{

    if (argv[1] && strcmp(argv[1], "-l")) {
        return NEST_BAD_PARAMETERS;
    }

    return NestListFiles( *con, printFileInfo, argv[1] );
}

NestReplyStatus
doLStats( NestConnection *con, char **argv )
{
    printf(  "%10d Total transfers\n"
             "%10d Reads\n"
             "%10d Writes\n"
             "%10lld Total bytes moved\n"
             "%10lld Total bytes read\n"
             "%10lld Total bytes written\n"
             "%10.2f MB/s total\n"
             "%10.2f MB/s reads\n"
             "%10.2f MB/s writes\n"
             "%10d commands\n"
             "%10d errors\n",
            stats.reads + stats.writes,
            stats.reads,
            stats.writes,
            stats.bytesread + stats.byteswrote,
            stats.bytesread,
            stats.byteswrote,
            NestBandwidth( stats.bytesread + stats.byteswrote,
						   stats.timereads + stats.timewrites ),
            NestBandwidth( stats.bytesread, stats.timereads ),
            NestBandwidth( stats.byteswrote, stats.timewrites ),
            stats.total_commands,
            stats.total_errors
    );

    return NEST_SUCCESS;
}

NestReplyStatus
doLList( NestConnection *con, char **argv )
{
    DIR *dir;
    struct dirent *d;

    dir = opendir(".");
    if(!dir) return ErrnoToNest(errno);

    while( (d=readdir(dir)) ) {
        printf("%s\n",d->d_name);
    }

    closedir(dir);

    return NEST_SUCCESS;
}

NestReplyStatus
doChangeDir( NestConnection *con, char **argv )
{
    NestReplyStatus reply;
    reply = NestChangeDir( argv[1], *con );
    getPwd(con);
    return reply;
}

NestReplyStatus
doLChangeDir( NestConnection *con, char **argv )
{
    int result;
    result = chdir(argv[1]);
    if(result==0) {
        return NEST_SUCCESS;
    } else {
        return ErrnoToNest(errno);
    }
}

NestReplyStatus
doLPrintDir( NestConnection *con, char **argv )
{
    char *path;
    path = getcwd(0,0);
    if(path) {
        printf("%s\n",path);
        NestFree(path);
        return NEST_SUCCESS;
    } else {
        return ErrnoToNest(errno);
    }
}

NestReplyStatus
doRemoveDir( NestConnection *con, char **argv )
{
	if( !strcmp(argv[0],"-r") || !strcmp(argv[0],"-rf") ) {
		return NestRmRecursive( argv[2], *con );
	} else {
		return NestRmDir( argv[1], *con );
	}
}

NestReplyStatus
doPrintDir( NestConnection *con, char **argv )
{
    NestReplyStatus status = getPwd( con );
    printf( "%s is remote pwd\n", remote_dir );
    return status;
}

NestReplyStatus
doKill( NestConnection *con, char **argv )
{
    NestReplyStatus status = NestRemoteShutdown( *con );
    if ( status == NEST_SUCCESS ) {
        *con = -1;
    }
    return status;
}

NestReplyStatus
doDebug( NestConnection *con, char **argv )
{
    return NestDebug(argv[1],*con);
}

NestReplyStatus
doTime( NestConnection *con, char **argv )
{
    long timestamp;
    NestReplyStatus status = NestGetTimestamp( timestamp, *con );
    if ( status == NEST_SUCCESS ) {
        printf( "NeST time is %s", ctime( (time_t *)&timestamp ) );
    }
    return status;
}

/* This is called from a signal handler, so it cannot use stdio. */

static void alarm_handler( int sig )
{
	char *msg = "\nTimeout.\n";
	write(2,msg,strlen(msg));
	exit(1);
}

NestReplyStatus
doTimeout( NestConnection *con, char **argv )
{
    int seconds = atoi(argv[1]);

    signal(SIGALRM,alarm_handler);

    if(seconds<=0) {
        alarm(0);
        printf("Timeout cancelled.\n");
    } else {
        alarm(seconds);
        printf("Timeout is %d seconds.\n",seconds);
    }

    return NEST_SUCCESS;
}

NestReplyStatus
doSync( NestConnection *con, char **argv )
{
    return NestSync(*con);
}

NestReplyStatus
doFormat( NestConnection *con, char **argv )
{
    NestReplyStatus reply;
    float size, total;
    reply = NestFormat(size, total, *con);
    if(reply==NEST_SUCCESS) {
        printf("format: %.0f available, %.0f total\n",size, total);
    }
    return reply;
}

NestReplyStatus
doAvail( NestConnection *con, char **argv )
{
    NestReplyStatus reply;
    nestLotSize size, total;
    reply = NestAvailableSpace(size, total, *con);
    if (reply==NEST_SUCCESS) {
		char	abuf[32], tbuf[32];
        printf( "avail: %s available, %s total\n",
				NestFormatLotSize( size,  abuf, sizeof( abuf ), true ),
				NestFormatLotSize( total, tbuf, sizeof( tbuf ), true )
			   );
    }
    return reply;
}

NestReplyStatus
doRequestAd( NestConnection *con, char **argv )
{
    char *ad;
    unsigned adlen;
    NestReplyStatus status = NestRequestAd( *con, &ad, &adlen );
    if ( status == NEST_SUCCESS ) {
        printf( "%s\n", ad );
    }
    delete( ad );
    return status;
}

NestReplyStatus
doUser( NestConnection *con, char **argv )
{

    if (!argv[1]) {
        return NEST_BAD_PARAMETERS;
    }

    if (!strcmp(argv[1], "add")) {
        if (!argv[2] || !argv[3]) {
            return NEST_BAD_PARAMETERS;
        }
        if(!argv[4])
            return NestAddUser( *con, argv[2], argv[3]);
        else
            return NestAddUser( *con, argv[2], argv[3], argv[4]);
    }

    if (!strcmp(argv[1], "delete")) {
        if (!argv[2]) {
            return NEST_BAD_PARAMETERS;
        }
        return NestDelUser( *con, argv[2] );
    }

    return NEST_BAD_PARAMETERS;
}


NestReplyStatus
doGroup( NestConnection *con, char **argv )
{

    if (!argv[1]) {
        return NEST_BAD_PARAMETERS;
    }

    if (!strcmp(argv[1], "add")) {
        if (!argv[2]) {
            return NEST_BAD_PARAMETERS;
        }
        return NestAddGroup( *con, argv[2] );
    }

    if (!strcmp(argv[1], "delete")) {
        if (!argv[2]) {
            return NEST_BAD_PARAMETERS;
        }
        return NestDelGroup( *con, argv[2] );
    }

    if (!strcmp(argv[1], "adduser")) {
        if (!argv[2] || !argv[3]) {
            return NEST_BAD_PARAMETERS;
        }
        return NestAddUserToGroup( *con, argv[2], argv[3] );
    }

    if (!strcmp(argv[1], "deleteuser")) {
        if (!argv[2] || !argv[3]) {
            return NEST_BAD_PARAMETERS;
        }
        return NestDelUserFromGroup( *con, argv[2], argv[3] );
    }

    return NEST_BAD_PARAMETERS;
}

NestReplyStatus
doLot( NestConnection *con, char **argv )
{

    NestLotHandle   handle;
    NestLotList     *list = NULL;
    NestReplyStatus reply;

    if (!argv[1]) {
        argv[1] = "user_lots";
    }

    if (!strcmp(argv[1], "user") || !strcmp(argv[1], "group")) {
		bool          userlot;
        int           duration = 0;
        char          *user = NULL;
		char          *group = NULL;
        char          *parent_user = NULL;
		char          *parent_group = NULL;
        NestLotHandle parent_handle;
        int           lot_type = G_LOT;
        int           args_used=2;
        rsv_id_t      rsv_id; 

		// Is this a user or group lot?
        if(!strcmp(argv[1], "user")) { 
			userlot = true;
		} else {
			userlot = false;
		}

		// Get the group name
		if ( !userlot ) {
			if ( !argv[args_used] ) {
				return NEST_BAD_PARAMETERS;
			}
			group = argv[args_used];
			args_used++;
		}

        // Get the lot size
		if ( !argv[args_used] ) {
			return NEST_BAD_PARAMETERS;
		}
        nestLotSize  bytes;
		if (  NestParseLotSize( argv[args_used], &bytes ) != NEST_SUCCESS ) {
			return NEST_BAD_PARAMETERS;
		}
		if ( bytes <= 0 ) {
			return NEST_BAD_PARAMETERS;
		}
		args_used++;

		if ( argv[args_used] && ( (argv[args_used][0] == '.')   ||
								  isdigit(argv[args_used][0]) )  ) {
			reply = NestParseLotTime( argv[args_used], &duration );
			if ( reply!= NEST_SUCCESS ) {
				return NEST_BAD_PARAMETERS;
			}
			args_used++;
		}

		// Now, all the optional parameters
        NestAdd_id_to_handle(&parent_handle, 0);
        while(argv[args_used]){

			if(!strcmp(argv[args_used], "-t")){
				args_used++;
				if(!argv[args_used]){
					return NEST_BAD_PARAMETERS;
				}
				if(!strcmp(argv[args_used], "G") || 
				   !strcmp(argv[args_used], "g")
				   )
					lot_type = G_LOT;
				else if(!strcmp(argv[args_used], "B") ||
						!strcmp(argv[args_used], "b")
						)
                        lot_type = BE_LOT;    
				else
					return NEST_BAD_PARAMETERS;
				args_used++;

			} else if(!strcmp(argv[args_used], "-u")){

				args_used++;
				if(!argv[args_used]){
					return NEST_BAD_PARAMETERS;
				}
				user = argv[args_used];
				args_used++;

			} else if(!strcmp(argv[args_used], "-U")){
				args_used++;
				if(!argv[args_used]){
					return NEST_BAD_PARAMETERS;
				}
				parent_user = argv[args_used];
				args_used++;

			} else if(!strcmp(argv[args_used], "-G")) {
				args_used++;
				if(!argv[args_used]){
					return NEST_BAD_PARAMETERS;
				}
				parent_group = argv[args_used];
				args_used++;

			} else if(!strcmp(argv[args_used], "-L")) {
				args_used++;
				if(!argv[args_used]){ 
					return NEST_BAD_PARAMETERS;
				}
				sscanf( argv[args_used], "%lu", &rsv_id );
				NestAdd_id_to_handle(&parent_handle, rsv_id);
				args_used++;

			} else {
                return NEST_BAD_PARAMETERS;
			}
        }

        if( userlot ) {
            reply = NestRequestUserLot(&handle, *con, bytes, duration, user, 
                                       parent_user, parent_group,
                                       &parent_handle, lot_type );
        } else {
            reply = NestRequestGroupLot(&handle, *con, bytes, duration, group,
                                        parent_user, parent_group,
                                        &parent_handle, lot_type );
        }
        
        if (reply == NEST_SUCCESS) {
			NestLot stat;
			reply = NestGetLotStat( &stat, *con, &handle );
			if ( reply == NEST_SUCCESS ) {
				reply = ShowNestLot( &stat );
				printf( "\n" );
			}
        }
        return reply;
    }

    if (!strcmp(argv[1], "user_lots")) {
        if(argv[2])
            reply = NestListUserLots(&list, *con, argv[2]);
        else
            reply = NestListUserLots(&list, *con);
        if (reply == NEST_SUCCESS) {
            return ShowNestLotList(list);
        }
        return reply;
    }

    if (!strcmp(argv[1], "group_lots")) {
        if (!argv[2]) {
            return NEST_BAD_PARAMETERS;
        }
        reply = NestListGroupLots(&list, *con, argv[2]);
        if (reply == NEST_SUCCESS) {
            return ShowNestLotList(list);
        }
        return reply;
    }

    if (!strcmp(argv[1], "stat")) {
        if (!argv[2]) {
            return NEST_BAD_PARAMETERS;
        }
        NestLotHandle    lot_handle;
        if (NestString_to_handle( &lot_handle, argv[2] ) < 0) {
            return NEST_BAD_PARAMETERS;
        }
		NestLot stat;
        reply = NestGetLotStat(&stat, *con, &lot_handle);
        if (reply == NEST_SUCCESS) {
            reply = ShowNestLot(&stat);
            printf("\n");
        }
        return reply;
    }

    if (!strcmp(argv[1], "update")) {
        nestLotSize  size;
        int          duration = 0;

        if (!argv[2] || !argv[3]) {
            return NEST_BAD_PARAMETERS;
        }

		if ( NestParseLotSize( argv[3], &size ) != NEST_SUCCESS ) {
            return NEST_BAD_PARAMETERS;
		}
        if (argv[4]) {
			if ( NestParseLotTime( argv[4], &duration ) != NEST_SUCCESS ) {
				return NEST_BAD_PARAMETERS;
			}
        }

        NestLotHandle    lot_handle;
        if (NestString_to_handle( &lot_handle, argv[2] ) < 0) {
            return NEST_BAD_PARAMETERS;
        }

        reply = NestUpdateLot( *con, &lot_handle, size, duration );
        if (reply == NEST_SUCCESS) {
			NestLot stat;
			reply = NestGetLotStat( &stat, *con, &lot_handle );
			if ( reply == NEST_SUCCESS ) {
				reply = ShowNestLot( &stat );
				printf( "\n" );
			}
        }
		return reply;
    }

    if (!strcmp(argv[1], "change_t")) {
        int lot_type = G_LOT;
        int  duration = 0;
       
        if (!argv[2] || !argv[3]) {
            return NEST_BAD_PARAMETERS;
        } 
    
        if(!strcmp(argv[3], "G") || !strcmp(argv[3], "g")) {
            lot_type = G_LOT;
        } else if(!strcmp(argv[3], "B") || !strcmp(argv[3], "b")) {
            lot_type = BE_LOT;
        } else {
            return NEST_BAD_PARAMETERS;
        }
    
        if (argv[4]) { 
			if ( NestParseLotTime( argv[4], &duration ) != NEST_SUCCESS ) {
				return NEST_BAD_PARAMETERS;
			}
        }
        
        NestLotHandle    lot_handle;
        if (NestString_to_handle( &lot_handle, argv[2] ) < 0) {
            return NEST_BAD_PARAMETERS;
        }
    
        return NestChangeLotType(*con, &lot_handle, lot_type, duration);
    }

    if (!strcmp(argv[1], "movefile")) {

        // Quick parm check
        if (!argv[2] || !argv[3] || !argv[4] || !argv[5]) {
            return NEST_BAD_PARAMETERS;
        }

        // Extract the lot handles from the command line
        NestLotHandle    source_handle, dest_handle;
        if (NestString_to_handle( &source_handle, argv[3] ) < 0) {
            return NEST_BAD_PARAMETERS;
        }
        if (NestString_to_handle( &dest_handle, argv[4] ) < 0) {
            return NEST_BAD_PARAMETERS;
        }

        // How much to move
        nestLotSize  bytes;
		if ( NestParseLotSize( argv[5], &bytes ) != NEST_SUCCESS ) {
			return NEST_BAD_PARAMETERS;
		}
		if ( bytes < 0 ) {
			return NEST_BAD_PARAMETERS;
		}

        // Do it
        return NestLotMoveFile( *con, argv[2], &source_handle,
                                &dest_handle, bytes );
    }

    if ( (!strcmp(argv[1], "remove")) || (!strcmp(argv[1], "rm")) ) {
        if (!argv[2]) {
            return NEST_BAD_PARAMETERS;
        }
        NestLotHandle    lot_handle;
        if (NestString_to_handle( &lot_handle, argv[2] ) < 0) {
            return NEST_BAD_PARAMETERS;
        }

        return NestTerminateLot(*con, &lot_handle);
    }

    if (!strcmp(argv[1], "add_user")) {
        if (!argv[2] || !argv[3]) {
            return NEST_BAD_PARAMETERS;
        }
        NestLotHandle lot_handle;
        if (NestString_to_handle( &lot_handle, argv[2] ) < 0) {
            return NEST_BAD_PARAMETERS;
        }
        
        return NestAddUserToLot(*con, &lot_handle, argv[3]);
    }

    if (!strcmp(argv[1], "delete_user")) {
        if (!argv[2] || !argv[3]) { 
            return NEST_BAD_PARAMETERS;
        }
        NestLotHandle lot_handle;
        if (NestString_to_handle( &lot_handle, argv[2] ) < 0) {
            return NEST_BAD_PARAMETERS;
        }

        return NestDeleteUserFromLot(*con, &lot_handle, argv[3]);
    }

    if (!strcmp(argv[1], "attach")) {
        if (!argv[2] || !argv[3]) {
            return NEST_BAD_PARAMETERS;
        }
        NestLotHandle lot_handle;
        if (NestString_to_handle( &lot_handle, argv[2] ) < 0) {
            return NEST_BAD_PARAMETERS;
        }

        return NestAttachLotToFile(*con, &lot_handle, argv[3]);
    }
        
    if (!strcmp(argv[1], "detach")) {
        if (!argv[2] || !argv[3]) {
            return NEST_BAD_PARAMETERS;
        }
        NestLotHandle lot_handle;
        if (NestString_to_handle( &lot_handle, argv[2] ) < 0) {
            return NEST_BAD_PARAMETERS;
        }

        return NestDetachLotFromFile(*con, &lot_handle, argv[3]);
    }

    return NEST_BAD_PARAMETERS;
}

NestReplyStatus
doCollection( NestConnection *con, char **argv )
{

	// Get the collection name
	char	*name = argv[2];

	// What is the request?
    if (!strcmp(argv[1], "add") ) {
		return NestNameCollection( argv[3], name, *con );
	}
	// Delete request
    else if (!strcmp(argv[1], "delete")) {
		return NestUnnameCollection( name, *con );
	}
	// Set date request
	else if (!strcmp(argv[1], "date")) {
		time_t		thetime = time( NULL );
		if ( argv[3] ) {
			if ( isdigit(*argv[3]) ) {
				thetime = (time_t) atol( argv[3] );
			} else {
				return NEST_BAD_PARAMETERS;
			}
		}
		return NestDateCollection( name, thetime, *con );
	}
	// Set Version request
	else if (!strcmp(argv[1], "version")) {
		if ( ! argv[3] ) {
			return NEST_BAD_PARAMETERS;
		}
		return NestVersionCollection( name, argv[3], *con );
	}
	// Set comment request
	else if (!strcmp(argv[1], "comment")) {
		if ( ! argv[3] ) {
			return NEST_BAD_PARAMETERS;
		}
		return NestCommentCollection( name, argv[3], *con );
	}

    return NEST_BAD_PARAMETERS;
}

NestReplyStatus
ShowLotHandle(NestLotHandle *handle)
{

    char str[NEST_NAMELEN];

    if (!handle) {
        return NEST_LOT_NOT_FOUND;
    }

    if (NestHandle_to_string(handle, str, NEST_NAMELEN) < 0) {
        return NEST_UNKNOWN_ERROR;
    }
    printf("Lot Handle - %s\n", str);

    return NEST_SUCCESS;
}

NestReplyStatus
ShowNestLot(NestLot *arg)
{

    if (!arg) {
        return NEST_LOT_NOT_FOUND;
    }

    printf("Lot Handle \t\t: %s", arg->lot_handle);
    if(!strcmp(arg->parent_lot_handle, "0"))
        printf("\nParent Lot Handle \t: %s", "-");
    else
        printf("\nParent Lot Handle \t: %s", arg->parent_lot_handle);

    if(arg->lot_type == G_LOT)
        printf("\nLot Type \t\t: %s", "Guaranteed");
    else
        printf("\nLot Type \t\t: %s", "Best-effort");

    if (arg->is_user) {
        printf("\nOwnerUser \t\t: %s", arg->owner);
    }
    else {
        printf("\nOwnerGroup \t\t: %s", arg->owner);
    }

	char tmp[64];
    printf("\nLot Device \t\t: %s", arg->dev);
    printf("\nLot Total Size \t\t: %s",
		   NestFormatLotSize( arg->size, tmp, sizeof( tmp ), true )  );
    printf("\nLot Size \t\t: %s",
		   NestFormatLotSize( arg->size, tmp, sizeof( tmp ), true )  );
    printf("\nLot Used \t\t: %s (%6.2f%%)",
		   NestFormatLotSize( arg->used, tmp, sizeof( tmp ), true ),
		   (arg->used * 100.0) / arg->size);

    time_t  time = (int)arg->start_time;
    struct tm *tm_val = localtime(&time);
    printf( "\nStart time \t\t: %d/%02d/%d %02d:%02d:%02d",
			tm_val->tm_mon+1,
			tm_val->tm_mday,
			tm_val->tm_year+1900,
			tm_val->tm_hour,
			tm_val->tm_min,
			tm_val->tm_sec );
    if(arg->lot_type == G_LOT){
        printf("\nDuration \t\t: %s",
			   NestFormatLotTime( arg->duration, tmp, sizeof(tmp), true ) );
        printf("\nTime left \t\t: %s",
			   NestFormatLotTime( arg->time_left, tmp, sizeof(tmp), true ) );
    }
    //  printf("\nCost \t\t: %d", arg->cost);
    printf("\nLot Users \t\t: %s", arg->lot_users); 
    return NEST_SUCCESS;
}


NestReplyStatus
ShowNestLotList(NestLotList *list)
{

    NestLotList *tmp;

    if (!list) {
        return NEST_LOT_NOT_FOUND;
    }

    printf("\n");
    while (list) {
        tmp = list->next;
        ShowNestLot(&list->stat);
        printf("\n\n");
        list = tmp;
    }

    return NEST_SUCCESS;
}


NestReplyStatus
doHelp( NestConnection *con, char **argv )
{

    struct command *c;
    if (argv[1]) {
        for( c=command_list; c->name; c++ ) {
            if(!strcmp(argv[1],c->name)) {
                printf("\n%15s %s\n%s",c->name,c->args,c->help);
                break;
            }
        }
        if(!c->name) {
            printf("\n Unknown command '%s'",argv[1]);
        }
        printf("\n\n To see all commands, type 'help'\n\n");
    } else {
        int i = 0;
        printf( "\n" );
        for( c=command_list; c->name; c++ ) {
            printf("%15s",c->name);
            if ( ++i%5==0 ) printf( "\n" );
        }
        printf("\n\n For detailed information, type 'help <command>'\n\n");
    }
    return NEST_SUCCESS;
}

NestReplyStatus
doQuit( NestConnection *con, char **argv )
{
    if ( stats.reads + stats.writes > 0 ) {
        doLStats(con, argv);
    }
    doClose(con,argv);
    if(stats.total_errors>0) {
        exit(1);
    } else {
        exit(0);
    }
    return NEST_SUCCESS;
}


void
doInitNestReadline()
{
    return;
}

int
ProcessAndSend( NestConnection *con,
				char **arg_list,
				int arg_count )
{
    struct command *c;
    NestReplyStatus reply;

	for( c=command_list; c->name; c++ ) {
		if(!strcmp(arg_list[0],c->name)) {
			stats.total_commands++;
			if( arg_count<c->minargs || arg_count>c->maxargs ) {
				printf("%s: wrong number of arguments (type 'help %s' "
					   "for more help)\n",arg_list[0],arg_list[0]);
			} else if( c->must_be_connected && con<0 ) {
				printf("%s: not connected\n",arg_list[0]);
			} else {
				reply = c->action(con, arg_list);
				if(reply!=NEST_SUCCESS) {
					printf("%s: %s\n",arg_list[0],
						   NestErrorString(reply));
					stats.total_errors++;
				}
			}
			break;
		}
	}

	if( !c->name ) {
		printf("%s: unknown command (type 'help' for help)\n",
			   arg_list[0]);
	}
	return 0;
}

int
ProcessInteractive( NestConnection *con )
{
    char *line = NULL;
    char prompt[32];

    doInitNestReadline();

    while(1) {
        if(line) {
            NestFree(line);
            line = NULL;
        }
        
        if(con>=0) {
            NestSnprintf( prompt, sizeof( prompt ),
						  "chirp:%s:%s> ", remote_host, remote_dir );
			prompt[sizeof(prompt)-1] = '\0';
        } else {
            strcpy( prompt, " chirp:> " );
        }

        if ( ! ( line = readline( prompt ) ) ) {
			break;
		}
        if (strcmp(line, "")) {
            add_history(line); // TODO -- only if different from last line
        }

		int arg_count;
		char **arg_list = NULL;
        NestParseLine( line, &arg_list, &arg_count );
        if( !arg_list )
			continue;
        if( !arg_list[0] )
			continue;
        arg_count--;

		ProcessAndSend( con, arg_list, arg_count );
		NestFreeArgv( arg_list, arg_count );
    }

    printf("\n");
    return 0;
}

int
ProcessNonInteractive(
	NestConnection *con,
	int argc,
	char *argv[],
	int argno )
{
	for(  ;  argno < argc;  argno++ )
	{
		int arg_count;
		char **arg_list = NULL;
        NestParseLine( argv[argno], &arg_list, &arg_count );
        if( !arg_list ) continue;
        if( !arg_list[0] ) continue;
        arg_count--;

        if(con>=0) {
            printf( "chirp:%s:%s> %s\n",
					remote_host, remote_dir, argv[argno] );
        } else {
            printf( "chirp:> %s\n", argv[argno] );
        }

		ProcessAndSend( con, arg_list, arg_count );
		NestFreeArgv( arg_list, arg_count );
	}

	return 0;
}

int
main( int argc, char *argv[] )
{
    NestConnection con=-1;
    NestReplyStatus reply;

	// Process the command line...
	char	*server = NULL, *port = NULL;
	int		argno = 1;

	// Swallow the first arg as server, next as port?
	if ( argno < argc ) {
		server = argv[argno++];
	}
	if ( ( argno < argc ) && ( isdigit( *argv[argno] ) ) ) {
		port = argv[argno++];
	}

	// Walk through the remainder of the command line
	bool	interactive = true;
	for( ;  interactive && argno<argc;  argno++ ) {
		char	*carg = argv[argno];
		char	*narg = argv[argno+1];

		if ( carg[0] == '-' ) {
			switch( carg[1] ) {
			case '\0':
				interactive = false;
				break;
			case 'p':
				if ( narg && isdigit( *narg ) ) {
					port = narg;
				} else {
					fprintf( stderr,
							 "usage: nest_client [server] [-p port]..\n" );
					exit( 1 );
				}
				argno++;
				break;
			default:
				fprintf( stderr,
						 "usage: nest_client [server] [-p port]..\n" );
			}
		}
	}

	// Connect to the server
    if( server ) {
		char	*largv[3];
		largv[1] = server;
		largv[2] = port;
        reply = doOpen(&con,largv);
        if(reply!=NEST_SUCCESS) {
            fprintf(stderr,"couldn't connect to %s: %s\n",argv[1],
                    NestErrorString(reply));
            return 1;
        }
    }

	// Handle the actuall session itself now (interactive or non-interactive)
	if ( interactive ) {
		ProcessInteractive( &con );
	} else {
		ProcessNonInteractive( &con, argc, argv, argno );
	}

    doQuit( &con, 0 );
}
