/***************************Copyright-DO-NOT-REMOVE-THIS-LINE**
  *
  * Condor Software Copyright Notice
  * Copyright (C) 1990-2005, Condor Team, Computer Sciences Department,
  * University of Wisconsin-Madison, WI.
  *
  * This source code is covered by the Condor Public License, which can
  * be found in the accompanying LICENSE.TXT file, or online at
  * www.condorproject.org.
  *
  * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
  * AND THE UNIVERSITY OF WISCONSIN-MADISON "AS IS" AND ANY EXPRESS OR
  * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
  * WARRANTIES OF MERCHANTABILITY, OF SATISFACTORY QUALITY, AND FITNESS
  * FOR A PARTICULAR PURPOSE OR USE ARE DISCLAIMED. THE COPYRIGHT
  * HOLDERS AND CONTRIBUTORS AND THE UNIVERSITY OF WISCONSIN-MADISON
  * MAKE NO MAKE NO REPRESENTATION THAT THE SOFTWARE, MODIFICATIONS,
  * ENHANCEMENTS OR DERIVATIVE WORKS THEREOF, WILL NOT INFRINGE ANY
  * PATENT, COPYRIGHT, TRADEMARK, TRADE SECRET OR OTHER PROPRIETARY
  * RIGHT.
  *
  ****************************Copyright-DO-NOT-REMOVE-THIS-LINE**/

// This file is used for NeST function prototypes, etc. that are
// commonly used in the NeST server.

// Any such functions that are also used by the client are now
// defined in "nest_shared.h".


#ifndef _NEST_COMMON_H
#define _NEST_COMMON_H

#include <arpa/inet.h>
#include <assert.h>
#include <dirent.h>
#include <errno.h>
#include <fcntl.h>
#include <netdb.h>
#include <netinet/in.h>
#include <signal.h>
#include <stddef.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <ctype.h>
#include <sys/socket.h>
#include <sys/stat.h>
#include <sys/time.h>
#include <sys/types.h>
#include <sys/param.h>
#include <sys/uio.h>
#include <sys/wait.h>
#include <sys/utsname.h>
#include <syslog.h>
#include <time.h>

#include <pthread.h>

#include "nest_shared.h"

#include <list>
#include <string>

#include "query.h"
#include <pwd.h>
#include <stdarg.h>
#include <stropts.h>

using namespace std;


#define PROXY_MONITOR_PORT 	7879
#define PROXY_BLOCK_SIZE  	65536
#define BACKLOG           	5

#define NEST_NAMELEN		(NEST_MESSAGESIZE/8)

#define FILEBUFFER			65536
#define ERROR_EXIT      	-1
#define BILLION				1000000000
#define MILLION				1000000
#define THOUSAND			1000
#define PROCESS_POOL_SIZE	15    // TODO hangs during pop creation when > 15
#define ENTIRE_FILE			-1    // a flag to specify to r/w to eof

#define NEST_REPLY_CHECK(X)     if ((X) != NEST_SUCCESS) {\
                                    return (X);\
                                }

        // a way to make sure that only dispatch thread goes where it should

void NestcheckPid( const char *file, const char *function );
#define CHECK_PID NestcheckPid( __FILE__, __FUNCTION__ )

#define DIR_MODE            0700
#define FILE_MODE           0600
#define WRITE_FLAGS         (O_WRONLY | O_CREAT | O_NONBLOCK) // | O_TRUNC
#define APPEND_FLAGS        (O_WRONLY | O_CREAT | O_NONBLOCK | O_APPEND )
#define READ_FLAGS          (O_RDONLY)
#define TRANSFER_TYPES      16

#define     NEST_READ_ACCESS        0x001
#define     NEST_WRITE_ACCESS       0x002
#define     NEST_RW_ACCESS          (NEST_READ_ACCESS | NEST_WRITE_ACCESS)
#define     NEST_LIST_ACCESS        0x004
#define     NEST_DELETE_ACCESS      0x008
#define     NEST_INSERT_ACCESS      0x010
#define     NEST_MODIFYAC_ACCESS    0x020

#define     NEST_MAX(x,y)   ( ((x)>(y)) ? (x) : (y) )
#define     NEST_MIN(x,y)   ( ((x)<(y)) ? (x) : (y) )

extern char*        NEST_TRANSFER_NAMES[];

/*
    it would be
    #define NEST_MAX_LOTS ~(1 << (8 * sizeof(long int) - 1))
    if its signed long int
*/
#define NEST_MAX_LOTS    ~((long unsigned int) 0);

#include "nest_storage_common.h"

#include "nest_debug.h"
#include "nest_arch_specific.h"
#include "nest_auth.h"

#include "nest_error.h"


#define  NEST_IDENTITIES_VIEW         "NestIdentitiesView"
#define  NEST_FILE_CATALOG_VIEW       "NestFileCatalogView"
#define  NEST_LOT_CATALOG_VIEW        "NestLotCatalogView"
#define  NEST_DEVICES_VIEW            "NestDevicesView"
#define  NEST_QUOTA_CATALOG_VIEW      "NestQuotaCatalogView"
#define  NEST_COLLECTION_VIEW         "NestCollectionView"

/*
    This is both the key_prefix and the ClassAd Type
    for all types except Identities
*/
#define     STORAGE_OBJECT_KEY_PREFIX   "NestStorageObject-"
#define     USER_KEY_PREFIX             "NestUser-"
#define     GROUP_KEY_PREFIX            "NestGroup-"
#define     IDENTITY_TYPE               "NestIdentity-"
#define     USERLIST_KEY                "UserListKey"
#define     LOT_KEY_PREFIX              "NestLot-"
#define     QUOTA_KEY_PREFIX            "NestQuotaStatus-"
#define     DEVICE_KEY_PREFIX           "NestDevice-"
#define     TOTAL_USERS_KEY             "TotalNestUsers"
#define     TOTAL_GROUPS_KEY            "TotalNestGroups"
#define     RSV_ID_KEY                  "RsvId"
#define     COLLECTION_KEY              "NestCollection-"


#define     NEST_LINK_MAXLOOP    10


extern bool     NestEnableLots; 
extern const char    *NestAdminUser;
extern user_id_t AdminUid;
extern const char    *NestSystemUser;
extern const char    *NestAnonymousUser;
extern const char    *SystemPwdFile;
extern const char    *SystemGrpFile;
extern const char    *GridMapFile;
extern const char    *MetaDir;
extern const char    *ActualGroup;
extern user_id_t ActualGid;
extern pthread_mutex_t nest_common_mutex;
extern pthread_mutex_t nest_malloc_mutex;
extern pthread_mutex_t nest_strdup_mutex;


#define CLASSAD_LOGFILE         "./ClassAdCollection.Log"
#define CLASSAD_STORAGEFILE     "./ClassAdStorageFile"
#define CLASSAD_CHECKPOINTFILE  "./ClassAdCheckPointFile"

typedef
struct grid_mapfile_entry {
    char    identity[NEST_MAXPATHLEN];
    char    name[NEST_NAMELEN];
} grid_mapfile_entry;

    // this is important
void initCommonMutex();
void lockCommonMutex( pthread_mutex_t *, char *, int );
void unlockCommonMutex( pthread_mutex_t *, char *, int );

	// Export a couple of bits from nest_server.C
const char *getRootDir( void );


    //  user management utilities

int     Nestset_grid_mapfile( const char *fname );
grid_mapfile_entry  *Nestget_grid_mapfile_entry(int state);
int     Nestinsert_grid_mapfile_entry(const char *fname,
                                      const char *identity,
                                      const char *name);
int     Nestend_grid_mapfile(int state);

struct passwd*    Nestnew_pw_ent();
void    Nestfree_pw_ent(struct passwd *);
void    Nestpw_assign(struct passwd *, struct passwd *);
int        Nestfgetpwuid(user_id_t, struct passwd *, FILE *);
int        Nestfgetpwnam(const char *, struct passwd *, FILE *);
int        Nestmy_getpwuid(user_id_t, struct passwd *, const char *);
int        Nestmy_getpwnam(const char *, struct passwd *, const char *);
int         NestGet_uid(user_id_t *, const char *, const char *);
int         NestGet_uname( char *uname,
						   user_id_t u,
						   const char *fname = SystemPwdFile);

    //  grp management utilities
int        Nestgetgrpent(group_ent *, FILE *);
int        Nestputgrpent(group_ent *, FILE *);
int        Nestgetgrpnam(const char *, group_ent *, const char *);
int        Nestgetgrpgid(user_id_t, group_ent *, const char *);

    // fdset utilities
int NestFd_print( int debug_level, fd_set *fds, int maxfds );
int NestFd_print( int debug_level, fd_set *fds, int maxfds, bool use_fstat );
void NestFd_set(   int fd, fd_set *fds );
void NestFd_clear( int fd, fd_set *fds );
int NestFd_isset( int fd, fd_set *fds );

    // socket functions
int  NestAccept   ( int sockfd, struct sockaddr *cliaddr, socklen_t *addrlen );
int  NestBind     ( int sockfd, const struct sockaddr *addr, socklen_t addrlen);
NestReplyStatus  NestConnect  
                  ( int sockfd, struct sockaddr *serv_addr, int addrlen );
int  NestClose    ( int sockfd );
int  NestShutdown ( int sockfd, int how );
void NestListen   ( int sockfd, int backlog );
int  NestSocket   ( int family, int type, int protocol );
int  NestsetSocket( int domain, int type, int protocol, int port );

void NestGetsockname( int sockfd, struct sockaddr *localaddr,
        socklen_t *addrlen );
void NestSetsockopt ( int sockfd, int level, int name, void *val,
        socklen_t len );
int  NestSelect( int numfds, fd_set * readfds, fd_set * writefds,
        fd_set * exceptfds, struct timeval * timeout );
NestReplyStatus NestMergeFdsets( fd_set *dest, fd_set *src, int maxfd );

int NestDup   ( int filedes );
int NestDup2  ( int filedes, int filedes2 );
int NestSpipe ( int fd[2] );

char *NestGetUserName( char *user );

    // network functions
struct hostent* NestGethostbyname( const char *hostname );
const char* NestInet_ntop(
        int family, const void*addrptr, char*strptr, ssize_t len);
char *NestDns_lookup( const struct sockaddr_in *from );
char *NestGetIP( struct in_addr addr );

    // descriptor functions
int     NestCreat ( const char *pathname, mode_t mode );
int     NestOpen  ( const char *pathname, int flags, mode_t mode = FILE_MODE );
off_t   NestLseek ( int fd, off_t offset, int whence );
int     NestRename( const char *oldpath, const char *newpath );
int     NestStat  ( const char *file, nest_stat *st );
int     NestFstat ( int fd, nest_stat *st );
int     NestLstat ( const char *file, nest_stat *st );
int     NestStatFS( const char *path, nest_fsstat *s );
int     NestSymlink( const char *oldpath, const char *newpath );
int     NestReadlink( const char *path, char *buf, size_t buf_len );
int     NestUnlink( const char *file );
int     NestTruncUnlink( char *file );
int     NestSync( );

ssize_t NestReadn ( int fd, void *vptr, ssize_t n );
ssize_t NestRead  ( int fd, void *vptr, ssize_t n );
ssize_t NestPread ( int fd, void *vptr, size_t count, off_t offset );
ssize_t NestWriten( int fd, const void *vptr, ssize_t n );
ssize_t NestWrite ( int fd, const void *vptr, ssize_t n );
ssize_t NestPwrite( int fd, const void *buf, size_t count, off_t offset );
void    NestClearFileFlags( int fd, int flags );
void    NestSetFileFlags( int fd, int flags );
int     NestChmod ( const char *pathname, mode_t mode );
int     NestFchmod( int fd, mode_t mode );
int     NestChown ( const char *pathname, uid_t uid, gid_t gid );
int     NestLchown ( const char *pathname, uid_t uid, gid_t gid );
int     NestFchown( int fd, uid_t uid, gid_t gid );

int NestLogMessage( int fd, char * msg );

NestReplyStatus NestGetcwd( char *buffer, size_t size );
void NestChdir( const char *pathname );
int  Nestsafe_Chdir( const char *pathname );
int  NestMkdir( const char *pathname, mode_t mode );
int  NestRmdir( const char *pathname );

int     NestChangeUser( const char *name );
uid_t    NestGetuid( );
gid_t   NestGetgid( );

int        NestSeteuid(uid_t uid);
int     NestSetegid(gid_t gid);

int NestSetenv( const char *name, const char *value, int overwrite );

void NestSignal( int signum, void (*handler)(int));

void NestBufferInit( char * buffer, unsigned long size );


int NestGetFileflags( int transfer_type );

pid_t NestGetpid( );
pid_t NestSetpgrp( );



char *NestFdGets( char *buffer, ssize_t bufsize, int fd );

NestReplyStatus NestSendFd( int sd, int fd );

int NestReceiveFd( int sd );

class nestProtocol;
bool BlockingTransfer( nestProtocol *src, nestProtocol *dst, long size );

//double  NestClassAdLotList_to_double(ExprTree *lot_list);
double  NestGetLotDuration(char *duration);
nestLotSize  NestStrLotList_to_LotSize(const char *lot_str); 

NestReplyStatus
NestAdd_ACL( NestACL **result,
         const char *name,
         int is_user,
         nest_access_rights_t right );

NestReplyStatus
NestAdd_entry_to_list(string &new_members,
                  string members,
                  string user);

NestReplyStatus
NestDelete_entry_from_list(string &new_members,
                       string members,
                       string user);

void NestNanosleep( unsigned long nanos );

int NestPathToDevice( const char *path, char *device );
int NestDeviceToPath( const char *device, char *path );

int NestFSType( const char *device, char *type );

struct timeval * NestNextTimeout( int seconds, struct timeval *last );

const char *NestUname( );

const char *NestGetModifiedIdentity( const char *identity );
const char *NestGetUserFromIdentity( const char *identity );
bool NestGetMapfileEntry( const char *name, grid_mapfile_entry *match );

const char *NestGetIdentityFromUser( const char *user );

NestReplyStatus NestCreateFullPath( const char *path );

void NestCollapsePath( const char *inpath, char *outpath );

#endif
