/***************************Copyright-DO-NOT-REMOVE-THIS-LINE**
  *
  * Condor Software Copyright Notice
  * Copyright (C) 1990-2005, Condor Team, Computer Sciences Department,
  * University of Wisconsin-Madison, WI.
  *
  * This source code is covered by the Condor Public License, which can
  * be found in the accompanying LICENSE.TXT file, or online at
  * www.condorproject.org.
  *
  * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
  * AND THE UNIVERSITY OF WISCONSIN-MADISON "AS IS" AND ANY EXPRESS OR
  * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
  * WARRANTIES OF MERCHANTABILITY, OF SATISFACTORY QUALITY, AND FITNESS
  * FOR A PARTICULAR PURPOSE OR USE ARE DISCLAIMED. THE COPYRIGHT
  * HOLDERS AND CONTRIBUTORS AND THE UNIVERSITY OF WISCONSIN-MADISON
  * MAKE NO MAKE NO REPRESENTATION THAT THE SOFTWARE, MODIFICATIONS,
  * ENHANCEMENTS OR DERIVATIVE WORKS THEREOF, WILL NOT INFRINGE ANY
  * PATENT, COPYRIGHT, TRADEMARK, TRADE SECRET OR OTHER PROPRIETARY
  * RIGHT.
  *
  ****************************Copyright-DO-NOT-REMOVE-THIS-LINE**/

#include "nest_device_manager.h"

DeviceManager::DeviceManager(void)
{
    devices = NULL;
	device = NULL;
    return;
}

DeviceManager::~DeviceManager(void)
{
	delete devices;
	delete device;
    return;
}

int
DeviceManager::add_entry( NestDevice *dev )
{
    if ( !dev || !this->devices || this->device) {
        return -1;
    }

    string  str_key;
	string  str_class_ad = "";

    str_key = DEVICE_KEY_PREFIX;
    str_key = str_key + dev->getDevName( );

	char tmp[64];
	ClassAd *dev_ad = new ClassAd;
	dev_ad->InsertAttr( "Type", DEVICE_KEY_PREFIX );
	dev_ad->InsertAttr( "Device", dev->getDevName() );
    NestSnprintf( tmp, sizeof(tmp), "%lld", dev->getCapacity() );
	dev_ad->InsertAttr( "Capacity", tmp );
    NestSnprintf( tmp, sizeof(tmp), "%lld", dev->getUserUsage() );
	dev_ad->InsertAttr( "UserUsage", tmp );
    NestSnprintf( tmp, sizeof(tmp), "%lld", dev->getGroupUsage() );
	dev_ad->InsertAttr( "GroupUsage", tmp );

#if 0
    str_class_ad =  str_class_ad +
		"[  Type = \"" + DEVICE_KEY_PREFIX + "\"; " +
		"   Device = \"" + dev->getDevName() + "\"; " +
		"   Capacity = " + str_capacity + "; " +
		"   UserUsage = " + str_user_usage + "; " +
		"   GroupUsage = " + str_group_usage +
		"]";

    ClassAd *device_class_ad = this->parser.ParseClassAd( str_class_ad, true);
#endif
	
    if (!dev_ad) {
        return -1;
    }

    if (!this->devices->AddClassAd( str_key, dev_ad )) {
        return -1;
    }

	this->device = dev;
    return 0;
}

const NestDevice *
DeviceManager::get_device_entry( const char *devname )
{
	return get_entry( devname );
}

NestDevice *
DeviceManager::get_entry( const char *dev )
{
    if (!dev || !this->devices) {
        return NULL;
    }

    if (!strcmp( dev, device->getDevName() )) {
        return device;
    }

    return NULL;
}

nestLotSize
DeviceManager::free_space(const char *dev, int is_user)
{
    const NestDevice   *disk;
    nestLotSize usage;

    disk = this->get_entry(dev);

    if (!disk) {
        return 0;
    }

    if (is_user) {
        usage = disk->getUserUsage();
    }
    else {
        usage = disk->getGroupUsage();
    }

    nestLotSize free_space = disk->getCapacity() - usage;

    if (free_space < 0) {
        /*
            Should never happen!
        */
        nest_warn( __FILE__, __FUNCTION__,
				   "Should never see this message! "
				   "== capacity<total_usage!\n");
		nest_debug( NEST_DEBUG_STORAGE, "cap = %lld, usage = %lld\n",
					disk->getCapacity(), usage );
        return 0;
    }

    return free_space;
}

nestLotSize
DeviceManager::free_space(int is_user)
{
    return this->free_space( device->getDevName(), is_user );
}

const NestDevice *
DeviceManager::get_best_device( void )
{
	return device;
}

int
DeviceManager::incr_usage(const char *dev, nestLotSize size, int is_user)
{
    NestDevice    *disk = get_entry(dev);
    if (!disk) {
        return -1;
    }
    if (is_user) {
		nestLotSize	cur = disk->getUserUsage( );
        cur += size;
		disk->setUserUsage( cur );
    }
    else {
		nestLotSize	cur = disk->getGroupUsage( );
        cur += size;
		disk->setGroupUsage( cur );
    }
    return 0;
}

int
DeviceManager::decr_usage(const char *dev, nestLotSize size, int is_user)
{
    NestDevice    *disk = get_entry(dev);
    if (!disk) {
        return -1;
    }
    if (is_user) {
		nestLotSize	cur = disk->getUserUsage( );
        cur -= size;
		disk->setUserUsage( cur );
    }
    else {
		nestLotSize	cur = disk->getGroupUsage( );
        cur -= size;
		disk->setGroupUsage( cur );
    }
    return 0;
}

NestReplyStatus
DeviceManager::initialize(ClassAdCollection *resources,
                          const char *dev )
{

    nest_debug( NEST_DEBUG_STORAGE,
                "DeviceManager::init(%s)\n",
                dev );

    this->devices = resources;
    this->query.Bind(resources);
    this->devices_view = NEST_DEVICES_VIEW;

    ViewName    parent_view = "root";

    string view_constraint(""),
           view_rank(""),
           view_partition_expr("");

    view_constraint = view_constraint +
		" other.Type == \"" + DEVICE_KEY_PREFIX + "\" ";

    if (!this->devices->CreateSubView( this->devices_view,
                                       parent_view,
                                       view_constraint,
                                       view_rank,
                                       view_partition_expr ) ) {
        nest_debug( NEST_DEBUG_ALWAYS,
                    "DeviceManager Couldn't create view for itself!\n");
        nest_debug(NEST_DEBUG_ALWAYS, "CondorErrMsg says: %s",
            CondorErrMsg.c_str());
        return NEST_CLASSAD_ERROR;
    }

    char path[NEST_MAXPATHLEN];

    if(!NestDeviceToPath(dev,path)) {
        nest_debug( NEST_DEBUG_ALWAYS,
                    "Device %s not recognized by getfsspec()-DeviceManager!\n",
                    dev );
        return ErrnoToNest(errno);
    }

	// Let's learn what we can about the f/s on the disk...
    nest_fsstat     s_fs;
    if (NestStatFS(path, &s_fs) < 0) {
        nest_perror("statfs failed:");
        nest_debug( NEST_DEBUG_ALWAYS,
                    "Cannot tolerate statfs fail!\n");
        return ErrnoToNest(errno);
    }

	// Build the disk entry
    NestDevice       *disk = new NestDevice( dev );

	disk->setBlockSize( s_fs.f_bsize );
	disk->setUserUsage( 0 );
	disk->setGroupUsage( 0 );
	disk->setCapacity( (s_fs.f_bavail * (nestLotSize) s_fs.f_bsize) );

    if ( this->add_entry( disk ) < 0 ) {
        nest_debug( NEST_DEBUG_STORAGE,
                    "Couldn't add device class ad to collection!\n" );
        return NEST_CLASSAD_ERROR;
    }

	nest_debug( NEST_DEBUG_STORAGE,
                "%s: U.Usage=%lld G.Usage=%lld Cap=%lld, bsize=%ld\n",
                disk->getDevName(),
                disk->getUserUsage(),
                disk->getGroupUsage(),
                disk->getCapacity(),
				disk->getBlockSize() );

    return NEST_SUCCESS;
}

NestReplyStatus
DeviceManager::recover( ClassAdCollection *resources )
{

    this->query.Bind(resources);
    this->devices = resources;
    this->devices_view = NEST_DEVICES_VIEW;

    this->query.Query(this->devices_view, NULL);
    this->query.ToFirst();

    string query_key;

    if (!this->query.Current(query_key)) {
        return NEST_UNKNOWN_ERROR;  //  No device found!
    }

    ClassAd *device_classad = this->devices->GetClassAd( query_key );
    if (!device_classad) {
        return NEST_CLASSAD_ERROR;
    }

	char	dev[NEST_NAMELEN];
	char	tmp[64];
    device_classad->EvaluateAttrString("Device", dev, NEST_NAMELEN );
	device = new NestDevice( dev );

    device_classad->EvaluateAttrString("Capacity", tmp, sizeof( tmp ) );
	device->setCapacity( atoll( tmp ) );

    device_classad->EvaluateAttrString("UserUsage", tmp, sizeof( tmp ) );
	device->setUserUsage( atoll( tmp ) );

    device_classad->EvaluateAttrString("GroupUsage", tmp, sizeof( tmp ) );
	device->setGroupUsage( atoll( tmp ) );

    this->query.Query(NEST_LOT_CATALOG_VIEW, NULL);
    this->query.ToFirst();

    if (!this->query.Current(query_key)) {
        return NEST_SUCCESS;
    }

    nestLotSize  size;
    nestLotSize  tot_user_usage = 0;
    nestLotSize  tot_group_usage = 0;

    bool    is_user,
            result;


    do {
        ClassAd *lot_class_ad = this->devices->GetClassAd(query_key);

        if (!lot_class_ad) {
            continue;
        }

		char tmp[64];
        result = lot_class_ad->EvaluateAttrString( "Size", tmp, sizeof(tmp) );
		size = atoll( tmp );
        result &= lot_class_ad->EvaluateAttrBool( "IsUser", is_user );

        if (!result) {
            continue;
        }

        if (is_user) {
            tot_user_usage += size;
        }
        else {
            tot_group_usage += size;
        }

    } while (this->query.Next(query_key));

	device->setUserUsage( tot_user_usage );
    device->setGroupUsage( tot_group_usage );

    if ( (device->getCapacity() < device->getUserUsage() ) ||
		 (device->getCapacity() < device->getGroupUsage()) ) {
        nest_debug( NEST_DEBUG_STORAGE,
               "Weird! The sum of all lots granted is>device %s capacity!\n",
               device->getDevName() );
    }

    nest_fsstat   s_fs;
    char path[NEST_MAXPATHLEN];

    if(!NestDeviceToPath( dev, path) ) {
        nest_debug( NEST_DEBUG_STORAGE,
                    "No information about the device %s !\n",
                    dev );
        return NEST_UNKNOWN_ERROR;
    }

    if (NestStatFS( path, &s_fs ) < 0) {
        nest_perror("statfs failed");
        nest_debug( NEST_DEBUG_STORAGE,
                    "cannot tolerate statfs fail!\n");
        return NEST_UNKNOWN_ERROR;
    }
	device->setBlockSize( s_fs.f_bsize );

	// Store it
	this->device = device;

    return NEST_SUCCESS;
}

NestReplyStatus
DeviceManager::synchronize()
{
    return NEST_SUCCESS;
}

NestReplyStatus
DeviceManager::shutdown()
{
    return NEST_SUCCESS;
}




