/***************************Copyright-DO-NOT-REMOVE-THIS-LINE**
  *
  * Condor Software Copyright Notice
  * Copyright (C) 1990-2005, Condor Team, Computer Sciences Department,
  * University of Wisconsin-Madison, WI.
  *
  * This source code is covered by the Condor Public License, which can
  * be found in the accompanying LICENSE.TXT file, or online at
  * www.condorproject.org.
  *
  * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
  * AND THE UNIVERSITY OF WISCONSIN-MADISON "AS IS" AND ANY EXPRESS OR
  * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
  * WARRANTIES OF MERCHANTABILITY, OF SATISFACTORY QUALITY, AND FITNESS
  * FOR A PARTICULAR PURPOSE OR USE ARE DISCLAIMED. THE COPYRIGHT
  * HOLDERS AND CONTRIBUTORS AND THE UNIVERSITY OF WISCONSIN-MADISON
  * MAKE NO MAKE NO REPRESENTATION THAT THE SOFTWARE, MODIFICATIONS,
  * ENHANCEMENTS OR DERIVATIVE WORKS THEREOF, WILL NOT INFRINGE ANY
  * PATENT, COPYRIGHT, TRADEMARK, TRADE SECRET OR OTHER PROPRIETARY
  * RIGHT.
  *
  ****************************Copyright-DO-NOT-REMOVE-THIS-LINE**/

extern "C" {
#include "globus_gridftp_server.h"
}
#include "nest_speak.h"
#include "nest_dsi_session.h"
#include "nest_dsi_transfer.h"
#include "nest_dsi_recv.h"
#include "nest_dsi_send.h"
#include "nest_dsi_stat.h"
#include "nest_dsi_finish.h"
#include "nest_dsi_config.h"
#include "nest_dsi_logger.h"

//using namespace std;

#define GlobusErrorNotDoneYet()                                             \
    globus_error_put(                                                       \
        globus_error_construct_error(                                       \
            NULL,                                                           \
            NULL,                                                           \
            1,                                                              \
            __FILE__,                                                       \
            "NeST DSI",                                                     \
            __LINE__,                                                       \
            "No real implementation yet."))

static
globus_version_t local_version =
{
    0,			// major version number
    1,			// minor version number
    1095172060,	// time stamp
    0			// branch ID
};


static nestDsiConfig	*config = NULL;
#ifndef DSI_NAME
# ifdef PIC
#  define DSI_NAME "nestdsi_pic"
# else
#  define DSI_NAME "nestdsi"
# endif
#endif
static char *registry_symbol = DSI_NAME;

// The public logger symbol
nestDsiLogger	*_logger;


static
void
nest_dsi_session_init(
    globus_gfs_operation_t              op,
	globus_gfs_session_info_t			*session_info )
{
	_logger->logEvent( NEST_DSI_LOG_SESSION,
					   "Session Init: here; username='%s', cred=%p\n",
					   session_info->username, session_info->del_cred );

	// Send the session start finish

	// Allocate a new session object
	nestDsiSession	*session = new nestDsiSession(
		config->getServerName(), config->getServerPort(), session_info );
	if( session->connect( op ) ) {
		_logger->logEvent( NEST_DSI_LOG_ALWAYS,
						   "Deleting session & returning\n" );

		nestDsiFinishSession	finish( op, NULL );
		finish.setCode( GLOBUS_FAILURE );
		finish.setResult( session->getStatus(), session->getErrName() );
		finish.finished( );

		delete session;
		return;
	}

	// Finished OK
	nestDsiFinishSession	finish( op, session );
	finish.finished( true, GLOBUS_SUCCESS );
}

//
//  this is exception.  just clean up and leave.
//
void
nest_dsi_session_destroy(
    void *                              user_arg)
{
	nestDsiSession		*session = (nestDsiSession *) user_arg;

	_logger->logEvent( NEST_DSI_LOG_SESSION,
					   "Session Destroy: destroying connection %p\n",
					   session );

	// Close the connection to the nest server
	if ( session ) {
		delete session;
	}
}

static
void
nest_dsi_stat(
    globus_gfs_operation_t              op,
    globus_gfs_stat_info_t *            stat_info,
    void *                              user_arg)
{
	nestDsiSession		*session = (nestDsiSession *) user_arg;

	// First, stat it...
	nestDsiStat		stat( session );
	stat.doStat( op, stat_info );

}

static
void
nest_dsi_command(
    globus_gfs_operation_t				op,
    globus_gfs_command_info_t *         cmd_info,
    void *                              user_arg)
{
	nestDsiSession	*session = (nestDsiSession *) user_arg;
	bool			handled = false;
	NestReplyStatus	status;
	nestDsiFinishCmd finish( op, cmd_info );

	switch( cmd_info->command )
	{
	case GLOBUS_GFS_CMD_MKD:
		handled = true;
		_logger->logEvent( NEST_DSI_LOG_COMMANDS,
						   "MkDir '%s'\n", cmd_info->pathname );
		status = NestMkDir( cmd_info->pathname, session->getConnection( ) );
		finish.setCreatedDir( cmd_info->pathname );
		break;
    case GLOBUS_GFS_CMD_RMD:
		handled = true;
		_logger->logEvent( NEST_DSI_LOG_COMMANDS, "RmDir '%s'\n",
						   cmd_info->pathname);
		status = NestRmDir( cmd_info->pathname, session->getConnection( ) );
		break;
    case GLOBUS_GFS_CMD_DELE:
		handled = true;
		_logger->logEvent( NEST_DSI_LOG_COMMANDS, "RemoveFile '%s'\n",
						   cmd_info->pathname );
		status = NestRemoveFile( cmd_info->pathname,
								 session->getConnection( ) );
		break;
    case GLOBUS_GFS_CMD_RNTO:
		handled = true;
		_logger->logEvent( NEST_DSI_LOG_COMMANDS, "RemoveFile '%s'->'%s'\n",
						   cmd_info->rnfr_pathname, cmd_info->pathname );
		status = NestRenameFile( cmd_info->rnfr_pathname,
								 cmd_info->pathname,
								 session->getConnection( ) );
		break;
    case GLOBUS_GFS_CMD_RNFR:
		break;
    case GLOBUS_GFS_CMD_CKSM:
		break;
    case GLOBUS_GFS_CMD_SITE_CHMOD:
		break;
    case GLOBUS_GFS_CMD_SITE_DSI:
		break;
	default:
		break;
	}

	// If it's not handled, return a "not done yet"
	if ( ! handled )
	{
		_logger->logEvent( NEST_DSI_LOG_ALWAYS,
						   "Command: %d not handled\n", cmd_info->command );
		finish.finished( true, GlobusErrorNotDoneYet() );
		return;
	}

	// Return status back...
	finish.setResult( status );
	finish.finished( );
}

static
void
nest_dsi_event (
    globus_gfs_event_info_t *           event_info,
    void *                              user_arg )
{
	//nestDsiTransfer	*transfer = (nestDsiTransfer *) user_arg;

	_logger->logEvent( NEST_DSI_LOG_ALWAYS,
					   "Event: not handled\n" );
}

static
void
nest_dsi_recv_cb (
    globus_gfs_operation_t              op,
    globus_result_t                     result,
    globus_byte_t *                     buffer,
    globus_size_t                       nbytes,
    globus_off_t                        offset,
    globus_bool_t                       eof,
    void *                              user_arg )
{
	nestDsiRecv	*recv = ( nestDsiRecv *) user_arg;
	int			status;

	// Send the data to NeST
	status = recv->sendToNest( op, result, buffer, nbytes, offset, eof );
	if ( status ) {
		_logger->logEvent( NEST_DSI_LOG_ALWAYS,
						   "RECV: sendToNest failed -- cancelling\n" );
		recv->cancel( op );
		return;
	}

	// And, setup for the next one
	if ( eof ) {
		globus_gridftp_server_finished_transfer( op, GLOBUS_SUCCESS );
		delete recv;
		_logger->logEvent( NEST_DSI_LOG_XFER,
						   "RECVcb: All done!\n" );
	} else {
		_logger->logEvent( NEST_DSI_LOG_XFER_FULL,
						   "RECVcb: Setting up next callback\n" );
		recv->setup( op, nest_dsi_recv_cb );
	}

	// Done
}

static
void
nest_dsi_recv(
    globus_gfs_operation_t   			op,
    globus_gfs_transfer_info_t *        transfer_info,
    void *                              user_arg)
{
	nestDsiSession	*session = (nestDsiSession *) user_arg;
	nestDsiRecv		*recv = NULL;

	// Create the transfer object
	recv = new nestDsiRecv( session, transfer_info );
	if ( recv->init( op ) ) {
		_logger->logEvent( NEST_DSI_LOG_ALWAYS,
						   "RECV: Transfer init failed\n" );
		delete recv;
		return;
	}

	// Misc globus results as we proceed...
	if ( recv->setup( op, nest_dsi_recv_cb ) ) {
		_logger->logEvent( NEST_DSI_LOG_ALWAYS,
						   "RECV: Setup failed\n" );
	}

	return;
}

static
void
nest_dsi_send_cb (
    globus_gfs_operation_t              op,
    globus_result_t                     result,
    globus_byte_t *                     buffer,
    globus_size_t                       nbytes,
    void *                              user_arg )
{
	nestDsiSend	*send = ( nestDsiSend *) user_arg;

	// Note that this send is complete
	int num = send->writeCB( op, result, buffer, nbytes );
	if ( num < 0 ) {
		send->cancel( op );
		return;
	}

	// Read the first block from NeST
	int status = send->readFromNest( nbytes );
	if ( status ) {
		_logger->logEvent( NEST_DSI_LOG_ALWAYS,
						   "sendcb: readFromNest() failed\n" );
		send->cancel( op );
		return;
	} else if ( status > 0 ) {
		_logger->logEvent( NEST_DSI_LOG_ALWAYS,
						   "sendcb: readFromNest() returned %d; done\n",
						   status );
		return;
	}

	// All out of data?
	if ( 0 == send->getBufferBytes( ) ) {
		send->finished( op );
		delete send;
		return;
	}

	// Misc globus results as we proceed...
	if ( send->setup( op, nest_dsi_send_cb ) ) {
		_logger->logEvent( NEST_DSI_LOG_ALWAYS,
						   "sendcb: Setup failed\n" );
	}

	// Done
	return;
}

static
void
nest_dsi_send(
    globus_gfs_operation_t              op,
    globus_gfs_transfer_info_t *        transfer_info,
    void *                              user_arg)
{
	nestDsiSession		*session = (nestDsiSession *) user_arg;
	nestDsiSend		*send = NULL;

	// Create the transfer object
	send = new nestDsiSend( session, transfer_info );
	if ( send->init( op ) ) {
		_logger->logEvent( NEST_DSI_LOG_ALWAYS,
						   "SEND: Transfer init failed\n" );
		delete send;
		return;
	}

	// Read the first block from NeST
	globus_size_t	nbytes;
	int status = send->readFromNest( nbytes );
	if ( status < 0 ) {
		send->cancel( op );
		return;
	}

	// Misc globus results as we proceed...
	if ( nbytes > 0 ) {
		if ( send->setup( op, nest_dsi_send_cb ) ) {
		_logger->logEvent( NEST_DSI_LOG_ALWAYS,
						   "SEND: Setup failed\n" );
		}
	}

	return;
}

static
int
nest_dsi_activate(void);

static
int
nest_dsi_deactivate(void);

GlobusExtensionDefineModule(globus_gridftp_server_nestdsi) =
{
    "globus_gridftp_server_" DSI_NAME,
    nest_dsi_activate,
    nest_dsi_deactivate,
    NULL,
    NULL,
    &local_version
};

static globus_gfs_storage_iface_t	nest_dsi_iface = {
	GLOBUS_GFS_DSI_DESCRIPTOR_BLOCKING,	// Descriptor

	nest_dsi_session_init,			// Init func
	nest_dsi_session_destroy,		// Destroy func

	NULL,							// List func
	nest_dsi_send,					// Send func
	nest_dsi_recv,					// Receive func
	nest_dsi_event,					// Trev func 

	NULL,							// active
	NULL,							// passive
	NULL,							// data destroy

	nest_dsi_command, 				// Command
	nest_dsi_stat,					// Stat

	NULL							// Set cred
};

static
int
nest_dsi_activate(void)
{

	// Instantiate the config and logger objets
	config = new nestDsiConfig;
	_logger = new nestDsiLogger( config );
	_logger->logEvent( NEST_DSI_LOG_ALWAYS,
					   "Activating NeST DSI '%s', server=%s, port=%d\n",
					   registry_symbol,
					   config->getServerName(), config->getServerPort() );

	// Register myself
    globus_extension_registry_add(
        GLOBUS_GFS_DSI_REGISTRY,
        registry_symbol,
        GlobusExtensionMyModule(globus_gridftp_server_nestdsi),
        &nest_dsi_iface);
    
    return 0;
}

static
int
nest_dsi_deactivate(void)
{
	_logger->logEvent( NEST_DSI_LOG_ALWAYS,
					   "Deactivating NeST DSI\n" );
    globus_extension_registry_remove(
        GLOBUS_GFS_DSI_REGISTRY, registry_symbol );

    return 0;
}
