/***************************Copyright-DO-NOT-REMOVE-THIS-LINE**
  *
  * Condor Software Copyright Notice
  * Copyright (C) 1990-2005, Condor Team, Computer Sciences Department,
  * University of Wisconsin-Madison, WI.
  *
  * This source code is covered by the Condor Public License, which can
  * be found in the accompanying LICENSE.TXT file, or online at
  * www.condorproject.org.
  *
  * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
  * AND THE UNIVERSITY OF WISCONSIN-MADISON "AS IS" AND ANY EXPRESS OR
  * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
  * WARRANTIES OF MERCHANTABILITY, OF SATISFACTORY QUALITY, AND FITNESS
  * FOR A PARTICULAR PURPOSE OR USE ARE DISCLAIMED. THE COPYRIGHT
  * HOLDERS AND CONTRIBUTORS AND THE UNIVERSITY OF WISCONSIN-MADISON
  * MAKE NO MAKE NO REPRESENTATION THAT THE SOFTWARE, MODIFICATIONS,
  * ENHANCEMENTS OR DERIVATIVE WORKS THEREOF, WILL NOT INFRINGE ANY
  * PATENT, COPYRIGHT, TRADEMARK, TRADE SECRET OR OTHER PROPRIETARY
  * RIGHT.
  *
  ****************************Copyright-DO-NOT-REMOVE-THIS-LINE**/

#ifndef NEST_DSI_GFSFINISH_H
#define NEST_DSI_GFSFINISH_H

#include "nest_common.h"
extern "C" {
#include "globus_gridftp_server.h"
}

// Base "Finish" class
class nestDsiSession;
class nestDsiFinish
{
public:
	nestDsiFinish( globus_gfs_operation_t		gfsOp );
	virtual ~nestDsiFinish( void );

	// Code to help with "globus_result_t" handling
	const char * TranslateNestStatus(
		NestReplyStatus status );
	void nestDsiFinish::setResult( 
		NestReplyStatus	status,
		const char		*string = NULL );
	void nestDsiFinish::setResult(
		int				errno,
		const char		*string = NULL );
	void nestDsiFinish::setResult(
		globus_result_t	result );
	globus_result_t genResult(
		const char		*func,
		int				my_errno );

  protected:
	globus_gfs_operation_t		gfsOp;
	globus_result_t				result;
	bool						result_set;
};

// Base "finish" class with "Info"
class nestDsiFinishInfo : public nestDsiFinish
{
  public:
	nestDsiFinishInfo( globus_gfs_operation_t		gfsOp,
					   globus_gfs_operation_type_t	type );
	virtual ~nestDsiFinishInfo( void );

	// Store basic info
	int setCode( int code );

	// Call this to actually send the finish
	virtual void finished(
		bool			valid = false,
		globus_result_t	result = GLOBUS_SUCCESS );

  protected:
	globus_gfs_finished_info_t	info;
};

// Finish a session create
class nestDsiFinishSession : public nestDsiFinishInfo
{
  public:
	nestDsiFinishSession( globus_gfs_operation_t	gfsOp,
						  nestDsiSession			*session );
	virtual ~nestDsiFinishSession( void );

	// Store basic info
	int setSession( nestDsiSession *session );

  private:

};

class nestDsiFinishCmd : public nestDsiFinishInfo
{
  public:
	nestDsiFinishCmd( globus_gfs_operation_t			gfsOp,
					  const globus_gfs_command_info_t	*cmd );
	virtual ~nestDsiFinishCmd( void );

	// Store basic info
	int setChecksum( const char *cksum );
	int setCreatedDir( const char *path );

  private:

};


class nestDsiFinishStat : public nestDsiFinish
{
  public:
	nestDsiFinishStat( globus_gfs_operation_t gfsOp );
	~nestDsiFinishStat( );

	// Generate the actual finished events
	void finished( void );
	void finished(
		globus_gfs_stat_t	*statarray,
		int					count );

  private:
	globus_gfs_finished_info_t	*info;
};

class nestDsiFinishTransfer : public nestDsiFinish
{
  public:
	nestDsiFinishTransfer( globus_gfs_operation_t gfsOp );
	~nestDsiFinishTransfer( );

	// Generate the actual finished events
	void finished( void );
	void finished( NestReplyStatus status );

  private:
};

#endif

