/***************************Copyright-DO-NOT-REMOVE-THIS-LINE**
  *
  * Condor Software Copyright Notice
  * Copyright (C) 1990-2005, Condor Team, Computer Sciences Department,
  * University of Wisconsin-Madison, WI.
  *
  * This source code is covered by the Condor Public License, which can
  * be found in the accompanying LICENSE.TXT file, or online at
  * www.condorproject.org.
  *
  * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
  * AND THE UNIVERSITY OF WISCONSIN-MADISON "AS IS" AND ANY EXPRESS OR
  * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
  * WARRANTIES OF MERCHANTABILITY, OF SATISFACTORY QUALITY, AND FITNESS
  * FOR A PARTICULAR PURPOSE OR USE ARE DISCLAIMED. THE COPYRIGHT
  * HOLDERS AND CONTRIBUTORS AND THE UNIVERSITY OF WISCONSIN-MADISON
  * MAKE NO MAKE NO REPRESENTATION THAT THE SOFTWARE, MODIFICATIONS,
  * ENHANCEMENTS OR DERIVATIVE WORKS THEREOF, WILL NOT INFRINGE ANY
  * PATENT, COPYRIGHT, TRADEMARK, TRADE SECRET OR OTHER PROPRIETARY
  * RIGHT.
  *
  ****************************Copyright-DO-NOT-REMOVE-THIS-LINE**/

#ifndef NEST_DSI_LOGGER_H
#define NEST_DSI_LOGGER_H

#include <stdarg.h>
#include <pthread.h>
#include "nest_dsi_config.h"

#define NEST_DSI_LOG_NONE			0x00000000
#define NEST_DSI_LOG_ALWAYS			0x00000001
#define NEST_DSI_LOG_FULLDEBUG		0x00000002
#define NEST_DSI_LOG_SESSION		0x00000004
#define NEST_DSI_LOG_XFER			0x00000008
#define NEST_DSI_LOG_XFER_FULL		0x00000010
#define NEST_DSI_LOG_COMMANDS		0x00000020
#define NEST_DSI_LOG_STAT			0x00000040
#define NEST_DSI_LOG_FINISH			0x00000080
#define NEST_DSI_LOG_ALL			0xffffffff

class nestDsiLogger
{
  public:
	nestDsiLogger( const nestDsiConfig *config );
	~nestDsiLogger( void );

	void logEvent(   unsigned level, const char *format, ... );
	void logEventVa( unsigned level, const char *format, va_list args );

  private:
	const char	*log_file;
	const char	*old_file;
	unsigned	log_flags;
	long		max_size;
	int			lock_count;
	bool		rotating;
	enum { LOG_STDOUT, LOG_STDERR, LOG_FILE } log_type;
	pthread_mutex_t mutex;
	void Lock( void );
	void Unlock( void );
};

// Declare the external logger that's public
extern nestDsiLogger	*_logger;

#endif
