/***************************Copyright-DO-NOT-REMOVE-THIS-LINE**
  *
  * Condor Software Copyright Notice
  * Copyright (C) 1990-2005, Condor Team, Computer Sciences Department,
  * University of Wisconsin-Madison, WI.
  *
  * This source code is covered by the Condor Public License, which can
  * be found in the accompanying LICENSE.TXT file, or online at
  * www.condorproject.org.
  *
  * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
  * AND THE UNIVERSITY OF WISCONSIN-MADISON "AS IS" AND ANY EXPRESS OR
  * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
  * WARRANTIES OF MERCHANTABILITY, OF SATISFACTORY QUALITY, AND FITNESS
  * FOR A PARTICULAR PURPOSE OR USE ARE DISCLAIMED. THE COPYRIGHT
  * HOLDERS AND CONTRIBUTORS AND THE UNIVERSITY OF WISCONSIN-MADISON
  * MAKE NO MAKE NO REPRESENTATION THAT THE SOFTWARE, MODIFICATIONS,
  * ENHANCEMENTS OR DERIVATIVE WORKS THEREOF, WILL NOT INFRINGE ANY
  * PATENT, COPYRIGHT, TRADEMARK, TRADE SECRET OR OTHER PROPRIETARY
  * RIGHT.
  *
  ****************************Copyright-DO-NOT-REMOVE-THIS-LINE**/

#include "nest_dsi_session.h"
#include "nest_dsi_logger.h"

nestDsiSession::nestDsiSession( const char *server,
						  int port,
						  const globus_gfs_session_info_t *info )
{
	copy_session_info( info );

	// Finally, we are currently not connected!
	this->connected = false;
	this->server = strdup( server );
	this->port = port;
	return;
}
nestDsiSession::nestDsiSession( nestDsiSession *old )
{
	copy_session_info( &(old->session_info) );
	this->server = strdup( old->server );
	this->port = old->port;
	this->connected = false;
	return;
}
nestDsiSession::~nestDsiSession( void )
{
	disconnect( );

	if ( this->session_info.username ) {
		free( (void*) this->session_info.username );
		this->session_info.username = NULL;
	}

	if ( this->session_info.password ) {
		free( (void*) this->session_info.password );
		this->session_info.password = NULL;
	}

	if ( this->session_info.subject ) {
		free( (void*) this->session_info.subject );
		this->session_info.subject = NULL;
	}

	if ( this->session_info.cookie ) {
		free( (void*) this->session_info.cookie );
		this->session_info.cookie = NULL;
	}

	if ( this->session_info.host_id ) {
		free( (void*) this->session_info.host_id );
		this->session_info.host_id = NULL;
	}

	if( this->server ) {
		free( (void*) this->server );
	}

}

int
nestDsiSession::connect( globus_gfs_operation_t gfsOp )
{
	if ( connected ) {
		return 0;
	}

	// Connect to the NeST server
	_logger->logEvent( NEST_DSI_LOG_SESSION,
					   "Connecting to nest server '%s' @ %d...\n",
					   server, port );
	NestReplyStatus status = NestOpenConnection(
		connection,
		server,
		port,
		session_info.username,
		session_info.del_cred );

	// And, how'd it go?
	if ( status != NEST_SUCCESS ) {
		_logger->logEvent( NEST_DSI_LOG_ALWAYS,
						   "Connection to %s/%d failed\n", server, port );
		this->status = status;
		return -1;
	}

	//finish.setSession( (nestDsiSession *) this );
	_logger->logEvent( NEST_DSI_LOG_SESSION,
					   "Connection ok\n" );
	connected = true;
	this->status = 0;
	return 0;
}

int
nestDsiSession::disconnect( void )
{
	if ( connected ) {
		(void) NestCloseConnection( connection );
		connected = false;
	}
	return 0;
}

int
nestDsiSession::copy_session_info( const globus_gfs_session_info_t *info )
{
	this->session_info.del_cred = info->del_cred;

	// Copy out the strings
	if ( info->username ) {
		this->session_info.username = strdup( info->username );
	} else {
		this->session_info.username = NULL;
	}
	if ( info->password ) {
		this->session_info.password = strdup( info->password );
	} else {
		this->session_info.password = NULL;
	}
	if ( info->subject ) {
		this->session_info.subject = strdup( info->subject );
	} else {
		this->session_info.subject = NULL;
	}
	if ( info->cookie ) {
		this->session_info.cookie = strdup( info->cookie );
	} else {
		this->session_info.cookie = NULL;
	}
	if ( info->host_id ) {
		this->session_info.host_id = strdup( info->host_id );
	} else {
		this->session_info.host_id = NULL;
	}

	return 0;
}
