/***************************Copyright-DO-NOT-REMOVE-THIS-LINE**
  *
  * Condor Software Copyright Notice
  * Copyright (C) 1990-2005, Condor Team, Computer Sciences Department,
  * University of Wisconsin-Madison, WI.
  *
  * This source code is covered by the Condor Public License, which can
  * be found in the accompanying LICENSE.TXT file, or online at
  * www.condorproject.org.
  *
  * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
  * AND THE UNIVERSITY OF WISCONSIN-MADISON "AS IS" AND ANY EXPRESS OR
  * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
  * WARRANTIES OF MERCHANTABILITY, OF SATISFACTORY QUALITY, AND FITNESS
  * FOR A PARTICULAR PURPOSE OR USE ARE DISCLAIMED. THE COPYRIGHT
  * HOLDERS AND CONTRIBUTORS AND THE UNIVERSITY OF WISCONSIN-MADISON
  * MAKE NO MAKE NO REPRESENTATION THAT THE SOFTWARE, MODIFICATIONS,
  * ENHANCEMENTS OR DERIVATIVE WORKS THEREOF, WILL NOT INFRINGE ANY
  * PATENT, COPYRIGHT, TRADEMARK, TRADE SECRET OR OTHER PROPRIETARY
  * RIGHT.
  *
  ****************************Copyright-DO-NOT-REMOVE-THIS-LINE**/

extern "C" {
#include "globus_gridftp_server.h"
}

#include "nest_dsi_stat.h"
#include "nest_dsi_finish.h"
#include "nest_dsi_logger.h"

nestDsiStat::nestDsiStat( nestDsiSession *parent_session )
{
	this->parent_session = parent_session;
	this->session = NULL;
	this->list = NULL;
	this->end = NULL;
	this->count = 0;

}

nestDsiStat::~nestDsiStat( void )
{

}

void
cStoreFileInfo( const NestFileInfo *finfo, void *arg )
{
	nestDsiStat	*stat = (nestDsiStat *) arg;
	stat->storeFileInfo( finfo );
}

void
nestDsiStat::storeFileInfo( const NestFileInfo *finfo )
{

	// Allocate a buffer to store it in
	NestFileInfo	*newinfo;
	newinfo = (NestFileInfo*) calloc( 1, sizeof(NestFileInfo) );
	if ( ! newinfo ) {
		return;
	}

	// Copy it...
	memcpy( newinfo, finfo, sizeof( NestFileInfo ) );
	newinfo->next = NULL;

	// If it's the first one...
	if ( ! list ) {
		list = newinfo;
		end = newinfo;
	} else {
		// Otherwise, store it at the tail of the list...
		end->next = newinfo;
		end = newinfo;
	}
	count++;
}

void
nestDsiStat::storeDir( const char *name )
{
	NestFileInfo	info;

	memset( &info, 0, sizeof( info ) );
	info.type = NEST_DIR;
	int now = (int) time( NULL );
	info.atime = now;
	info.ctime = now;
	info.mtime = now;
	strcpy( info.name, name );

	storeFileInfo( &info );
}

int
nestDsiStat::doStat( globus_gfs_operation_t		op,
					 globus_gfs_stat_info_t		*stat_info )
{
	NestFileInfo		info;
	NestReplyStatus		status;
	nestDsiFinishStat	finish( op );

	GlobusGFSName(nestDsiStat::doStat);

	// Connect to the NeST
	// Create the NeST session to handle the stat
	session = new nestDsiSession( parent_session );
	status = session->connect( op );

	// How'd it go?
	if ( GLOBUS_SUCCESS != status ) {
		_logger->logEvent( NEST_DSI_LOG_ALWAYS,
						   "stat::doStat - Connection failed\n" );
		finish.setResult( session->getStatus(), session->getErrName() );
		finish.finished( );
		delete session;
		return -1;
	}

	// How'd it go?
    status = NestFileinfo( session->getConnection( ),
						   stat_info->pathname,
						   &info );
	if ( NEST_SUCCESS != status ) {
		_logger->logEvent( NEST_DSI_LOG_ALWAYS,
						   "stat::doStat - NestFileInfo of '%s' failed\n",
						   stat_info->pathname );
		finish.setResult( status );
		finish.finished( );
		delete session;
		return -1;
	}

	// Don't handle links, at least for now...
	if ( info.type == NEST_LINK ) {
		_logger->logEvent( NEST_DSI_LOG_STAT,
						   "stat::doStat - Is a link\n" );
		finish.setResult( NEST_FILE_NOT_FOUND, "doStat" );
		finish.finished( );
		delete session;
		return -1;
	}

	// If it's a file, pretty easy to handle...
	if ( info.type == NEST_FILE ) {
		stat_info->file_only = true;
		globus_gfs_stat_t	statbuf;

		// Fill in the statbuf & done
		CopyFileInfo( &info, &statbuf );
		finish.finished( &statbuf, 1 );
		delete session;
		return 0;
	}

	// Here, it's a directory.  Read it.
	stat_info->file_only = false;
	status = NestChangeDir( stat_info->pathname,
							session->getConnection( ) );
	if ( NEST_SUCCESS != status ) {
		_logger->logEvent( NEST_DSI_LOG_ALWAYS,
						   "stat::doStat - chdir to '%s' failed\n",
						   stat_info->pathname );
		finish.setResult( session->getStatus(), session->getErrName() );
		finish.finished( );
		delete session;
		return -1;
	}

	// Insert "." and ".." entries.
	storeDir( "." );
	storeDir( ".." );

	// Now, the real dir contents
	status = NestListFiles( session->getConnection( ),
							cStoreFileInfo,
							this );
	if ( NEST_SUCCESS != status ) {
		_logger->logEvent( NEST_DSI_LOG_ALWAYS,
						   "stat::doStat - list file '%s' of failed\n",
						   stat_info->pathname );
		finish.setResult( session->getStatus(), session->getErrName() );
		finish.finished( );
		delete session;
		return -1;
	}
	_logger->logEvent( NEST_DSI_LOG_STAT,
					   "stat::doStat - Found %d files\n", count );

	// Empty directory?
	if ( 0 == count ) {
		finish.finished( NULL, 0 );
		delete session;
		return 0;
	}

	// Ok, now generate an array of stat buffers from the list
	globus_gfs_stat_t	*stat_array;
	stat_array =
		(globus_gfs_stat_t *)calloc( count, sizeof( globus_gfs_stat_t ) );
	if ( info.type == NEST_LINK ) {
		finish.setResult( NEST_OUT_OF_RESOURCES, "doStat" );
		finish.finished( );
		delete session;
		return -1;
	}

	// Walk through the list
	NestFileInfo	*cur = list;
	int				num = 0;
	int				total_count = count;
	while( cur ) {
		globus_gfs_stat_t *statbuf = &stat_array[num];

		// Copy the info
		assert( num < total_count );
		CopyFileInfo( cur, statbuf );

		// Go to the next one in the array
		num++;

		// Adjust the pointers, free up the current one
		NestFileInfo	*tmp = cur;
		cur = cur->next;
		free( tmp );
		list = cur;
		count--;
	}
	end = NULL;
	assert( 0 == count );
	assert( NULL == list );

	// Send it off
	finish.finished( stat_array, total_count );

	// And, free up the memory allocated
	for ( num = 0;  num < total_count;  num++ ) {
		FreeStatbuf( &stat_array[num] );
	}
	free( stat_array );

	// Clean up
	delete session;

	// Ok, done
	return 0;
}

void
nestDsiStat::CopyFileInfo( const NestFileInfo *info,
						   globus_gfs_stat_t *statbuf )
{
	if ( info->type == NEST_DIR ) {
		statbuf->mode = ( S_IFDIR | S_IRWXU );
	} else {
		statbuf->mode = ( S_IFREG | S_IRWXU );
	}
	statbuf->nlink = 1;

	statbuf->name = strdup( info->name );

	statbuf->uid = getuid( );		// TODO
	statbuf->gid = getgid( );		// TODO

	statbuf->size = info->size;

	statbuf->atime = info->atime;
	statbuf->ctime = info->ctime;
	statbuf->mtime = info->mtime;
}

void
nestDsiStat::FreeStatbuf( globus_gfs_stat_t *statbuf )
{
	free( statbuf->name );
}
