/***************************Copyright-DO-NOT-REMOVE-THIS-LINE**
  *
  * Condor Software Copyright Notice
  * Copyright (C) 1990-2005, Condor Team, Computer Sciences Department,
  * University of Wisconsin-Madison, WI.
  *
  * This source code is covered by the Condor Public License, which can
  * be found in the accompanying LICENSE.TXT file, or online at
  * www.condorproject.org.
  *
  * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
  * AND THE UNIVERSITY OF WISCONSIN-MADISON "AS IS" AND ANY EXPRESS OR
  * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
  * WARRANTIES OF MERCHANTABILITY, OF SATISFACTORY QUALITY, AND FITNESS
  * FOR A PARTICULAR PURPOSE OR USE ARE DISCLAIMED. THE COPYRIGHT
  * HOLDERS AND CONTRIBUTORS AND THE UNIVERSITY OF WISCONSIN-MADISON
  * MAKE NO MAKE NO REPRESENTATION THAT THE SOFTWARE, MODIFICATIONS,
  * ENHANCEMENTS OR DERIVATIVE WORKS THEREOF, WILL NOT INFRINGE ANY
  * PATENT, COPYRIGHT, TRADEMARK, TRADE SECRET OR OTHER PROPRIETARY
  * RIGHT.
  *
  ****************************Copyright-DO-NOT-REMOVE-THIS-LINE**/

extern "C" {
#include "globus_gridftp_server.h"
}

#include "nest_dsi_transfer.h"
#include "nest_dsi_finish.h"
#include "nest_dsi_logger.h"

static int	transfer_num = 0;

nestDsiTransfer::nestDsiTransfer(
	nestDsiSession				*parent_session,
	globus_gfs_transfer_info_t	*transfer_info )
{
	this->parent_session = parent_session;
	this->session = NULL;
	this->num = transfer_num++;
	this->buffer_size = 0;
	this->buffer = NULL;
	this->transfer_status = NEST_SUCCESS;
	memcpy( &(this->info),
			transfer_info,
			sizeof( globus_gfs_transfer_info_t) );
	this->info.pathname = strdup( transfer_info->pathname );
}

nestDsiTransfer::~nestDsiTransfer( void )
{
	// Free up the tranfer buffer
	if( buffer ) {
		free( buffer );
	}

	// Delete the session
	if( session ) {
		session->disconnect( );
		delete session;
	}

	// Free up tranfer info pieces that we've dup'ed
	if ( info.pathname ) {
		free( info.pathname );
	}
}

int
nestDsiTransfer::init( globus_gfs_operation_t op )
{
    int				optimal_count;
    globus_size_t	block_size;

	// Get block size & concurrency
    globus_gridftp_server_get_optimal_concurrency( op, &optimal_count );
    globus_gridftp_server_get_block_size( op, &block_size );
    globus_assert( optimal_count > 0 && block_size > 0 );

	// Allocate the transfer buffer
	buffer_size = block_size;
	this->buffer = malloc( buffer_size );
	if ( ! this->buffer ) {
		cancel( op, NEST_OUT_OF_RESOURCES, "malloc" );
		return -1;
	}
	this->buffer_size = buffer_size;

	// Connect to the NeST
	// Create the NeST session to handle the transfer
	this->session = new nestDsiSession( parent_session );

	// Connect to the NeST
	if( session->connect( op ) ) {
		_logger->logEvent( NEST_DSI_LOG_ALWAYS,
						   "nestDsiTransfer::init : session connect failed\n"
						   );
		cancel( op );
		return -1;
	}

	// Start the transfer
	_logger->logEvent( NEST_DSI_LOG_XFER,
					   "nestDsiTransfer::init : Starting transfer\n" );
	globus_gridftp_server_begin_transfer(
		op, 
		0,
		this );

	// Ok, done
	return 0;
}

void
nestDsiTransfer::finished( globus_gfs_operation_t op )
{
	nestDsiFinishTransfer	finish( op );
	finish.finished( );
}

void
nestDsiTransfer::cancel( globus_gfs_operation_t op,
						 NestReplyStatus status,
						 const char *string )
{

	nestDsiFinishTransfer	finish( op );

	if ( NULL == string ) {
		string = "Unknown";
	}

	// Set the result thingy
	if ( NEST_SUCCESS != session->getStatus( ) ) {
		finish.setResult( session->getStatus(), session->getErrName() );
	} else if ( status ) {
		finish.setResult( status, string );
	} else if ( NEST_SUCCESS != transfer_status ) {
		_logger->logEvent( NEST_DSI_LOG_XFER,
						   "Setting result to %d\n", transfer_status );
		finish.setResult( transfer_status );
	}

	finish.finished( );
}
