/***************************Copyright-DO-NOT-REMOVE-THIS-LINE**
  *
  * Condor Software Copyright Notice
  * Copyright (C) 1990-2005, Condor Team, Computer Sciences Department,
  * University of Wisconsin-Madison, WI.
  *
  * This source code is covered by the Condor Public License, which can
  * be found in the accompanying LICENSE.TXT file, or online at
  * www.condorproject.org.
  *
  * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
  * AND THE UNIVERSITY OF WISCONSIN-MADISON "AS IS" AND ANY EXPRESS OR
  * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
  * WARRANTIES OF MERCHANTABILITY, OF SATISFACTORY QUALITY, AND FITNESS
  * FOR A PARTICULAR PURPOSE OR USE ARE DISCLAIMED. THE COPYRIGHT
  * HOLDERS AND CONTRIBUTORS AND THE UNIVERSITY OF WISCONSIN-MADISON
  * MAKE NO MAKE NO REPRESENTATION THAT THE SOFTWARE, MODIFICATIONS,
  * ENHANCEMENTS OR DERIVATIVE WORKS THEREOF, WILL NOT INFRINGE ANY
  * PATENT, COPYRIGHT, TRADEMARK, TRADE SECRET OR OTHER PROPRIETARY
  * RIGHT.
  *
  ****************************Copyright-DO-NOT-REMOVE-THIS-LINE**/

#ifndef     _NEST_FILE_CATALOG_H_
#define     _NEST_FILE_CATALOG_H_

#include "nest_common.h"
#include "nest_user.h"

class nestFileCatalog
{

  public:

	nestFileCatalog( );
	~nestFileCatalog();

	bool    is_file_present(const char *file);

	bool    is_system_file(const char *file);

	/*
	  All 'file' names are represented from the respective user's
	  logical root and are 'logical' names!
	*/
	NestReplyStatus add_entry(const char *logical_file,
							  const char *physical_file,
							  const char *logical_dir,
							  const nestUser *user,
							  nest_obj_type_t file_type);

	/*
	  For files and directories
	*/
	NestReplyStatus add_entry(const char *file,
							  const nestUser *user,
							  nest_obj_type_t type);

	/*
	  For sym-links
	*/
	NestReplyStatus add_entry(const char *newpath,
							  const char *oldpath,
							  const nestUser *user,
							  nest_obj_type_t type);

	NestReplyStatus delete_entry(const char *file);

	NestReplyStatus
        get_system_link_physical_name(char *physical_filename,
                                      const char *logical_filename);

	NestReplyStatus get_physical_name(char *physical_filename,
									  const char *logical_filename);

	NestReplyStatus get_logical_name(char *logical_filename,
									 const char *logical_filename);

	NestReplyStatus get_logical_name(char *logical_filename,
									 const char *filename,
									 const nestUser *user);

	NestReplyStatus get_reverse_logical_name( char *dest,
											  const char *src );

	NestReplyStatus get_owner_user( const char *file,
									nestUser *user );

	NestReplyStatus get_type(nest_obj_type_t *type,
							 const char *file);

	NestReplyStatus get_size(nestFileSize  *size,
							 const char *file);

	NestReplyStatus set_size(const char *file,
							 nestFileSize  size);

	NestReplyStatus set_transfer_flag(const char *file,
									  bool flag);

	NestReplyStatus get_perm(const char *file,
							 const nestUserOrGroup *ug,
							 nest_access_rights_t *result);

	NestReplyStatus set_perm(const char *file,
							 const nestUserOrGroup *ug,
							 nest_access_rights_t argument);

	NestReplyStatus get_acl(NestACL **result,
							const char *file,
							int is_user);

	NestReplyStatus get_file_classad(const char *file,
									 ClassAd** file_classad);

	NestReplyStatus get_file_uniquifier(const char *physical_filename,
										NestFilehandle* uniquifier );

	NestReplyStatus get_parent_dir(char *dest,
								   const char *src);

	NestReplyStatus get_dest_file(char *dest,
								  const char *src,
								  const nestUser *user);

	bool    is_system_dir_access_allowed(const char *physical_dirname,
										 user_id_t uid,
										 user_id_t owner_uid,
										 nest_access_rights_t flags);

	bool    is_system_file_access_allowed(const char *physical_filename,
										  user_id_t uid,
										  user_id_t owner_uid,
										  nest_access_rights_t flags);

	bool    is_system_entry_access_allowed(const char *physical_filename,
										   user_id_t uid,
										   nest_access_rights_t flags);

	bool is_anonymous_access_allowed(const char *physical_filename,
									 nest_access_rights_t flags);


	bool is_system_access_allowed(const char *file,
								  const nestUser *user,
								  nest_access_rights_t flags);

	bool is_access_allowed(const char *file,
						   const nestUser *user,
						   nest_access_rights_t flags);
	// Same as above, but less painful name
	bool access( const char *file,
				 const nestUser *user, 
				 nest_access_rights_t flags)
		{
			return is_access_allowed( file, user, flags );
		};

	bool is_access_allowed_in_parent(const char *file,
									 const nestUser *user,
									 nest_access_rights_t flags);
	// Same as above, but less painful name
	bool paccess( const char *file,
				 const nestUser *user, 
				 nest_access_rights_t flags)
		{
			return is_access_allowed_in_parent( file, user, flags );
		};

	bool is_dir_empty( const char *logical_dir );

	NestReplyStatus system_dir_list( NestFileInfo **result,
									 const char *physical_dir );

	NestReplyStatus dir_list( NestFileInfo **result,
							  const char *logical_dir );

	NestReplyStatus attach_lot_to_file( const char *logical_file,
										const NestLotHandle *handle );

	NestReplyStatus detach_lot_from_file( const char *logical_file,
										  const NestLotHandle *handle );

	NestReplyStatus initialize(ClassAdCollection *resources,
							   const char *rootdir);

	NestReplyStatus recover(ClassAdCollection *resources,
							const char *rootdir);

	NestReplyStatus synchronize();
	NestReplyStatus shutdown();


  private:
	ClassAdCollection       *file_catalog;

	ClassAdParser           parser;
	ClassAdUnParser         unparser;
	LocalCollectionQuery    query;

	nestUser				*everybody;

	string                  root_dir;
	ViewName                file_catalog_view;

	NestReplyStatus get_physical_name(string &physical_name,
									  string logical_name,
									  const nestUser *user,
									  string lot_handle);

	bool            is_access_allowed_for_entry(const char *file,
												const nestUser *user,
												nest_access_rights_t flags);


};

extern nestFileCatalog	*file_catalog;

#endif
