/***************************Copyright-DO-NOT-REMOVE-THIS-LINE**
  *
  * Condor Software Copyright Notice
  * Copyright (C) 1990-2005, Condor Team, Computer Sciences Department,
  * University of Wisconsin-Madison, WI.
  *
  * This source code is covered by the Condor Public License, which can
  * be found in the accompanying LICENSE.TXT file, or online at
  * www.condorproject.org.
  *
  * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
  * AND THE UNIVERSITY OF WISCONSIN-MADISON "AS IS" AND ANY EXPRESS OR
  * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
  * WARRANTIES OF MERCHANTABILITY, OF SATISFACTORY QUALITY, AND FITNESS
  * FOR A PARTICULAR PURPOSE OR USE ARE DISCLAIMED. THE COPYRIGHT
  * HOLDERS AND CONTRIBUTORS AND THE UNIVERSITY OF WISCONSIN-MADISON
  * MAKE NO MAKE NO REPRESENTATION THAT THE SOFTWARE, MODIFICATIONS,
  * ENHANCEMENTS OR DERIVATIVE WORKS THEREOF, WILL NOT INFRINGE ANY
  * PATENT, COPYRIGHT, TRADEMARK, TRADE SECRET OR OTHER PROPRIETARY
  * RIGHT.
  *
  ****************************Copyright-DO-NOT-REMOVE-THIS-LINE**/

#ifndef    _NEST_IDENTITY_MANAGER_H_

#define    _NEST_IDENTITY_MANAGER_H_

#include "nest_user.h"
#include "nest_common.h"

class nestIdentityManager
{

  public:
	nestIdentityManager( );
	~nestIdentityManager();

	NestReplyStatus		set_userlist( const char *userlist );
	NestReplyStatus		set_denylist ( const char *denylist );

	NestReplyStatus     get_total_nest_users(int *total_nest_users);
	NestReplyStatus     get_total_nest_groups(int *total_nest_groups);
	bool                is_user_present(const nestUser *user);
	bool                is_group_present(const nestGroup *group);
	bool                is_active_group(const nestGroup *group);
	bool                is_valid_nest_user(const nestUser *user,
										   bool is_init = false);

	bool                is_user_access_allowed(const nestUser *user);

	bool                is_group_access_allowed(const nestGroup *group,
												const nestUser *user,
												bool allow_members = false);

	NestReplyStatus     get_best_group(nestGroup *group,
									   const nestUser *user);

	NestReplyStatus     get_group_owner(const nestGroup *group,
										nestUser *owner);

	NestReplyStatus     set_default_group(const nestUser *user,
										  const nestGroup *group,
										  bool test_presence = true);

	NestReplyStatus     reset_default_group(const nestUser *user,
											const nestGroup *group);

	NestReplyStatus     add_user_in_group(const nestUser *cur_user,
										  const nestUser *user,
										  const nestGroup *group );

	NestReplyStatus     delete_user_in_group(const nestUser *cur_user,
											 const nestUser *user,
											 const nestGroup *group );

	NestReplyStatus     add_user( const nestUser *new_user,
								  bool log_existing = true );

	NestReplyStatus     update_usage(const nestUserOrGroup *name,
									 nestLotSize more_size);

	NestReplyStatus     delete_user(const nestUser *user,
									bool log_existing = true );

	NestReplyStatus     add_group( const nestUser *cur_user,
								   const nestGroup *group );

	NestReplyStatus     delete_group(const nestUser *cur_user,
									 const nestGroup *group );
									 

	NestReplyStatus     get_uid(user_id_t *uid, const nestUser *user);
	NestReplyStatus     get_gid(user_id_t *gid, const nestGroup *group);
	NestReplyStatus     get_space_used(nestLotSize *size, 
									   const nestUserOrGroup *ug ); 

	NestReplyStatus     initialize(ClassAdCollection *resources);
	NestReplyStatus     recover(ClassAdCollection *resources);

	NestReplyStatus     shutdown();
	NestReplyStatus     synchronize(bool is_init = false);

  private:

	ClassAdCollection       *Identities;

	ViewName                identities_view;
	ClassAdParser           parser;
	ClassAdUnParser         unparser;
	LocalCollectionQuery    query;

	string                  prev_userlist;
	string                  userlist;
	string                  prev_denylist;
	string                  denylist;

	NestReplyStatus   get_actual_nest_group(char *actual_nest_group);

	bool              is_user_in_group(const nestUser *user,
									   const nestGroup *group);

	bool              is_group_in_user(const nestUser *user,
									   const nestGroup *group);

	NestReplyStatus   add_user_to_group_members(const nestUser *user,
												const nestGroup *group);

	NestReplyStatus   add_group_to_user_groups(const nestUser *user,
											   const nestGroup *group);

	NestReplyStatus   delete_user_from_group_members(const nestUser *user,
													 const nestGroup *group);

	NestReplyStatus   delete_group_from_user_groups(const nestUser *user,
													const nestGroup *group);

	bool              is_present_in_list(const char *userpool,
										 const nestUser *user);

};

extern nestIdentityManager	*identity_manager;

#endif
