/***************************Copyright-DO-NOT-REMOVE-THIS-LINE**
  *
  * Condor Software Copyright Notice
  * Copyright (C) 1990-2005, Condor Team, Computer Sciences Department,
  * University of Wisconsin-Madison, WI.
  *
  * This source code is covered by the Condor Public License, which can
  * be found in the accompanying LICENSE.TXT file, or online at
  * www.condorproject.org.
  *
  * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
  * AND THE UNIVERSITY OF WISCONSIN-MADISON "AS IS" AND ANY EXPRESS OR
  * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
  * WARRANTIES OF MERCHANTABILITY, OF SATISFACTORY QUALITY, AND FITNESS
  * FOR A PARTICULAR PURPOSE OR USE ARE DISCLAIMED. THE COPYRIGHT
  * HOLDERS AND CONTRIBUTORS AND THE UNIVERSITY OF WISCONSIN-MADISON
  * MAKE NO MAKE NO REPRESENTATION THAT THE SOFTWARE, MODIFICATIONS,
  * ENHANCEMENTS OR DERIVATIVE WORKS THEREOF, WILL NOT INFRINGE ANY
  * PATENT, COPYRIGHT, TRADEMARK, TRADE SECRET OR OTHER PROPRIETARY
  * RIGHT.
  *
  ****************************Copyright-DO-NOT-REMOVE-THIS-LINE**/

#ifndef _NEST_IO_MANAGER_H
#define _NEST_IO_MANAGER_H

#include "nest_common.h"
#include "nest_client.h"
#include "nest_client_pool.h"

// Because of interdependency hell, we need to forward define these...
struct ThreadData;

class nestIoManager
{

  public:

	nestIoManager ( void );
	~nestIoManager ( void );

    int init( void );

    void sigChild( int signal, siginfo_t *info, void *context );

    void makeProgress( );

    void taskDone( nestClient *client, NestReplyStatus status );
    
    NestReplyStatus addTask( nestClient *client );
    
    nestClient *getFrontTask( void );

	void addDoneTask( nestClient * );

	void signalNest( int thread_id );

	NestReplyStatus workerRequestHandler(
		int worker_id,
		nestClient *client );

    void debugLevel( long level );
    
    int processFunction( int id );

    ssize_t writeAd( nestProtocol *ad );

	int getMaxFd( void ) { return maxfd; };

  private:
    void setUpModelList( );

    nestClient *selectFavoriteProtocol( );
    nestClient *selectShortestFile( );

    nestClient *selectLowestStride( );
    unsigned long findLowestStride( );
    int makeSchedule();
    NestReplyStatus makeStrides();
    void strideDone( nestClient *, long len );

    NestReplyStatus lockMutex();
    NestReplyStatus unlockMutex();

    
    nestClient *selectQueuedTask( );

    NestReplyStatus startTask( nestClient *client );
    
	// Statistics
    double			bandwidth;
    double			total_bandwidth;
    unsigned long	kbsIn;
    unsigned long	kbsOut;
    unsigned long	transfers;
    unsigned long	errors;
    int				tickets;


	typedef struct
	{
		int current_tasks;
		unsigned long kbs_consumed;
		int stride;
	} ProtocolStats;
	ProtocolStats protocol_stats[PROTOCOL_COUNT];

	int admission_control;
	nestClientPool *task_queue;
	//pthread_mutex_t *mutex;

	int max_concurrency;
	int current_tasks;
	unsigned long ave_concurrency;
	unsigned long total_sizes;

	unsigned long unscheduled_tasks;
	unsigned long fifoscheduled_tasks;

	int		selection_criteria;
	int 	scheduling_policy;

	// TODO  make it so using favorites does not require admit control
	// keep seperate count of current favorites
	// only allow others if current favorites <= threshold (suggest 0)
	// have seperate function called isFavorite( nestClient *)

	int favoriteProtocol;

	// *** From POT ***

	// These need to be public so that they can be accessed by the threads
    pthread_mutex_t			mutex;
    int						model_id;
    int						server_pipe[2];
    nestClientPool			*doneTasks;
    nestClientPool			*pendingTasks;

    NestReplyStatus clearNestSignal( );
    
    int poolSize;
	ThreadData	*first_thread;

    int maxfd;

};

extern nestIoManager	*io_manager;

#if 0
struct ThreadData;
class nestConcurrencyPot : public nestConcurrencyModel
{
  public:
    
    nestConcurrencyPot(
		int pool_size,
		fd_set *readfds,
		fd_set *writefds,
		int id );
    
    virtual ~nestConcurrencyPot( void );
    
    virtual NestReplyStatus addTask( nestClient *client );
    
    virtual NestReplyStatus makeProgress( fd_set *r_ready, fd_set *w_ready );

    bool directCallback();

};
#endif

#endif
