/***************************Copyright-DO-NOT-REMOVE-THIS-LINE**
  *
  * Condor Software Copyright Notice
  * Copyright (C) 1990-2005, Condor Team, Computer Sciences Department,
  * University of Wisconsin-Madison, WI.
  *
  * This source code is covered by the Condor Public License, which can
  * be found in the accompanying LICENSE.TXT file, or online at
  * www.condorproject.org.
  *
  * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
  * AND THE UNIVERSITY OF WISCONSIN-MADISON "AS IS" AND ANY EXPRESS OR
  * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
  * WARRANTIES OF MERCHANTABILITY, OF SATISFACTORY QUALITY, AND FITNESS
  * FOR A PARTICULAR PURPOSE OR USE ARE DISCLAIMED. THE COPYRIGHT
  * HOLDERS AND CONTRIBUTORS AND THE UNIVERSITY OF WISCONSIN-MADISON
  * MAKE NO MAKE NO REPRESENTATION THAT THE SOFTWARE, MODIFICATIONS,
  * ENHANCEMENTS OR DERIVATIVE WORKS THEREOF, WILL NOT INFRINGE ANY
  * PATENT, COPYRIGHT, TRADEMARK, TRADE SECRET OR OTHER PROPRIETARY
  * RIGHT.
  *
  ****************************Copyright-DO-NOT-REMOVE-THIS-LINE**/


#include "nest_lot_manager.h"
/*
    this is the part I dont like - including nest_storage_manager.h

    used to terminate relevant files while terminating Lots - a round-about,
    but cleaner alternative is possible! the 'catch' is to catch implicit
    terminate_lot() that might be triggered on a lot request!
*/

#include "nest_storage_manager.h"
#include "nest_config.h"

// A little helper class...
class lotHandleList {
public:
	lotHandleList( NestLotHandle handle, const nestUserOrGroup *owner )
		{
			this->handle = handle;
			this->lot_owner = copyOwner( owner );
			this->next = NULL;
		};
	lotHandleList( string &handle_str, nestUserOrGroup *owner )
		{
			NestString_to_handle(&this->handle, handle_str.c_str() );
			this->lot_owner = copyOwner( owner );
			this->next = NULL;
		};
	~lotHandleList( void )
		{ delete lot_owner; };

	lotHandleList *setNext( lotHandleList *next )
		{ this->next = next; return next; };
	lotHandleList *getNext( void ) const
		{ return next; };
	NestLotHandle *getHandle( void )
		{ return &handle; };
	const nestUserOrGroup *getOwner( void ) const
		{ return lot_owner; };

private:
	nestUserOrGroup *copyOwner( const nestUserOrGroup *owner )
		{
			if ( owner->isUser() ) {
				return new nestUser( dynamic_cast<const nestUser*>( owner) );
			} else {
				return new nestGroup( dynamic_cast<const nestGroup*>( owner) );
			}
		}

	NestLotHandle   handle;
	nestUserOrGroup	*lot_owner;
	lotHandleList	*next;
};


nestLotManager::nestLotManager( void )
{
    this->lot_catalog = NULL;
    this->QuotaMngr = NULL;
    this->DeviceMngr = NULL;
    this->storage_space = 0;
    this->root_dir = NULL;

	this->max_lot_duration = INFINITE_LOT_DURATION;
	this->default_lot_duration = DEFAULT_LOT_DURATION;

}

nestLotManager::~nestLotManager()
{
	if ( this->root_dir ) {
		NestFree( (void*) this->root_dir );
		this->root_dir = NULL;
	}
}

NestReplyStatus
nestLotManager::setup( const char *rootDir,
					   nestLotSize storage_space )
{
    this->storage_space = storage_space;
	this->root_dir = NestStrdup( rootDir );
	if ( ! this->root_dir ) {
		return NEST_OUT_OF_RESOURCES;
	}
    return NEST_SUCCESS;
}

bool
nestLotManager::kernel_quota_enabled()
{
    return has_kernel_quota;
}

//    rsv_id 0 is not used!
rsv_id_t
nestLotManager::get_next_rsv_id()
{

    //  Just to make sure it doesn't loop forever!

    rsv_id_t count = 1;

    while (count++) {
        NestLotHandle   tmp_handle;
        if (NestAdd_id_to_handle(&tmp_handle, this->next_rsv_id) < 0) {
            this->next_rsv_id ++;
            continue;
        }

        char    str_id[NEST_NAMELEN];
        if (NestHandle_to_string(&tmp_handle, str_id, NEST_NAMELEN) < 0) {
            this->next_rsv_id ++;
            continue;
        }

        string str_key;
        str_key = LOT_KEY_PREFIX;
        str_key = str_key + str_id;

        ClassAd *lot_class_ad = this->lot_catalog->GetClassAd(str_key);
        if (!lot_class_ad) {
            this->next_rsv_id++;
			nest_debug( NEST_DEBUG_LOTS,
						"get_next_rsv_id: Returning lot ID %d\n",
						this->next_rsv_id - 1 );
            return (this->next_rsv_id - 1);
        }
        this->next_rsv_id ++;
    }

    return 0;
}

NestReplyStatus
nestLotManager::sync_next_rsv_id()
{
    string str_key = RSV_ID_KEY;

    char str_rsv_id[NEST_NAMELEN];
    NestSnprintf(str_rsv_id, NEST_NAMELEN, "%d", this->next_rsv_id);
    
    string modify_classad = "";
    modify_classad = modify_classad + 
		"[ RsvId = " + str_rsv_id + "]";
    modify_classad =
		" [ Updates = " + modify_classad + " ] ";
    
    ClassAd *mod_classad = parser.ParseClassAd( modify_classad, true );
    if(!mod_classad){
        return NEST_CLASSAD_ERROR;
    }
    if (!this->lot_catalog->ModifyClassAd( str_key, mod_classad )) {
        return NEST_CLASSAD_ERROR;
    }
    
	return NEST_SUCCESS;
}

NestReplyStatus
nestLotManager::set_actual_name( const nestUserOrGroup *perc_name )
{
    string key;

    key = perc_name->isUser() ? USER_KEY_PREFIX : GROUP_KEY_PREFIX;
    key += perc_name->getName( );

    ClassAd *id_ad = this->lot_catalog->GetClassAd(key);

    if (!id_ad) {
        return perc_name->isUser() ?
			NEST_USER_NOT_FOUND : NEST_GROUP_NOT_FOUND;
    }

    string act_name;

    if (!id_ad->EvaluateAttrString( "ActualName", act_name )) {
        return NEST_CLASSAD_ERROR;
    }

	perc_name->setActual( &act_name );
    return NEST_SUCCESS;
}

NestReplyStatus
nestLotManager::LotClassAd_to_NestLot(ClassAd *lot_ad,
                                      NestLotInternal *rstat)
{

    if (!lot_ad || !rstat) {
        return NEST_BAD_PARAMETERS;
    }

	rstat->clean( );

    bool    result = true;
    result &= lot_ad->EvaluateAttrString( "LotHandle",
                                          rstat->lot_handle,
                                          NEST_NAMELEN);

	result &= lot_ad->EvaluateAttrString( "ParentLotHandle",
                                          rstat->parent_lot_handle,
                                          NEST_NAMELEN);

    ExprTree *user_list = lot_ad->Lookup( "LotUsers" );

    if(!user_list) {
        return NEST_CLASSAD_ERROR;
    }
    string str_user_list;
    unparser.Unparse( str_user_list, user_list);
    
    strcpy(rstat->lot_users, str_user_list.c_str());

	string	owner;
    result &= lot_ad->EvaluateAttrString( "Owner", owner );

	bool	is_user;
    result &= lot_ad->EvaluateAttrBool( "IsUser", is_user );
	if ( is_user ) {
		rstat->owner = new nestUser( &owner );
	} else {
		rstat->owner = new nestGroup( &owner );
	}

	char	tmp[64];

    result &= lot_ad->EvaluateAttrString( "Device", rstat->dev, NEST_NAMELEN);

    result &= lot_ad->EvaluateAttrString( "TotalSize", tmp, sizeof(tmp) );
	rstat->total_size = atoll( tmp );

    result &= lot_ad->EvaluateAttrString( "Size", tmp, sizeof(tmp) );
	rstat->size = atoll( tmp );

    result &= lot_ad->EvaluateAttrString( "Used", tmp, sizeof( tmp ) );
	rstat->used = atoll( tmp );

    result &= lot_ad->EvaluateAttrNumber( "StartTime", rstat->start_time );

    result &= lot_ad->EvaluateAttrNumber( "Duration", rstat->duration );

    result &= lot_ad->EvaluateAttrNumber( "LotType", rstat->lot_type );

    result &= lot_ad->EvaluateAttrInt( "Cost", rstat->cost );

    int  time_now = NestTime( NULL );
    rstat->time_left = (rstat->start_time + rstat->duration) - time_now;

    if (rstat->time_left < 0) {
        rstat->time_left = 0;
    }
    else if (rstat->time_left > rstat->duration) {
		rstat->time_left = rstat->duration;
    }

    if (!result) {
        return NEST_CLASSAD_ERROR;
    }

    return NEST_SUCCESS;
}

nestLotSize
nestLotManager::delete_dead_rsvs(nestLotSize req_size,
                                 int is_user,
                                 bool aggressive_cleanup )
{

    /*
        Assumption - the ClassAd collection returns the
            lot-classads with increasing expiry time

        if not - Got to do something extra to find the
        'oldest' dead Lots!

        If aggressive_cleanup is set to true fairness may not be
        maintained! Users already using less than average storage
        space may find their best effort or dead lots being removed.
        If agressive_cleanup is set to false only those lots
        belonging to users using more than average lot space will be
        removed.

    */

    if(aggressive_cleanup) {
        nest_debug( NEST_DEBUG_LOTS,
                    "Doing Aggressive Cleanup,does not try to maintain "
                    "fairness\n");
    } else {
		nest_debug(NEST_DEBUG_LOTS,
				   "Doing Normal Cleanup,tries to maintain fairness\n");
	}

    NestReplyStatus reply;

    nestLotSize more_free_size = 0;
    bool aggressive_user_cleanup, aggressive_group_cleanup;

    int total_nest_users = 0;
	int total_nest_groups = 0;
    double average_user_space  = 0.0;
	double average_group_space = 0.0;
    
    reply = identity_manager->get_total_nest_users( &total_nest_users );
    if(reply != NEST_SUCCESS){
        nest_debug( NEST_DEBUG_LOTS,
                    "Error trying to get total nest users, \
                     falling back to agressive cleanup for users\n");
        aggressive_user_cleanup = true;
    } else {
        aggressive_user_cleanup = false;
		/* 
         * Less one user to account for nest_everybody 
         * Tune the percentage allocated to users as required
         */
        
        average_user_space =
			(storage_space * 0.5) / (total_nest_users - 1.0);
    }

    reply = identity_manager->get_total_nest_groups( &total_nest_groups );
    if(reply != NEST_SUCCESS) {
        nest_debug( NEST_DEBUG_LOTS,
                    "Error trying to get total nest groups, \
                     falling back to agressive cleanup for groups\n");
        aggressive_group_cleanup = true;
    } else {
        aggressive_group_cleanup = false;
		/* 
         * Tune the percentage allocated to groups as required
         */
         average_group_space =
			 (storage_space * 0.5) / (total_nest_groups - 1.0 );
    }
    
    /*
        1)Get total space - storage_space;
        2)Get total users - available from NestIdentitiesView.
        3)Calculate space per user - storage_space / total_users
        4)Retrieve dead lots from only users who use more than the avg
        space

        5)If this fails then look at lots of other users only
        6)If this also fails then return fail else return success
        This requires breaking this function into two
    */

    this->query.Query(this->lot_catalog_view, NULL);
    this->query.ToFirst();

    string query_key;

    if (!this->query.Current(query_key)) {
        return 0;
    }

    ClassAd         *lot_class_ad;
    string          str_handle;
    string          str_lot_owner;
    int             cur_time =NestTime( NULL );
    nestLotSize     size;
    int             start_time;
    bool            lot_is_user;
    int             duration;
    int             lot_type;
    lotHandleList   *dead_lots = NULL;

    do {
        lot_class_ad = this->lot_catalog->GetClassAd(query_key);

        if (!lot_class_ad) {
            nest_debug(NEST_DEBUG_LOTS,
					   "Weird - The Query gave a key"
					   " of a non-existing ClassAd!\n");
            continue;
        }

        bool    result;
		char	tmp[64];
        result = lot_class_ad->EvaluateAttrString( "LotHandle", str_handle );
        result &= lot_class_ad->EvaluateAttrString( "Owner", str_lot_owner );
        result &= lot_class_ad->EvaluateAttrString( "Size", tmp, sizeof(tmp) );
		size = atoll( tmp );
        result &= lot_class_ad->EvaluateAttrNumber( "StartTime", start_time );
        result &= lot_class_ad->EvaluateAttrNumber( "Duration", duration );
        result &= lot_class_ad->EvaluateAttrNumber( "LotType", lot_type);
        result &= lot_class_ad->EvaluateAttrBool( "IsUser", lot_is_user);

        if (!result) {
            continue;
        }

        nestLotSize user_space_used,group_space_used;
		nestUserOrGroup	*lot_owner;
        if(lot_is_user) {
			lot_owner = new nestUser( &str_lot_owner );
        } else {
			lot_owner = new nestGroup( &str_lot_owner );
        }
		identity_manager->get_space_used( &user_space_used, lot_owner );
		nest_debug( NEST_DEBUG_LOTS, "Average user space : %f\n",
					average_user_space);

        if(!aggressive_cleanup && lot_is_user &&
		   user_space_used <= average_user_space){
            /* Show pity on this user - He/She is already using less than
               average user space
            */
			delete lot_owner;
            continue;
        }
        
        if(!aggressive_cleanup && !lot_is_user &&
		   group_space_used <= average_group_space){
            /* Show pity on this group - It is already using less than
               average group space
            */
			delete lot_owner;
            continue;
        }

        /*
            Assuming the ClassAdCollection would return
            lots sorted by their expiration date.
            If not got to do some extra work to terminate the
            "older" lots!
            Though this continues searching, just to make sure,
            which can be changed once its sure how the Collection
            works,
        */

        if ( (lot_type == BE_LOT) ||
             (cur_time > (start_time + duration)) ) {
            /*
                Here is a dead-one! Of course a best effort lot is as good
                as dead!
            */

			lotHandleList *tmp = new lotHandleList( str_handle, lot_owner );
			tmp->setNext( dead_lots );
            dead_lots = tmp;

            more_free_size += size;
            if (more_free_size >= req_size) {
                // thats big-enough!
				delete lot_owner;
                break;
            }
        }
		delete lot_owner;

    } while (this->query.Next(query_key));

    if (more_free_size < req_size) {
		lotHandleList	*cur = dead_lots;
        while ( cur ) {
            lotHandleList	*next = dead_lots->getNext( );
			delete cur;
            cur = next;
        }
        return 0;
    }

    while (dead_lots) {
        lotHandleList *next = dead_lots->getNext( );
        this->terminate_lot( dead_lots->getHandle(), dead_lots->getOwner() );
        delete dead_lots;
        dead_lots = next;
    }

    return more_free_size;
}

NestReplyStatus
nestLotManager::get_lot_and_share(const char *arg,
                                  char *str_lot_handle,
                                  nestLotSize *lot_share)
{
    int i=0;

    while (arg[i] && arg[i]!= ':') {
        str_lot_handle[i] = arg[i];
        i++;
    }
    str_lot_handle[i] = 0;

    if (!arg[i]) {
        *lot_share = 0;
        return NEST_SUCCESS;
    }

    i++;
    sscanf( arg + i, "%lld", lot_share );

    return NEST_SUCCESS;
}

NestReplyStatus
nestLotManager::change_lot_usage(const char *lot_key_arg,
                                 nestLotSize *size,
                                 bool   allocate_as_much)
{

    if ((*size) == 0) {
        return NEST_SUCCESS;
    }

    string  lot_key = lot_key_arg;

    ClassAd *lot_ad = this->lot_catalog->GetClassAd( lot_key );
    if (!lot_ad) {
        return NEST_LOT_NOT_FOUND;
    }

    nestLotSize  lot_size;
    nestLotSize  lot_used;

    bool    result = true;
	char	tmp[64];
    result  = lot_ad->EvaluateAttrString( "Size", tmp, sizeof( tmp ) );
	lot_size = atoll( tmp );
    result &= lot_ad->EvaluateAttrString( "Used", tmp, sizeof( tmp ) );
	lot_used = atoll( tmp );

    if (!result) {
        return NEST_CLASSAD_ERROR;
    }

    if ((allocate_as_much) && (*size > (lot_size - lot_used))) {
        (*size) = lot_size - lot_used;
        if ((*size) == 0) {
            return NEST_LOT_NOT_ENOUGH;
        }
    }

    if (allocate_as_much || (*size) <= (lot_size - lot_used)) {

        lot_used += *size;

        if (lot_used < 0) {
            nest_debug( NEST_DEBUG_LOTS,
                        "Updating lot_used has made it negative (%lld)\n",
                        lot_used );
            lot_used = 0;
        }

        char    str_used[NEST_NAMELEN];
        NestSnprintf(str_used, NEST_NAMELEN, "\"%lld\"", lot_used);

        string modify_classad = "";
        modify_classad = modify_classad +
			" [ Used = " + str_used + " ] ";

        modify_classad =
			" [ Updates = " + modify_classad + " ] ";

		nest_debug( NEST_DEBUG_LOTS, "Updating lot usage: '%s'\n",
					modify_classad.c_str() );
        ClassAd *mod_classad = parser.ParseClassAd( modify_classad, true );

        if (!this->lot_catalog->ModifyClassAd( lot_key, mod_classad )) {
            return NEST_CLASSAD_ERROR;
        }

        return NEST_SUCCESS;
    }

    return NEST_LOT_NOT_ENOUGH;
}

NestReplyStatus
nestLotManager::add_lot_to_file(const char *lot_handle,
                                nestLotSize size,
                                const char *file_key_arg)
{

    string          file_key = file_key_arg;

    if (size == 0) {
        return NEST_SUCCESS;
    }

    nest_debug( NEST_DEBUG_STORAGE,
                "add_lot_to_file %s %lld %s\n",
                lot_handle,
                size,
                file_key_arg );

    ClassAd *file_class_ad = this->lot_catalog->GetClassAd(file_key);
    if (!file_class_ad) {
        return NEST_FILE_NOT_FOUND;
    }

    ExprTree *lot_list = file_class_ad->Lookup( "LotHandle" );
    if (!lot_list) {
        return NEST_CLASSAD_ERROR;
    }

    ExprListIterator    lot_itr((ExprList *) lot_list);
    Value   lot_val;
    char    lot_str[NEST_NAMELEN],
            str_lot_handle[NEST_NAMELEN],
            str_lot_share[NEST_NAMELEN];
    nestLotSize  lot_share;
    string  new_lot_attr = " { ";
    bool    first_attr = true;
    bool    match = false;

    lot_itr.CurrentValue( lot_val );
    if (!lot_val.IsStringValue( lot_str, NEST_NAMELEN ) ||
        !strcmp( lot_str, "")) {

        if (size < 0) {
            return NEST_LOT_INVALID;
        }

        NestSnprintf( str_lot_share, NEST_NAMELEN, ":%lld", size );
        new_lot_attr = new_lot_attr
			+ "\"" + lot_handle + str_lot_share + "\" } ";
        match = true;
        goto write_and_out;
    }

    //  delete lot_list;
    this->get_lot_and_share(lot_str, str_lot_handle, &lot_share);
    if (!strcmp(lot_handle, str_lot_handle)) {
        lot_share += size;
        match = true;
    }

    if (lot_share > 0)  {
        NestSnprintf( str_lot_share, NEST_NAMELEN, ":%lld", lot_share );
        new_lot_attr = new_lot_attr 
                    + "\"" 
                    + str_lot_handle 
                    + str_lot_share 
                    + "\"";
        first_attr = false;
    }

    while ( lot_itr.NextValue( lot_val ) ) {
        if (lot_val.IsStringValue( lot_str, NEST_NAMELEN )) {
            this->get_lot_and_share(lot_str, str_lot_handle, &lot_share);
            if (!strcmp(lot_handle, str_lot_handle)) {
                lot_share += size;
                match = true;
            }
            if (lot_share > 0) {
                NestSnprintf( str_lot_share, NEST_NAMELEN,
							  ":%lld", lot_share );
                if (!first_attr) {
                    new_lot_attr = new_lot_attr + ", ";
                }
                new_lot_attr = new_lot_attr + "\"" +
                               str_lot_handle + str_lot_share + "\"";
                first_attr = false;
            }
        }
    }

    if (!match) {

        if (size < 0) {
            return NEST_LOT_INVALID;
        }

        NestSnprintf( str_lot_share, NEST_NAMELEN, ":%lld", size );
        if (!first_attr) {
            new_lot_attr = new_lot_attr + ", ";
        }
        new_lot_attr = new_lot_attr + "\"" +
                       lot_handle + str_lot_share + "\"";
    }

    new_lot_attr = new_lot_attr + " } ";

write_and_out:
    string modify_classad = "";
    modify_classad = modify_classad +
		" [ LotHandle = " + new_lot_attr + " ] ";
    modify_classad =
		" [ Updates = " + modify_classad + " ] ";

    nest_debug( NEST_DEBUG_STORAGE,
                "Adding Lot handle to file '%s'\n",
                modify_classad.c_str() );

    ClassAd *mod_classad = parser.ParseClassAd( modify_classad, true );

    if (!this->lot_catalog->ModifyClassAd( file_key, mod_classad )) {
        return NEST_CLASSAD_ERROR;
    }

    return NEST_SUCCESS;
}

/*
    tries to find whether the new space required 'size'
    can be fit entirely in one of the existing Lots

    possible better alternative: find whether a combination
    of existing Lots can accomadate the required 'size'
*/
NestReplyStatus
nestLotManager::extend_current_lot(const char *file_key_arg,
                                   nestLotSize size)
{

    string  file_key = file_key_arg;

    ClassAd *file_ad = this->lot_catalog->GetClassAd( file_key );
    if (!file_ad) {
        return NEST_FILE_NOT_FOUND;
    }

    ExprTree *lot_list = file_ad->Lookup( "LotHandle" );
    if (!lot_list) {
        return NEST_CLASSAD_ERROR;
    }

    ExprListIterator    lot_itr((ExprList *) lot_list);

    Value   lot_val;
    char    lot_str[NEST_NAMELEN];

    lot_itr.CurrentValue( lot_val );
    if (!lot_val.IsStringValue( lot_str, NEST_NAMELEN ) ||
        !strcmp( lot_str, "")) {
        //  delete lot_list;
        return NEST_LOT_NOT_FOUND;
    }

    //  delete lot_list;

    char    lot_handle[NEST_NAMELEN];
    nestLotSize  lot_share;
    this->get_lot_and_share(lot_str, lot_handle, &lot_share);

    string  lot_key = LOT_KEY_PREFIX;
    lot_key = lot_key + lot_handle;

    NestReplyStatus reply = this->change_lot_usage(lot_key.c_str(), &size);

    if (reply == NEST_SUCCESS) {
        return this->add_lot_to_file( lot_handle,
                                      size,
                                      file_key_arg);
    }

    while ( lot_itr.NextValue( lot_val ) ) {
        if (lot_val.IsStringValue( lot_str, NEST_NAMELEN )) {

            this->get_lot_and_share(lot_str, lot_handle, &lot_share);

            lot_key = LOT_KEY_PREFIX;
            lot_key = lot_key + lot_handle;

            reply = this->change_lot_usage(lot_key.c_str(), &size);

            if (reply == NEST_SUCCESS) {
                return this->add_lot_to_file( lot_handle,
                                              size,
                                              file_key_arg);
            }
        }
    }

    return NEST_LOT_NOT_ENOUGH;
}


NestReplyStatus
nestLotManager::check_default_lot( const char *file_key_arg,
                                   nestLotSize size,
                                   const char *user_arg )
{

    string  user_key = USER_KEY_PREFIX,
            file_key = file_key_arg;
    user_key = user_key + user_arg;

    ClassAd *user_ad = this->lot_catalog->GetClassAd( user_key );
    if (!user_ad) {
        return NEST_USER_NOT_FOUND;
    }

    string lot_handle;
    if (!user_ad->EvaluateAttrString( "DefaultLotHandle", lot_handle ) ||
        !strcmp( lot_handle.c_str(), "")) {
        return NEST_CLASSAD_ERROR;
    }

    string lot_key = LOT_KEY_PREFIX;
    lot_key = lot_key + lot_handle;

    NestReplyStatus reply = this->change_lot_usage( lot_key.c_str(), &size );

    NEST_REPLY_CHECK(reply);

    return this->add_lot_to_file( lot_handle.c_str(),
                                  size,
                                  file_key.c_str());
}

NestReplyStatus
nestLotManager::reclaim_lot_space( const char *file_key_arg,
                                   nestLotSize size )
{

    string  file_key = file_key_arg;
    nestLotSize  to_be_reclaimed = 0 - size;
	nestLotSize  claim;

    ClassAd *file_ad = this->lot_catalog->GetClassAd( file_key );
    if (!file_ad) {
        return NEST_FILE_NOT_FOUND;
    }

    string  new_lot_attr = " { ";

    ExprTree *lot_list = file_ad->Lookup( "LotHandle" );
    if (!lot_list) {
        return NEST_CLASSAD_ERROR;
    }

    ExprListIterator    lot_itr((ExprList *) lot_list);
    Value   lot_val;
    char    lot_str[NEST_NAMELEN];

    lot_itr.CurrentValue( lot_val );
    if (!lot_val.IsStringValue( lot_str, NEST_NAMELEN ) ||
        !strcmp( lot_str, "")) {
        //  delete lot_list;
        return NEST_LOT_NOT_FOUND;
    }

    //  delete lot_list;

    char    lot_handle[NEST_NAMELEN],
            str_lot_share[NEST_NAMELEN];
    nestLotSize  lot_share;
    bool    first_attr = true;

    this->get_lot_and_share(lot_str, lot_handle, &lot_share);

    string  lot_key = LOT_KEY_PREFIX;
    lot_key = lot_key + lot_handle;

    claim = ( lot_share < to_be_reclaimed ?
			  0 - lot_share : 0 - to_be_reclaimed );
    NestReplyStatus reply = this->change_lot_usage(lot_key.c_str(), &claim);

    if (reply != NEST_SUCCESS) {
        claim = 0;
    } else {
        to_be_reclaimed += claim;
    }

    if (lot_share + claim) {
        NestSnprintf( str_lot_share, NEST_NAMELEN,
					  ":%lld", (lot_share + claim) );
        new_lot_attr =
			new_lot_attr + "\"" + lot_handle + str_lot_share + "\"";
        first_attr = false;
    }

    while ( lot_itr.NextValue( lot_val ) ) {
        if (lot_val.IsStringValue( lot_str, NEST_NAMELEN )) {

            this->get_lot_and_share(lot_str, lot_handle, &lot_share);

            lot_key = LOT_KEY_PREFIX;
            lot_key = lot_key + lot_handle;

            claim = lot_share < to_be_reclaimed ? -lot_share
                                                : -to_be_reclaimed;
            reply = this->change_lot_usage(lot_key.c_str(), &claim);

            if (reply != NEST_SUCCESS) {
                claim = 0;
            } else {
                to_be_reclaimed += claim;
            }

            if (lot_share + claim) {
                NestSnprintf( str_lot_share,
                              NEST_NAMELEN,
                              ":%lld",
                              (lot_share + claim) );
                if (!first_attr) {
                    new_lot_attr = new_lot_attr + ", ";
                }
                new_lot_attr = new_lot_attr + "\"" +
                               lot_handle + str_lot_share + "\"";
                first_attr = false;
            }
        }
    }

    new_lot_attr = new_lot_attr + " } ";

    string modify_classad = "";
    modify_classad = modify_classad + " [ LotHandle = " + new_lot_attr + " ] ";
    modify_classad = " [ Updates = " + modify_classad + " ] " ;
	nest_debug( NEST_DEBUG_LOTS, "Reclaiming lot space: '%s'\n",
				modify_classad.c_str() );

    ClassAd *mod_classad = parser.ParseClassAd( modify_classad, true );

    if (!this->lot_catalog->ModifyClassAd( file_key, mod_classad )) {
        return NEST_CLASSAD_ERROR;
    }

    return NEST_SUCCESS;
}

NestReplyStatus
nestLotManager::update_lot_usage(const char *logical_filename,
                                 nestLotSize size)
{
    NestReplyStatus reply;
    
    reply = update_file_lot_usage(logical_filename, size);
    if ( reply == NEST_LOT_NOT_ENOUGH || reply == NEST_SUCCESS ) {
        return reply;
    }
    
    reply = update_dir_lot_usage(logical_filename, size);
    if ( reply == NEST_LOT_NOT_ENOUGH || reply == NEST_SUCCESS ) {
        return reply;
    }
  
    reply = update_user_lot_usage(logical_filename, size);
    if ( reply == NEST_LOT_NOT_ENOUGH || reply == NEST_SUCCESS ) {
        return reply;
    }
  
    if(!has_kernel_quota) {
        reply = update_group_lot_usage(logical_filename, size);
        if ( reply == NEST_LOT_NOT_ENOUGH || reply == NEST_SUCCESS ) {
            return reply;
        }
    }

    return reply;
}

NestReplyStatus
nestLotManager::update_file_lot_usage(const char *logical_filename,
                                      nestLotSize size)
{
    if(size == 0) {
        return NEST_SUCCESS;
    }
    
    string file_key = STORAGE_OBJECT_KEY_PREFIX;
    file_key = file_key + logical_filename;

    ClassAd *file_class_ad = this->lot_catalog->GetClassAd(file_key);
    
    if (!file_class_ad) {
        return NEST_FILE_NOT_FOUND;
    }
   
    if (size < 0) { 
        return this->reclaim_lot_space( file_key.c_str(),
                                        size );
    }

    string attr = "AttachedLots";
    ExprTree *lot_list = file_class_ad->Lookup( attr );
    if (!lot_list) {
        return NEST_CLASSAD_ERROR;
    }

    ExprListIterator    lot_itr((ExprList *) lot_list);
    Value               lot_val;
    char                lot_str[NEST_NAMELEN];
    NestReplyStatus     reply;
    
    lot_itr.CurrentValue( lot_val );
    if (!lot_val.IsStringValue( lot_str, NEST_NAMELEN ) ||
        !strcmp( lot_str, "")) {
        return NEST_LOT_NOT_FOUND; 
    }

    nestLotSize  req_size;

    do {
        if (!lot_val.IsStringValue( lot_str, NEST_NAMELEN ) ||
            !strcmp( lot_str, "")) {
            continue; 
        }

        req_size = size;
        string lot_key = LOT_KEY_PREFIX;
        lot_key = lot_key + lot_str;
        
        reply = this->change_lot_usage( lot_key.c_str(), &req_size, true );

        if (reply != NEST_SUCCESS) {
            nest_debug( NEST_DEBUG_STORAGE,
                        "change_lot_usage failed - %s %lld %lld (%s)\n",
                        lot_key.c_str(),
                        size,
                        req_size,
                        NestErrorString(reply) );
            continue;
        }
        size -= req_size;
        reply = this->add_lot_to_file( lot_str, req_size, file_key.c_str());
        if (reply != NEST_SUCCESS) {
            //  what the hell! - too much trouble!!
        }
    }while((size > 0) && lot_itr.NextValue( lot_val ));

    if (!size) {
        return NEST_SUCCESS;
    }

    nest_debug( NEST_DEBUG_ALWAYS,
                "update_lot_usage - %s - "
                "You are in big trouble! \n"
                "Lot and File ClassAds have been updated! \n"
                " but finally ran out of Lot Space! \n"
                " you are unprepared for this 'race'! \n",
                logical_filename );

    return NEST_LOT_NOT_ENOUGH;
}

NestReplyStatus
nestLotManager::update_dir_lot_usage(const char *logical_file,
                                 nestLotSize size)
{
    char parent_dir[NEST_MAXPATHLEN];

    NestReplyStatus reply =
		file_catalog->get_parent_dir( parent_dir, logical_file);
    NEST_REPLY_CHECK(reply);
    
    return update_file_lot_usage(parent_dir, size);
}

NestReplyStatus
nestLotManager::update_user_lot_usage(const char *logical_filename,
                                 nestLotSize size)
{

    if (size == 0) {
        return NEST_SUCCESS;
    }

    string file_key = STORAGE_OBJECT_KEY_PREFIX;
    file_key = file_key + logical_filename;

    if (!this->lot_catalog->GetClassAd(file_key)) {
        return NEST_FILE_NOT_FOUND;
    }

    if (size < 0) {
        return this->reclaim_lot_space( file_key.c_str(),
                                        size );
    }

    NestReplyStatus reply = this->extend_current_lot(file_key.c_str(),
                                                     size);
    if (reply == NEST_SUCCESS) {
        /*
            one of the current Lots had enough free space
            - Life is easy -
        */
        return NEST_SUCCESS;
    }

    ClassAd *file_class_ad = this->lot_catalog->GetClassAd(file_key);
    if (!file_class_ad) {
        return NEST_FILE_NOT_FOUND;
    }

    bool    result;

    string  user;

    result = file_class_ad->EvaluateAttrString( "OwnerUser", user );

    if (!result) {
        return NEST_CLASSAD_ERROR;
    }

    if (this->check_default_lot(file_key.c_str(), size, user.c_str())
            == NEST_SUCCESS) {
        /*
            DefaultLot had enough space
        */
        return NEST_SUCCESS;
    }

    string constraint = "";
    constraint = constraint +
		"other.Owner == \"" + user + "\" && " +
		"other.IsUser == true";

    ExprTree *constraint_tree = this->parser.ParseExpression(constraint);
    if (!constraint_tree) {
        return NEST_CLASSAD_ERROR;
    }

    this->query.Query(this->lot_catalog_view, constraint_tree);
    delete constraint_tree;
    this->query.ToFirst();

    string lot_key;
    if (!this->query.Current(lot_key)) {
        return NEST_LOT_NOT_FOUND;
    }

    nestLotSize  req_size;

    do {
        char    lot_handle[NEST_NAMELEN];

        req_size = size;
        reply = this->change_lot_usage( lot_key.c_str(), &req_size, true );

        if (reply != NEST_SUCCESS) {
            nest_debug( NEST_DEBUG_STORAGE,
                        "change_lot_usage failed - %s %lld %lld (%s)\n",
                        lot_key.c_str(),
                        size,
                        req_size,
                        NestErrorString(reply) );
            continue;
        }

        size -= req_size;

        strcpy(lot_handle, lot_key.c_str() + NestStrlen(LOT_KEY_PREFIX));
        reply = this->add_lot_to_file( lot_handle,
                                       req_size,
                                       file_key.c_str());

        if (reply != NEST_SUCCESS) {
            //  what the hell! - too much trouble!!
            break;
        }

    } while (this->query.Next( lot_key ) && (size > 0));

    if (!size) {
        return NEST_SUCCESS;
    }

    nest_debug( NEST_DEBUG_ALWAYS,
                "update_lot_usage (user) - %s - "
                "You are in big trouble! \n"
                "Lot and File ClassAds have been updated! \n"
                " but finally ran out of Lot Space! \n"
                " you are unprepared for this 'race'! \n",
                user.c_str() );

    return NEST_LOT_NOT_ENOUGH;
}

NestReplyStatus
nestLotManager::update_group_lot_usage(const char *logical_filename,
                                       nestLotSize size)
{
    if (size == 0) {
        return NEST_SUCCESS;
    }

    string file_key = STORAGE_OBJECT_KEY_PREFIX;
    file_key = file_key + logical_filename;

    ClassAd *file_class_ad = this->lot_catalog->GetClassAd(file_key);
    if (!file_class_ad) {
        return NEST_FILE_NOT_FOUND;
    }

    if (size < 0) {
        return this->reclaim_lot_space( file_key.c_str(),
                                        size );
    }

    bool    result;
    string  user_str;
    result = file_class_ad->EvaluateAttrString( "OwnerUser", user_str );

    if (!result) {
        return NEST_CLASSAD_ERROR;
    }
    
    string user_key = USER_KEY_PREFIX;
    user_key = user_key + user_str;
    ClassAd *user_class_ad = lot_catalog->GetClassAd(user_key);
    
    if(!user_class_ad) {
        return NEST_CLASSAD_ERROR;
    }
    
    string attr("Groups");
    ExprTree *group_list = user_class_ad->Lookup( attr );
    if (!group_list) {
        return NEST_CLASSAD_ERROR;
    }

    ExprListIterator    group_itr((ExprList *) group_list);
    Value               group_val;
    nestLotSize              req_size;

    group_itr.CurrentValue( group_val );
    
    do {
		string		group_str;

        if (!group_val.IsStringValue( group_str ) ) {
            return NEST_CLASSAD_ERROR;
		}
		if ( group_str.length() == 0 ) {
            return NEST_CLASSAD_ERROR;
		}
		nestGroup	group( &group_str );

        if( group.match( NEST_NO_GROUP) )
            continue;
        
        NestLotListInternal *list = NULL, *tmp;
		nestUser	user( &user_str );
        NestReplyStatus reply = this->list_lots( &list, &user, &group );
    
        if (reply != NEST_SUCCESS && reply != NEST_LOT_NOT_FOUND) {
            nest_debug( NEST_DEBUG_LOTS,
                        "list_lots( %s ) failed - %s\n",
                        group_str.c_str(),
                        NestErrorString( reply ) );
            return reply;
        }

        if( reply == NEST_LOT_NOT_FOUND ) {
            continue;
        }
        
        while (list) {
            tmp = list->next;
            
            req_size = size;
            string lot_key = LOT_KEY_PREFIX;
            lot_key = lot_key + list->stat.lot_handle;

            reply = this->change_lot_usage( lot_key.c_str(), &req_size, true );
            
            if (reply != NEST_SUCCESS) {
                nest_debug( NEST_DEBUG_STORAGE,
                        "change_lot_usage failed - %s %lld %lld (%s)\n",
                        lot_key.c_str(),
                        size,
                        req_size,
                        NestErrorString(reply) );
                continue;
            }
            size -= req_size;
            reply = this->add_lot_to_file( list->stat.lot_handle, 
                                           req_size, 
                                           file_key.c_str()
                                         );
            if (reply != NEST_SUCCESS) {
                //  what the hell! - too much trouble!!
            }
            
            delete list;
            list = tmp;
        }
        
    }while((size > 0) && group_itr.NextValue( group_val ));

    if (!size) {
        return NEST_SUCCESS;
    }

    nest_debug( NEST_DEBUG_ALWAYS,
                "update_lot_usage (group) - %s - "
                "You are in big trouble! \n" 
                "Lot and File ClassAds have been updated! \n" 
                " but finally ran out of Lot Space! \n"
                " you are unprepared for this 'race'! \n",
                user_str.c_str() );
    
    return NEST_LOT_NOT_ENOUGH;
}

NestReplyStatus
nestLotManager::modify_lot_usage(const nestUser *user,
                                 const char *logical_filename,
                                 const NestLotHandle  *from_lot,
                                 nestLotSize from_share,
                                 const NestLotHandle  *to_lot)
{

    if (!user || !logical_filename || !from_lot || !to_lot) {
        return NEST_BAD_PARAMETERS;
    }

    NestLotInternal lot1, lot2;
    NestReplyStatus reply = stat_lot( from_lot, &lot1 );
    NEST_REPLY_CHECK(reply);
    reply = stat_lot( to_lot, &lot2);
    NEST_REPLY_CHECK(reply);

    if(lot1.owner->isUser() ) {
        if(!is_access_allowed_to_lot(from_lot, user))
            return NEST_NONPRIVELEDGED_USER;
    } else {
        if(!is_access_allowed_to_lot(from_lot, user, IS_GROUP))
            return NEST_NONPRIVELEDGED_USER;
    }
    
    if(lot2.owner->isUser() ) {
        if(!is_access_allowed_to_lot(to_lot, user))
            return NEST_NONPRIVELEDGED_USER;
    } else {
        if(!is_access_allowed_to_lot(to_lot, user, IS_GROUP))
            return NEST_NONPRIVELEDGED_USER;
    }

    string file_key = STORAGE_OBJECT_KEY_PREFIX;
    file_key = file_key + logical_filename;
    ClassAd *file_ad = this->lot_catalog->GetClassAd(file_key);

    if (!file_ad) {
        return NEST_FILE_NOT_FOUND;
    }

    ExprTree    *lot_list = file_ad->Lookup("LotHandle");
    if (!lot_list) {
        return NEST_LOT_NOT_FOUND;
    }

    ExprListIterator    lot_itr((ExprList *) lot_list);

    Value   lot_val;
    char    lot_str[NEST_MAXPATHLEN];

    lot_itr.CurrentValue( lot_val );
    if (!lot_val.IsStringValue( lot_str, NEST_MAXPATHLEN ) ||
        !strcmp( lot_str, "")) {
        //  delete lot_list;
        return NEST_LOT_NOT_FOUND;
    }

    char    str_from_handle[NEST_NAMELEN],
            str_to_handle[NEST_NAMELEN];

    if (NestHandle_to_string( from_lot, str_from_handle, NEST_NAMELEN) < 0) {
        return NEST_LOT_INVALID;
    }

    if (NestHandle_to_string( to_lot, str_to_handle, NEST_NAMELEN) < 0) {
        return NEST_LOT_INVALID;
    }

    //  delete lot_list;
    bool    valid_from_lot = false;
    char    lot_handle[NEST_NAMELEN];
    nestLotSize  lot_share;
    this->get_lot_and_share(lot_str, lot_handle, &lot_share);

    if (!strcmp(lot_handle, str_from_handle)) {
        if (lot_share < from_share) {
            return NEST_LOT_INVALID;
        }
        else {
            valid_from_lot = true;
            goto update_lot_attrs;
        }
    }

    while ( lot_itr.NextValue( lot_val ) ) {
        if (lot_val.IsStringValue( lot_str, NEST_NAMELEN )) {

            this->get_lot_and_share(lot_str, lot_handle, &lot_share);

            if (!strcmp(lot_handle, str_from_handle)) {
                if (lot_share < from_share) {
                    return NEST_LOT_INVALID;
                }
                else {
                    valid_from_lot = true;
                    goto update_lot_attrs;
                }
            }
        }
    }

update_lot_attrs:
    if (!valid_from_lot) {
        return NEST_LOT_INVALID;
    }

    bool            everything_ok;
    string          lot_key = LOT_KEY_PREFIX;

    lot_key = lot_key + str_to_handle;

    reply = this->change_lot_usage( lot_key.c_str(),
                                    &from_share );

    NEST_REPLY_CHECK(reply);

    reply = this->add_lot_to_file( str_to_handle,
                                   from_share,
                                   file_key.c_str() );
    everything_ok = (reply == NEST_SUCCESS);

    lot_key = LOT_KEY_PREFIX;
    lot_key = lot_key + str_from_handle;

    from_share = -from_share;
    reply = this->change_lot_usage( lot_key.c_str(),
                                    &from_share );
    everything_ok &= (reply == NEST_SUCCESS);

    reply = this->add_lot_to_file( str_from_handle,
                                   from_share,
                                   file_key.c_str() );
    everything_ok &= (reply == NEST_SUCCESS);

    if (!everything_ok) {
        nest_debug( NEST_DEBUG_STORAGE,
                    "Too much trouble - few updates out of all made it"
                    "and few didn't! - how bad!?\n" );
    }

    return NEST_SUCCESS;
}


NestReplyStatus
nestLotManager::set_user_default_lot( const NestLotHandle *handle,
                                      const nestUser *user )
{


    if (!user || !handle) {
        return NEST_BAD_PARAMETERS;
    }

    /*
        Don't allow even AdminUser to set his default to a Lot he doesn't own!
        I mean, don't allow unless you change the way terminate_lot works!

        Then, terminate_lot won't work as expected. terminate_lot
        only lists the Lot owner's files and deletes those present in the
        relevant Lot!

        So, if you allow Admin's files to be present in any Lot, not just the
        ones he owns, then the Lots in which the Admin owns some files will
        not be deleted, both good and bad, until the Admin removes
        those files by himself!
    */

    if (!is_access_allowed_to_lot(handle, user, false , false)) {
        return NEST_NONPRIVELEDGED_USER;
    }

    char    str_lot_handle[NEST_NAMELEN];
    if (NestHandle_to_string(handle, str_lot_handle, NEST_NAMELEN) < 0) {
        return false;
    }

    ViewName    ID_view = NEST_IDENTITIES_VIEW;

    string  key = USER_KEY_PREFIX;
    key += user->getName();

    ClassAd *user_ad = this->lot_catalog->GetClassAd( key );

    if (!user_ad) {
        return NEST_USER_NOT_FOUND;
    }

    bool    valid;
    if ((!user_ad->EvaluateAttrBool( "ValidEntry", valid )) ||
        (!valid)) {
        return NEST_USER_NOT_FOUND;
    }

    string modify_classad = "";
    modify_classad = modify_classad +
		" [ DefaultLotHandle = " + str_lot_handle + " ] ";
    modify_classad =
		" [ Updates = " + modify_classad + " ] ";
	nest_debug( NEST_DEBUG_LOTS, "Setting user default lot: '%s'\n",
				modify_classad.c_str() );

    ClassAd *mod_classad = parser.ParseClassAd( modify_classad, true );

    if (!this->lot_catalog->ModifyClassAd( key, mod_classad )) {
        return NEST_CLASSAD_ERROR;
    }

    return NEST_SUCCESS;
}

NestReplyStatus
nestLotManager::request_lot(NestLotHandle *handle,
                            const NestLotHandle *parent_handle,
                            const nestUserOrGroup *parent_name,
                            const nestUserOrGroup *name,
                            nestLotSize size,
                            int duration,
                            int lot_type,
            			    int cost)
{

    char            dev[NEST_MAXPATHLEN];
    nestLotSize     avl = size;
    int             start_time;
    rsv_id_t        id, par_id;
    NestLotInternal parent_lot;
    NestReplyStatus reply;
    bool            make_it_default = false;

    if (!handle) {
        return NEST_BAD_PARAMETERS;
    }

	// Make sure that the duration is within limits
	if ( duration <= 0 ) {
		duration = default_lot_duration;
	} else if (duration >= max_lot_duration) {
        duration = this->max_lot_duration;
    }

    NestGet_id_from_handle(parent_handle, &par_id);
    if(par_id){
        /* Requesting a sub lot from a parent lot.
           Check the available space and duration.
           Reduce the parent lot
         */

        reply = stat_lot( parent_handle, &parent_lot );
        NEST_REPLY_CHECK(reply);

        if(parent_lot.size - parent_lot.used < size)
            return NEST_LOT_NOT_ENOUGH;

        int  time_now = NestTime( NULL );
        parent_lot.time_left =
			(parent_lot.start_time + parent_lot.duration) - time_now;
        if(parent_lot.time_left < duration)
            /* minus 1 to take care of software delays */
            duration = parent_lot.time_left - 1;        
        if(duration < 0)
            duration = 0;

        reply = update_lot(parent_handle,parent_name,-size,0,0,false); 
        NEST_REPLY_CHECK(reply);
    }
        
    if(lot_type == BE_LOT)
        duration = 0;

    if(!par_id) {

        /*
            get_best_device returns the best device (in 'dev') and

            sets 'avl' to the available space in that device
            right now, it just returns the available space
            in the single device the NeST is managing!
        */

		const NestDevice *disk = DeviceMngr->get_best_device( );
		if ( ! disk ) {
			return NEST_BAD_PARAMETERS;
		}
		if ( name->isUser() ) {
			avl = disk->getAvailableUser( );
		} else {
			avl = disk->getAvailableGroup( );
		}
		strncpy( dev, disk->getDevName(), sizeof( dev ) );
    } else {
        strcpy( dev, parent_lot.dev );
    }

    if ( name->isUser() ) {
		const nestUser	*user = dynamic_cast<const nestUser*>( name );
		assert( user );
        make_it_default = (this->available_user_lot_space( user ) == 0);
    }

    start_time = NestTime( NULL );

    id = this->get_next_rsv_id();
    if (!id) {
        if(par_id)
            update_lot(parent_handle,parent_name,size,0,0,false); 
        return NEST_LOT_LIMIT_REACHED;
    }

    if(!par_id) {
        if (avl < size) {
            /*
                terminate/delete expired lots

                this part will get complicated if we start handling
                multiple devices!
            */

            /* First, try to do a fair cleanup */
            if (this->delete_dead_rsvs(size - avl, name->isUser(), false)
				<= 0) {
                /* Fair cleanup failed. Now try an agressive cleanup */
                if (this->delete_dead_rsvs(size - avl, name->isUser(), true)
					<= 0) {
                    return NEST_DISK_FULL;
                }
            }
        }
    }


	reply = identity_manager->update_usage( name, size );
    NEST_REPLY_CHECK(reply);

	if ( this->set_actual_name( name ) != NEST_SUCCESS ) {
        if(par_id) {
            update_lot(parent_handle,parent_name,size,0,0,false);
		}
        return name->isUser() ? NEST_USER_NOT_FOUND : NEST_GROUP_NOT_FOUND;
	}

#if 0
    char    actual_name[NEST_NAMELEN];
    if (this->get_actual_name(actual_name, name ) != NEST_SUCCESS) {
        if(par_id) {
            update_lot(parent_handle,parent_name,size,0,0,false);
		}
        return name->isUser() ? NEST_USER_NOT_FOUND : NEST_GROUP_NOT_FOUND;
    }
#endif

    if(has_kernel_quota) {
        if (QuotaMngr->incr_quota(name,
                                  dev,
                                  size) != NEST_SUCCESS) {
            if(par_id) {
                update_lot(parent_handle,parent_name,size,0,0,false);
            }
            return NEST_NO_QUOTA_ENFORCEMENT;
        }
    }

    if (NestAdd_id_to_handle(handle, id) < 0) {
        if(has_kernel_quota) { 
            QuotaMngr->decr_quota( name, dev, size);
        }
        if(par_id)
            update_lot(parent_handle,parent_name,size,0,0,false);
        return  NEST_BAD_PARAMETERS;
    }

    string str_key,
           str_lot_classad;

    char   str_id[NEST_NAMELEN],
           str_par_id[NEST_NAMELEN],
           str_is_user[NEST_NAMELEN],
           str_size[NEST_NAMELEN],
           str_start_time[NEST_NAMELEN],
           str_duration[NEST_NAMELEN],
           str_cost[NEST_NAMELEN],
           str_lot_type[NEST_NAMELEN];


    if (NestHandle_to_string(handle, str_id, NEST_NAMELEN) < 0) {
        if(has_kernel_quota) { 
            QuotaMngr->decr_quota( name, dev, size);
        }
        if(par_id)
            update_lot(parent_handle,parent_name,size,0,0,false);
        return NEST_BAD_PARAMETERS;
    }
    if (NestHandle_to_string(parent_handle, str_par_id, NEST_NAMELEN) < 0) {
        if(has_kernel_quota) { 
            QuotaMngr->decr_quota( name, dev, size);
        }
        if(par_id)
            update_lot(parent_handle,parent_name,size,0,0,false);
        return NEST_BAD_PARAMETERS;
    }

    NestSnprintf(str_is_user, NEST_NAMELEN, "%s",
				 name->isUser() ? "true" : "false");
    NestSnprintf(str_size, NEST_NAMELEN, "\"%lld\"", size);
    NestSnprintf(str_start_time, NEST_NAMELEN, "%d", start_time);
    NestSnprintf(str_duration, NEST_NAMELEN, "%d", duration);
    NestSnprintf(str_cost, NEST_NAMELEN, "%d", cost);
    NestSnprintf(str_lot_type, NEST_NAMELEN, "%d", lot_type);

    str_key = LOT_KEY_PREFIX;
    str_key = str_key + str_id;
    str_lot_classad =   "";
    str_lot_classad = str_lot_classad +
		"[  Type = \"" + LOT_KEY_PREFIX + "\"; " +
		"   LotHandle = \"" + str_id + "\"; " +
		"   ParentLotHandle = \"" + str_par_id + "\"; " +    
		"   Owner = \"" + name->getName() + "\"; " +
		"   LotUsers = {};" + 
		"   IsUser = " + str_is_user + "; " +
		"   Device = \"" + dev + "\"; " +
		"   TotalSize = " + str_size + "; " +
		"   Size = " + str_size + "; " +
		"   Used = \"0\"; " +
		"   StartTime = " + str_start_time + "; " +
		"   Duration = " + str_duration + "; " +
		"   Cost = " + str_cost + "; " +
		"   LotType = " + str_lot_type +
		" ]";

    ClassAd *lot_class_ad = parser.ParseClassAd( str_lot_classad, true);
    if (!lot_class_ad) {
        if(has_kernel_quota) { 
            QuotaMngr->decr_quota( name, dev, size);
        }
        if(par_id)
            update_lot(parent_handle,parent_name,size,0,0,false);
        return NEST_CLASSAD_ERROR;
    }

    if (!this->lot_catalog->AddClassAd( str_key, lot_class_ad )) {
        if(has_kernel_quota) { 
            QuotaMngr->decr_quota( name, dev, size);
        }
        if(par_id)
            update_lot(parent_handle,parent_name,size,0,0,false);
        return NEST_CLASSAD_ERROR;
    }

    if(!par_id)
        this->DeviceMngr->incr_usage( dev, size, name->isUser() );

    if (make_it_default && name->isUser() ) {
        this->set_user_default_lot( handle,
									dynamic_cast<const nestUser *>(name) );
    }

    sync_next_rsv_id();
    return NEST_SUCCESS;
}

NestReplyStatus
nestLotManager::request_user_lot(NestLotHandle *handle,
                                 const nestUser *user,
                                 nestLotSize size,
                                 int duration,
                                 const nestUser *lot_user,
								 const nestUserOrGroup *parent_ug,
                                 const NestLotHandle *parent_handle,
                                 int lot_type,
                                 int cost)
{
    NestReplyStatus			reply;
    nestUserOrGroup			*free_ug = NULL;
	const nestUserOrGroup	*parent = NULL;
	NestLotHandle			local_handle;

    if (!NestEnableLots) {
        return NEST_LOT_DISABLED;
    }

    if(!user || !lot_user || !parent_handle) {
        return NEST_BAD_PARAMETERS;
	}

    rsv_id_t rsv_id;
    NestGet_id_from_handle( parent_handle, &rsv_id ); 

    if(rsv_id) {
		if (! check_request_access( parent_handle, lot_user ) ) {
			return NEST_NONPRIVELEDGED_USER;
		}
		parent = free_ug = get_lot_user( parent_handle );

		if ( parent_ug && ( ! parent_ug->match( parent ) )  ) {
			reply = NEST_LOT_INVALID;
			goto error;
		}
	} else {
		if ( ! parent_ug ) {
            // requesting normal lot - not a sub lot
            if(! identity_manager->is_valid_nest_user( lot_user ) ) {
				reply = NEST_USER_NOT_FOUND;
				goto error;
			}
            if(! is_access_allowed_to_user( user, lot_user ) ) {
                reply = NEST_NONPRIVELEDGED_USER;
				goto error;
			}
        } else {
			reply = get_parent_lot( user, lot_user, parent_ug,
									&local_handle, size );
			if ( NEST_SUCCESS != reply ) {
				goto error;
			}
			parent_handle = &local_handle;
			parent = parent_ug;
        }
    }

	reply = this->request_lot(handle,
							  parent_handle,
							  parent,
							  lot_user,
							  size,
							  duration,
							  lot_type,
							  cost);

  error:
	// Free up the User/Group object if we allocated one
	if( free_ug ) {
		delete free_ug;
	}
	return reply;
}

NestReplyStatus
nestLotManager::request_group_lot(NestLotHandle *handle,
                                  const nestUser *user,
                                  nestLotSize size,
                                  int duration,
                                  const nestGroup *lot_group,
                                  const nestUserOrGroup *parent_ug,
                                  const NestLotHandle *parent_handle,
                                  int lot_type,
                				  int cost )
{
    NestReplyStatus			reply;
	nestUserOrGroup			*free_ug = NULL;
	const nestUserOrGroup	*parent = NULL;
	NestLotHandle			local_handle;

    if (!NestEnableLots) {
        return NEST_LOT_DISABLED;
    }

    if(has_kernel_quota) {
        return NEST_NOT_YET_IMPLEMENTED;
	}
    
    if(!user || !parent_handle) {
        return NEST_BAD_PARAMETERS;
    }

    rsv_id_t rsv_id;
    NestGet_id_from_handle( parent_handle, &rsv_id ); 

    if(rsv_id){
		if (! check_request_access( parent_handle, lot_group ) ) {
			return NEST_NONPRIVELEDGED_USER;
		}
		parent = free_ug = get_lot_user( parent_handle );
    } else {
		if ( ! parent_ug ) {
            // requesting normal lot - not a sub lot
            if(! identity_manager->is_group_present( lot_group) ) {
				reply = NEST_GROUP_NOT_FOUND;
                goto error;
			}
            if(!is_access_allowed_to_user( user, lot_group, false ) ) {
                reply = NEST_NONPRIVELEDGED_USER;
                goto error;
			}
        } else {
			reply = get_parent_lot(user, lot_group, parent_ug,
								   &local_handle, size);
			if ( NEST_SUCCESS != reply ) {
				goto error;
			}
			parent_handle = &local_handle;
			parent = parent_ug;
        }
    }
    
    reply = this->request_lot(handle,
							  parent_handle,
							  parent,
							  lot_group,
							  size,
							  duration,
							  lot_type,
							  cost );
  error:
	// Free up the User/Group object if we allocated one
	if( free_ug ) {
		delete free_ug;
	}
	return reply;

}

NestReplyStatus
nestLotManager::get_parent_lot(const nestUser *user,
							   const nestUserOrGroup *lot_ug,
                               const nestUserOrGroup *parent_ug,
                               NestLotHandle *parent_handle,
                               nestLotSize size )
{
    /* 
        Since no parent handle is specified, we need to search all lots
        of parent user and find one which has the required amount of free
        space and also the request access for user lot_user
    */
    NestReplyStatus reply;
    NestLotListInternal *lot_list;
    bool found_lot = false;

	// NRL -- is_user applies to const char *parent_user 
	reply = list_lots( &lot_list, user, parent_ug );
    NEST_REPLY_CHECK(reply);
 
    while(lot_list) {
        if(size <= lot_list->stat.size - lot_list->stat.used) {
            NestLotHandle handle;
            NestString_to_handle(&handle, lot_list->stat.lot_handle);
			reply = check_request_access( &handle, lot_ug );
#if 0
            reply = is_access_allowed_to_request( lot_ug, parent_ug, &handle);
#endif
            if(reply == NEST_SUCCESS) {
                found_lot = true;
                break;
            }
        }
        lot_list = lot_list->next;
    }
   
    if(found_lot) {
        NestString_to_handle(parent_handle, lot_list->stat.lot_handle); 
    }
 
    return found_lot ? NEST_SUCCESS : NEST_LOT_NOT_ENOUGH; 
}

bool
nestLotManager::check_request_access( const NestLotHandle	*parent_handle,
									  const nestUserOrGroup *ug )
{
	NestLotInternal	parent_lot;

	if ( stat_lot( parent_handle, &parent_lot ) != NEST_SUCCESS ) {
		return false;
	}

	return check_request_access( &parent_lot, ug );
}

bool
nestLotManager::check_request_access( const NestLotInternal	*parent_lot,
									  const nestUserOrGroup *ug )
{
	if ( ! parent_lot ) {
		return false;
	}

    string user_list( parent_lot->lot_users );
    string user( ug->getName() );

    if( user_list.find( user ) != string::npos ) {
        return true;
    } else {
        return false;
	}
}

nestUserOrGroup *
nestLotManager::get_lot_user( const NestLotHandle *handle )
{
	NestLotInternal	lot;
	if ( stat_lot( handle, &lot ) != NEST_SUCCESS ) {
		return NULL;
	}

	return get_lot_user( &lot );
}

nestUserOrGroup *
nestLotManager::get_lot_user( const NestLotInternal *lot )
{
	if ( ! lot ) {
		return NULL;
	}
	if ( lot->owner->isUser() ) {
		return new nestUser( dynamic_cast <nestUser *> (lot->owner) );
	} else {
		return new nestGroup( dynamic_cast <nestGroup *> (lot->owner) );
	}

}

bool 
nestLotManager::is_access_allowed_to_user(const nestUser *user,
                                          const nestUserOrGroup *lot_ug,
                                          bool allow_group_members,
                                          bool allow_admin)
{
    if(!user || !lot_ug) {
        return false;
    }

    if( allow_admin && user->isAdmin() ) {
        return true;
	}

	// If the requestee is a user, return OK if it matches the requesting user
    if( lot_ug->isUser() ) {
		return user->match( lot_ug );
    } else {    
		return identity_manager->is_group_access_allowed(
			dynamic_cast <const nestGroup *> (lot_ug),
			user,
			allow_group_members );
    }
}

bool
nestLotManager::is_access_allowed_to_lot(const NestLotHandle *handle,
                                         const nestUserOrGroup *ug,
                                         bool allow_group_members,
                                         bool allow_admin)
{

    if (!handle || !ug) {
        return false;
    }

	// This is safe; isAdmin() will return false for groups
	if ( ug->isAdmin() ) {
        return true;
    }

    char    str_id[NEST_NAMELEN];
    if (NestHandle_to_string(handle, str_id, NEST_NAMELEN) < 0) {
        return false;
    }

    string key = LOT_KEY_PREFIX;
    key += str_id;
    ClassAd *lot_class_ad = this->lot_catalog->GetClassAd(key);
    if (!lot_class_ad) {
        return false;
    }

    string name;
    bool result = lot_class_ad->EvaluateAttrString( "Owner", name);

    if (!result) {
        return false;
    }

	nestUser	tmpuser( &name );
    if( ug->isUser() ) {
		if ( ug->match( &tmpuser ) ) {
            return true;
		} else {
            return false;
        }
    } else {
		nestGroup	group( &name );
		if ( ug->match( &group ) ) {
           return true;
        }
		return identity_manager->is_group_access_allowed(
			dynamic_cast <const nestGroup *>( ug ),
			&tmpuser,
			allow_group_members );
    }

}

NestReplyStatus
nestLotManager::list_sub_lots(NestLotListInternal **result,
                              const NestLotHandle *handle)
{
    char str_id[NEST_NAMELEN];
    if(NestHandle_to_string(handle, str_id, NEST_NAMELEN) < 0) {
        return NEST_BAD_PARAMETERS;
	}
 
    string constraint = "";
    constraint = constraint + 
		"other.ParentLotHandle == \"" + str_id + "\"";

    ExprTree *constraint_tree = this->parser.ParseExpression(constraint);
    if (!constraint_tree) {
        return NEST_CLASSAD_ERROR;
    }

    this->query.Query(this->lot_catalog_view, constraint_tree);
	delete constraint_tree;
    this->query.ToFirst();

    string query_key;
    ClassAd *lot_class_ad;   
    if (!this->query.Current(query_key)) {
        return NEST_LOT_NOT_FOUND;
    }

    *result = NULL;

    do {
        lot_class_ad = this->lot_catalog->GetClassAd( query_key );

        if (!lot_class_ad) {
            nest_debug(NEST_DEBUG_LOTS,
					   "Weird - The Query gave a key of a "
					   "non-existing ClassAd!\n");
            continue;
        }

        string  ad_str;
        unparser.Unparse( ad_str, (ExprTree *) lot_class_ad );

        NestLotListInternal *tmp = new NestLotListInternal;

        this->LotClassAd_to_NestLot( lot_class_ad, &tmp->stat );
        tmp->next = *result;
        *result = tmp;

    } while ( query.Next(query_key) );

    return NEST_SUCCESS;    
}
                              
NestReplyStatus
nestLotManager::list_lots( NestLotListInternal **result,
						   const nestUser *req_user,
						   const nestUserOrGroup *lot_ug )
{
    if (!NestEnableLots) {
        return NEST_LOT_DISABLED;
    }
	if( lot_ug->isGroup() && has_kernel_quota ) {
		return NEST_NOT_YET_IMPLEMENTED;
	}

	// If we're passed a user to check against, do it here
	if( req_user ) {
		if ( lot_ug->isUser() ) {
			if(!identity_manager->is_valid_nest_user( 
				   dynamic_cast< const nestUser *> ( lot_ug ) ) ) {
				return NEST_USER_NOT_FOUND;
			}
		} else {
			if( !identity_manager->is_group_present(
					dynamic_cast< const nestGroup *> ( lot_ug ) ) ) {
				return NEST_GROUP_NOT_FOUND;
			}
		}

		if( ! is_access_allowed_to_user( req_user, lot_ug ) ) {
			return NEST_NONPRIVELEDGED_USER;
		}
	}


	// Do the real work here....
    string str_is_user = lot_ug->isUser() ? "true" : "false" ;
    string constraint = "";
    constraint = constraint +
		"other.Owner == \"" + lot_ug->getName() + "\" &&" +
		"other.IsUser == " + str_is_user;

    ExprTree *constraint_tree = this->parser.ParseExpression(constraint);
    if (!constraint_tree) {
        return NEST_CLASSAD_ERROR;
    }

    this->query.Query(this->lot_catalog_view, constraint_tree);
    delete constraint_tree;
    this->query.ToFirst();

    string query_key;
    ClassAd *lot_class_ad;

    if (!this->query.Current(query_key)) {
        return NEST_LOT_NOT_FOUND;
    }

    *result = NULL;

    do {
        lot_class_ad = this->lot_catalog->GetClassAd( query_key );

        if (!lot_class_ad) {
            nest_debug(NEST_DEBUG_LOTS,
					   "Weird - The Query gave a key of a"
					   " non-existing ClassAd!\n");
            continue;
        }

        string  ad_str;
        unparser.Unparse( ad_str, (ExprTree *) lot_class_ad );

        NestLotListInternal *tmp = new NestLotListInternal;
        this->LotClassAd_to_NestLot( lot_class_ad, &tmp->stat );
        tmp->next = *result;
        *result = tmp;

    } while (query.Next(query_key));

    return NEST_SUCCESS;

}

/*
    Note: more_size, more_duration, more_cost can be negative!

    I didn't name it new_size,... 'coz its the difference from the
    prev value and not new absolute values!
*/
NestReplyStatus
nestLotManager::update_lot(const NestLotHandle *handle,
                           const nestUserOrGroup *ug,
                           nestLotSize more_size,
                           int more_duration,
                           int more_cost,
                           bool update_total_size)
{
    
    NestReplyStatus reply;

    if (!NestEnableLots) {
        return NEST_LOT_DISABLED;
    }

    if (!handle) {
        return NEST_BAD_PARAMETERS;
    }

    NestLotInternal lot;
    reply = stat_lot( handle, &lot );
    NEST_REPLY_CHECK(reply);    

    bool lot_isuser = lot.owner->isUser( );
    if(lot_isuser ) {
        if(!is_access_allowed_to_lot(handle, ug))
            return NEST_NONPRIVELEDGED_USER;
    } else {
        if(!is_access_allowed_to_lot(handle, ug, false))
            return NEST_NONPRIVELEDGED_USER;
    }
    
    nestLotSize avl;

    bool has_parent = false;
    if(strcmp(lot.parent_lot_handle, "0"))
        has_parent = true;
        
    NestLotHandle parent_handle;
    NestLotInternal parent_lot;
    if(has_parent) {
        if( NestString_to_handle(&parent_handle,lot.parent_lot_handle) < 0 ) {
            return NEST_LOT_INVALID;
        }
        reply = stat_lot( &parent_handle, &parent_lot );
        NEST_REPLY_CHECK(reply); 
    }

	if ( this->set_actual_name( lot.owner ) != NEST_SUCCESS ) {
       return lot.owner->isUser() ? NEST_USER_NOT_FOUND : NEST_GROUP_NOT_FOUND;
	}
#if 0    
    char actual_name[NEST_NAMELEN];
    if (this->get_actual_name(actual_name, lot.owner, lot.is_user)
                != NEST_SUCCESS) {
        return lot.is_user ? NEST_USER_NOT_FOUND : NEST_GROUP_NOT_FOUND;
    }
#endif

    if (more_size > 0) {
        if(has_parent) {
            if(parent_lot.size - parent_lot.used < more_size) {
                return NEST_LOT_NOT_ENOUGH;
            }
            reply = update_lot(&parent_handle, parent_lot.owner,
							   -more_size, 0, 0, false);
            NEST_REPLY_CHECK(reply);
        } else {
            if ((avl = this->DeviceMngr->free_space(lot.dev, lot_isuser ))
                 <= 0) {
                return NEST_DISK_FULL;
            }

            if (avl < more_size) {
                if (this->delete_dead_rsvs(more_size - avl,
                                           lot_isuser) <= 0) {
                    return NEST_DISK_FULL;
                }
            }
        }

        if(has_kernel_quota) {
            if (QuotaMngr->incr_quota(lot.owner, lot.dev, more_size)
				!= NEST_SUCCESS) {
                if(has_parent) {
                    update_lot(&parent_handle, parent_lot.owner,
							   more_size, 0, 0, false);        
				}
                return NEST_QUOTA_ENFORCEMENT_ERROR;
            }
        }
    } else {

        //  -ve more_size

        if ((lot.size + more_size) < lot.used) {
            /*
                Trying to decrease below used!
            */
            return NEST_LOT_INVALID;
        }

        if(has_kernel_quota) {
            if (QuotaMngr->decr_quota(lot.owner,
                                      lot.dev,
                                      more_size) != NEST_SUCCESS) {
                return NEST_QUOTA_ENFORCEMENT_ERROR;
            }
        }

        if(has_parent) { 
            /* Now update the parent lot */
            reply = update_lot(&parent_handle, parent_lot.owner, -more_size, 0, 0, false);
            if(reply != NEST_SUCCESS) {
                if(has_kernel_quota) {
                    if (QuotaMngr->incr_quota(lot.owner,
                                              lot.dev,
                                              more_size) != NEST_SUCCESS) {
                        return NEST_QUOTA_ENFORCEMENT_ERROR;
                    }  
                }
            }
        }
    }

    /*
        dont worry, its ok to pass a negative more_size here!
    */
    this->DeviceMngr->incr_usage(lot.dev,
                                 more_size,
                                 lot_isuser );

    char    str_total_size[NEST_NAMELEN],
            str_size[NEST_NAMELEN],
            str_duration[NEST_NAMELEN],
            str_cost[NEST_NAMELEN];

    NestSnprintf( str_total_size, NEST_NAMELEN,
				  "\"%lld\"", lot.total_size + more_size);
    NestSnprintf(str_size, NEST_NAMELEN,
				 "\"%lld\"", lot.size + more_size);

    if(has_parent) {
        int  time_now = NestTime( NULL );
        parent_lot.time_left =
			(parent_lot.start_time + parent_lot.duration) - time_now;
        int time_left = lot.start_time + lot.duration - time_now;
        if(parent_lot.time_left < time_left + more_duration)
            more_duration = parent_lot.time_left - time_left - 1;
        if(more_duration < 0)
            more_duration = 0;
    }

    lot.duration += more_duration;
    if ( lot.duration >= this->max_lot_duration ) {
        lot.duration = this->max_lot_duration;
    }

    if(lot.duration < 0) {
        lot.duration = 0;
	}

    NestSnprintf(str_duration, NEST_NAMELEN, "%d", lot.duration);
    NestSnprintf(str_cost, NEST_NAMELEN, "%d", lot.cost + more_cost);

    string str_modify_classad = "";

    if(update_total_size) {
        str_modify_classad += str_modify_classad +
			"[  TotalSize = " + str_total_size + "; " +
			"   Size = " + str_size + "; " +
			"   Duration = " + str_duration + "; " +
			"   Cost = " + str_cost +
			"]";
    } else {
        str_modify_classad += str_modify_classad +
			"[  Size = " + str_size + "; " +
			"   Duration = " + str_duration + "; " +
			"   Cost = " + str_cost +
			"]";
    }

    str_modify_classad =
		" [ Updates = " + str_modify_classad + " ] ";

    ClassAd *modify_classad = parser.ParseClassAd( str_modify_classad, true );

    if(!modify_classad){
        /*
            Got to undo the incr/decr quota!!
        */

        if (more_size > 0) {
            if(has_kernel_quota) {
                QuotaMngr->decr_quota(lot.owner,
                                      lot.dev,
                                      more_size);
            }
            if(has_parent) {
                 update_lot(&parent_handle, parent_lot.owner, more_size,
							0, 0, false);
            }
        }
        else {
            if(has_parent) {
                update_lot(&parent_handle, parent_lot.owner, more_size,
						   0, 0, false); 
            }
            if(has_kernel_quota) {           
                QuotaMngr->incr_quota(lot.owner, lot.dev, more_size );
            }
        }

        return NEST_CLASSAD_ERROR;
    }

    string key = LOT_KEY_PREFIX;
    key += lot.lot_handle;
	nest_debug(NEST_DEBUG_LOTS,
			   "Modifing collection: key='%s'\n", key.c_str() );

    if (!this->lot_catalog->ModifyClassAd( key, modify_classad )) {
        /*
            Got to undo the incr/decr quota!!
        */

        if (more_size > 0) {
            if(has_kernel_quota) {
                QuotaMngr->decr_quota(lot.owner, lot.dev, more_size );
            }
            if(has_parent) {
                 update_lot(&parent_handle, parent_lot.owner, more_size,
							0, 0, false);
            }
        }
        else {
            if(has_parent) {
                update_lot(&parent_handle, parent_lot.owner, more_size,
						   0, 0, false); 
            }
            if(has_kernel_quota) {
                QuotaMngr->incr_quota( lot.owner, lot.dev, more_size );
            }
        }

        return NEST_CLASSAD_ERROR;
    }
    
	reply = identity_manager->update_usage( lot.owner, more_size);
    if (reply != NEST_SUCCESS) {
        /*
            Got to undo the incr/decr quota!!
        */

        if (more_size > 0) {
            if(has_kernel_quota) {
                QuotaMngr->decr_quota( lot.owner, lot.dev, more_size );
            }
            if(has_parent) {
                update_lot(&parent_handle, parent_lot.owner, more_size,
						   0, 0, false);
            }
        }
        else {
            if(has_parent) {
                update_lot(&parent_handle, parent_lot.owner, more_size,
						   0, 0, false);
            }
            if(has_kernel_quota) { 
                QuotaMngr->incr_quota( lot.owner, lot.dev, more_size);
            }
        }

        return reply;
    }
    return NEST_SUCCESS;
}

NestReplyStatus
nestLotManager::change_lot_type(const NestLotHandle *handle,
                                const nestUser *user,
                                int new_lot_type,
                                int more_duration )
{
    
    NestReplyStatus reply;

    if (!handle) {
        return NEST_BAD_PARAMETERS;
    }

    NestLotInternal lot;
    reply = stat_lot( handle, &lot );
    NEST_REPLY_CHECK(reply);

	bool lot_isuser = lot.owner->isUser( );
    if(lot_isuser) {
        if(!is_access_allowed_to_lot(handle, user))
            return NEST_NONPRIVELEDGED_USER;
    } else {
        if(!is_access_allowed_to_lot(handle, user, IS_GROUP, false))
            return NEST_NONPRIVELEDGED_USER;
    }

    char    str_id[NEST_NAMELEN];
    if (NestHandle_to_string(handle, str_id, NEST_NAMELEN) < 0) {
        return NEST_BAD_PARAMETERS;
    }

    if(lot.lot_type == new_lot_type) {
        return NEST_SUCCESS;
    }

    bool has_parent = false;
    if(strcmp(lot.parent_lot_handle, "0")) {
        has_parent = true;
    }
        
    NestLotHandle parent_handle;
    NestLotInternal parent_lot;
    if(has_parent) {
        if( NestString_to_handle(&parent_handle,lot.parent_lot_handle) < 0 ) {
            return NEST_LOT_INVALID;
        }
        reply = stat_lot( &parent_handle, &parent_lot );
        NEST_REPLY_CHECK(reply); 
    }

#if 0    
    char actual_name[NEST_NAMELEN];
    if (this->get_actual_name(actual_name, lot.owner, lot_isuser)
                != NEST_SUCCESS) {
        return lot_isuser ? NEST_USER_NOT_FOUND : NEST_GROUP_NOT_FOUND;
    }
#endif
	if ( this->set_actual_name( lot.owner ) != NEST_SUCCESS ) {
        return( lot.owner->isUser() ?
				NEST_USER_NOT_FOUND : NEST_GROUP_NOT_FOUND );
	}

    char    str_duration[NEST_NAMELEN];
    char    str_lot_type[NEST_NAMELEN];

    int time_now = NestTime( NULL );
    if(has_parent) {
        parent_lot.time_left =
			(parent_lot.start_time + parent_lot.duration) - time_now;
        if(parent_lot.time_left < more_duration)
            more_duration = parent_lot.time_left - 1;
        if(more_duration < 0 || new_lot_type == BE_LOT)
            more_duration = 0;
    }

    if(new_lot_type == BE_LOT) {
        more_duration = 0;
    }

    if ( more_duration >= max_lot_duration ) {
        more_duration = max_lot_duration;
    }
    lot.duration = time_now - lot.start_time + more_duration;
	// Why not: lot.duration += more_duration; ????
    if ( lot.duration >= max_lot_duration ) {
        lot.duration = max_lot_duration;
    }

    NestSnprintf(str_duration, NEST_NAMELEN, "%d", lot.duration);
    NestSnprintf(str_lot_type, NEST_NAMELEN, "%d", new_lot_type);

    string str_modify_classad = "";

    str_modify_classad += str_modify_classad +
		"[  LotType = " + str_lot_type + "; " +
		"   Duration = " + str_duration + 
		"]";

    str_modify_classad =
		" [ Updates = " + str_modify_classad + " ] ";

    ClassAd *modify_classad = parser.ParseClassAd( str_modify_classad, true );

    if(!modify_classad){
        return NEST_CLASSAD_ERROR;
    }

    string key = LOT_KEY_PREFIX;
    key += lot.lot_handle;
        
    if (!this->lot_catalog->ModifyClassAd( key, modify_classad )) {
        return NEST_CLASSAD_ERROR;
    }

    return NEST_SUCCESS;
}

NestReplyStatus
nestLotManager::stat_lot( const NestLotHandle *handle,
						  NestLotInternal *result )
                          
{
    char str_rsv_id[NEST_NAMELEN];
    if (NestHandle_to_string(handle, str_rsv_id, NEST_NAMELEN) < 0){
        return NEST_BAD_PARAMETERS;
    }
    string key = LOT_KEY_PREFIX;
    key += str_rsv_id;

    ClassAd *lot_class_ad = this->lot_catalog->GetClassAd(key);
    if (!lot_class_ad) {
        return NEST_LOT_NOT_FOUND;
    }

    return this->LotClassAd_to_NestLot( lot_class_ad, result );
}

NestReplyStatus
nestLotManager::stat_lot( const NestLotHandle *handle,
						  const nestUser *user,
						  NestLotInternal *result )
{
    if (!NestEnableLots) {
        return NEST_LOT_DISABLED;
	}

    if (!result || !user || !handle) {
        return NEST_BAD_PARAMETERS;
    }

    NestReplyStatus reply = stat_lot( handle, result);
    NEST_REPLY_CHECK(reply);

    if(result->owner->isUser() ) {
        if (!is_access_allowed_to_lot(handle, user)) {
            return NEST_NONPRIVELEDGED_USER;
        }
    } else {
        if (!is_access_allowed_to_lot(handle, user, IS_GROUP)) {
            return NEST_NONPRIVELEDGED_USER;
        }
    }       
   
    return NEST_SUCCESS;
}

NestReplyStatus
nestLotManager::terminate_lot(const NestLotHandle *handle,
                              const nestUserOrGroup *user,
                              int sched_next)
{

    NestReplyStatus reply;

    if (!handle) {
        return NEST_BAD_PARAMETERS;
    }

    NestLotInternal lot;
    reply = stat_lot( handle, &lot );
    NEST_REPLY_CHECK(reply);

	bool	lot_isuser = lot.owner->isUser( );
    if(lot_isuser) {
        if(!is_access_allowed_to_lot(handle, user))
            return NEST_NONPRIVELEDGED_USER;
    } else {
        if(!is_access_allowed_to_lot(handle, user, false))
            return NEST_NONPRIVELEDGED_USER;
    }

    /* if the lot contains sub lots - delete them first */
    NestLotListInternal *sub_lots;
    reply = list_sub_lots(&sub_lots, handle);
    if(reply != NEST_SUCCESS && reply != NEST_LOT_NOT_FOUND)
        return reply;
       
    if(reply == NEST_SUCCESS){
    /* Sub lots exist */
        while(sub_lots){
            NestLotHandle sub_handle;
            if( NestString_to_handle(&sub_handle,
									 sub_lots->stat.lot_handle) < 0)
                return NEST_LOT_INVALID;
            reply = terminate_lot(&sub_handle, sub_lots->stat.owner,
								  sched_next);
            if(reply != NEST_SUCCESS)
                return reply;
            sub_lots = sub_lots->next; 
        }
    } 

    char str_rsv_id[NEST_NAMELEN];
    if (NestHandle_to_string(handle, str_rsv_id, NEST_NAMELEN) < 0) {
        return NEST_BAD_PARAMETERS;
    }

    /*
        Note: Because of the way the following query is formulated, if the
        admin user tries to delete non-empty lots of any other user, it will
        not succeed; NEST_LOT_NOT_EMPTY will be returned. Since this seemed
        like a nice feedback to admin user, preventing accidental removal of
        lots, I am leaving it the way it is!
        If the behaviour needs to be changed such that admin user can delete 
        non-empty lots of other users just replace "user" with "name" in the
        following query. "name" is the username of the owner of the lot.
    */
    
    string constraint = "";
    constraint = constraint +
		"other.OwnerUser == \"" + user->getName() + "\"";

    ViewName    file_view = NEST_FILE_CATALOG_VIEW;
    ExprTree    *constraint_tree = parser.ParseExpression(constraint);
    this->query.Query(file_view, constraint_tree);
    delete constraint_tree;
    this->query.ToFirst();

    string      query_key;
    ClassAd     *file_ad;
    ExprTree    *lot_list;
    string      str_lot_list;
    char        target[NEST_NAMELEN];

    if (!query.Current(query_key)) {
        goto term_lot;
    }

    NestSnprintf( target, NEST_NAMELEN, "\"%s:", str_rsv_id );

    do {
        file_ad = this->lot_catalog->GetClassAd( query_key );

        if (!file_ad) {
            continue;
        }

        lot_list = file_ad->Lookup( "LotHandle" );
        if (!lot_list) {
            continue;
        }

        str_lot_list = "";
        this->unparser.Unparse( str_lot_list, lot_list );

        if (strstr(str_lot_list.c_str(), target) != NULL) {

            string  str_type;
            if (!file_ad->EvaluateAttrString( "FileType", str_type )) {
                continue;
            }
            string  logical_filename;
            if (!file_ad->EvaluateAttrString( "LogicalName",
											  logical_filename ) ) {
                continue;
            }

            if (strcmp(str_type.c_str(), "dir")) {
                char    filename[NEST_MAXPATHLEN];

                /*
                    I dont like this part at all :(
                    the only part that breaks the clear modularity
                    of the implementation
                */

                file_catalog->get_reverse_logical_name(
					filename, logical_filename.c_str() );

				// NRL -- TODO
				// This dynamic cast may fail under some circumstances.
				// I don't know how to solve this problem....
				const nestUser	*ruser =
					dynamic_cast <const nestUser *>( user );
				if ( ! ruser ) {
					nest_warn( __FILE__, __FUNCTION__,
							   "'%s' isn't a user: can't remove file '%s'"
							   " from lot during lot termination\n'",
							   user->getName(), filename );
				} else {
					reply = storage_manager->deleteFile( ruser, filename, "" );
					if ( reply != NEST_SUCCESS ) {
						nest_warn( __FILE__, __FUNCTION__,
								   "Error removing '%s' as user '%s' from lot"
								   " during lot termination\n",
								   user->getName(), filename );
					}
				}
            }
        }
    } while (this->query.Next(query_key));


term_lot:

    reply = stat_lot( handle, &lot );
    NEST_REPLY_CHECK(reply);

	reply = set_actual_name( lot.owner );
    NEST_REPLY_CHECK(reply);

    if(has_kernel_quota) {
        QuotaMngr->decr_quota( lot.owner, lot.dev, lot.size );
    }

    this->DeviceMngr->decr_usage( lot.dev, lot.size, lot_isuser );

    string key = LOT_KEY_PREFIX;
    key += lot.lot_handle;
      
    if (!this->lot_catalog->RemoveClassAd( key )) {
        nest_debug( NEST_DEBUG_LOTS,
                    "Couldn't remove Lot ClassAd from the collection!\n" );

        return NEST_CLASSAD_ERROR;
    }

	reply = identity_manager->update_usage( lot.owner, -lot.size );
    if(reply != NEST_SUCCESS){
        nest_debug( NEST_DEBUG_LOTS,
                    "Couldn't update user usage - will have incorrect behavior"
                    " during fairness cleanup!\n" );

        return reply;
    }

    /*
        Got to schedule next one if required -
        not in this version, though!
    */

    if (!sched_next) {
        /* Update the parent lot if present*/
        if(strcmp(lot.parent_lot_handle,"0")) {
            NestLotHandle parent_handle;
            NestLotInternal parent_lot;
            if(NestString_to_handle(&parent_handle,
									lot.parent_lot_handle) < 0) {
                return NEST_LOT_UPDATE_FAILED;
            }
            reply = stat_lot( &parent_handle, &parent_lot );
            NEST_REPLY_CHECK(reply);
            reply = update_lot( &parent_handle, parent_lot.owner,
								lot.size, 0, 0, false );
            NEST_REPLY_CHECK(reply);
        }

        return NEST_SUCCESS;
    }

    return NEST_NOT_YET_IMPLEMENTED;
}

nestLotSize
nestLotManager::free_space(int is_user)
{

    return this->DeviceMngr->free_space(is_user);
}

nestLotSize
nestLotManager::total_space( int is_user )
{
    return storage_space;
}

nestLotSize
nestLotManager::available_lot_space(NestLotHandle *handle)
{
    nestLotSize avail_space = 0;
    NestLotInternal lot;
    NestReplyStatus reply = stat_lot( handle, &lot );
    if(reply == NEST_SUCCESS) {
        avail_space = lot.size - lot.used;    
    }
    return avail_space;
}

nestLotSize
nestLotManager::available_lot_space( const nestUser *user,
                                     const char *logical_file )
{
    nestLotSize size;

    size = available_file_lot_space(logical_file);
    if(size >= 0) {
        return size;
    }
    size = available_dir_lot_space(logical_file);
    if(size >= 0) {
        return size;
    }
    size = available_user_lot_space(user);
    if(size >= 0) {
        return size;
    }

    if(!has_kernel_quota) {
        size = available_group_lot_space(user);
        if(size >= 0) {
            return size;
        }
    }
   
    return 0; 
}

nestLotSize
nestLotManager::available_file_lot_space( const char *logical_file )
{
    nestLotSize size = 0;
    bool found_lot = false;
    
    if(!logical_file) {
        return found_lot ? size : -1;
    }
                              
    ClassAd *file_class_ad;
    NestReplyStatus reply =
		file_catalog->get_file_classad(logical_file, &file_class_ad);
    if(reply != NEST_SUCCESS) {
        return found_lot ? size : -1;
    }

    string attr = "AttachedLots";
    ExprTree *lot_list = file_class_ad->Lookup( attr );

    if (!lot_list) {
        return found_lot ? size : -1;
    }

    ExprListIterator    lot_itr((ExprList *) lot_list);
    Value               lot_val;
    char                lot_str[NEST_NAMELEN];
    NestLotHandle       lot_handle;
    
    lot_itr.CurrentValue( lot_val );
    do {
        if (!lot_val.IsStringValue( lot_str, NEST_NAMELEN ) ||
            !strcmp( lot_str, "")) {
            return found_lot ? size : -1;
        } else {
            if(NestString_to_handle(&lot_handle, lot_str) < 0) {
                return found_lot ? size : -1;
            }
            nestLotSize tmp_size = available_lot_space( &lot_handle );
            if(tmp_size < 0) {
                return found_lot ? size : -1;
            } else {
                found_lot = true;
                size += tmp_size;
            }
        }   
    }while ( lot_itr.NextValue( lot_val ) );

    return found_lot ? size : -1;

}
    
nestLotSize
nestLotManager::available_dir_lot_space( const char *logical_file )
{
    char parent_dir[NEST_MAXPATHLEN];

    NestReplyStatus reply =
		file_catalog->get_parent_dir(parent_dir, logical_file);
    if(reply != NEST_SUCCESS) {
       return -1;
    } 
   
    return available_file_lot_space(parent_dir);
}

nestLotSize
nestLotManager::available_user_lot_space(const nestUser *user)
{

    NestLotListInternal *list = NULL,
                *tmp;
    nestLotSize      result = 0;
    NestReplyStatus reply = this->list_lots( &list, user, user);

    if (reply != NEST_SUCCESS) {
        nest_debug( NEST_DEBUG_LOTS,
                    "list_user_lots( %s ) failed - %s\n",
                    user->getName(),
                    NestErrorString( reply ) );
        if(reply == NEST_LOT_NOT_FOUND) {
            return -1;
        } else {
            return 0;
        }
    }

    while (list) {
        tmp = list->next;
        result += (list->stat.size - list->stat.used);
        delete list;
        list = tmp;
    }

    if (result < 0) {
        nest_warn( __FILE__, __FUNCTION__,
				   "Negative available_user_lot_space for user %s\n",
				   user->getName() );
        //result = 0;
    }
    else {
        nest_debug ( NEST_DEBUG_LOTS,
                     "Avail_user_lot_space( %s ) == %lld\n",
                     user->getName(),
                     result );
    }

    return result;
}

nestLotSize
nestLotManager::available_group_lot_space( const nestUser *user )
{
    bool found_lot = false;
    nestLotSize size = 0, tmp_size;
    NestReplyStatus reply;
    
    string key = USER_KEY_PREFIX;
    key +=  user->getName();

    ClassAd *user_class_ad = lot_catalog->GetClassAd(key);
    if (!user_class_ad) {
        return found_lot ? size : -1;
    }
    
    string attr("Groups");
    ExprTree *group_list = user_class_ad->Lookup( attr );
    
    if (!group_list) {
        return found_lot ? size : -1;
    }
    
    ExprListIterator    group_itr((ExprList *) group_list);
    Value               group_val;
    char                group_str[ NEST_NAMELEN ];
    
    group_itr.CurrentValue( group_val );
    do {
        if ( !group_val.IsStringValue( group_str, NEST_NAMELEN ) ||
            !strcmp( group_str, "")) {
            return found_lot ? size : -1;
        }
		if(!strcmp(group_str, NEST_NO_GROUP)) {
            continue;
        } else {
			nestGroup	group( group_str );
            reply = available_group_lot_space( user, &group, &tmp_size );
            if(reply != NEST_SUCCESS && reply != NEST_LOT_NOT_FOUND) {
                return found_lot ? size : -1;
            } else if(reply == NEST_SUCCESS) {
                found_lot = true;
                size += tmp_size;
            }
        }
    }while(group_itr.NextValue( group_val ) );

    return found_lot ? size : -1;
}

NestReplyStatus
nestLotManager::available_group_lot_space(const nestUser *user,
                                          const nestGroup *group,
                                          nestLotSize *size)
{

    NestLotListInternal *list = NULL,
                *tmp;
    nestLotSize      result = 0;

    NestReplyStatus reply = this->list_lots( &list,
											 user,
											 group );

    if (reply != NEST_SUCCESS) {
        nest_debug( NEST_DEBUG_LOTS,
                    "list_group_lots( %s ) failed - %s\n",
                    group->getName(),
                    NestErrorString( reply ) );
        return reply;
    }

    while (list) {
        tmp = list->next;
        result += (list->stat.size - list->stat.used);
        delete list;
        list = tmp;
    }

    if (result < 0) {
        nest_warn( __FILE__, __FUNCTION__,
				   "Negative available_group_lot_space for user %s\n",
                    group->getName() );
        result = 0;
    }
    else {
        nest_debug ( NEST_DEBUG_LOTS,
                     "Avail_group_lot_space( %s ) == %lld\n",
                     group->getName(),
                     result );
    }

    *size = result;
    return NEST_SUCCESS;

}

NestReplyStatus
nestLotManager::add_user_to_lot(const NestLotHandle   *handle,
                                const nestUser  *user,
                                const nestUser  *lot_user)
{
    if(!handle || !user || !lot_user) {
        return NEST_BAD_PARAMETERS;
    }

    NestLotInternal lot;
    NestReplyStatus reply = stat_lot( handle, &lot);
    NEST_REPLY_CHECK(reply);

    if(lot.owner->isUser() ) {
        if(!is_access_allowed_to_lot(handle, user))
            return NEST_NONPRIVELEDGED_USER;
    } else {
        if(!is_access_allowed_to_lot(handle, user, false))
            return NEST_NONPRIVELEDGED_USER;
    }

    string str_user_list(lot.lot_users);
    string str_new_user_list;
    string str_user_name(lot_user->getName() );

    bool need_to_modify = false;
    
    string str_find_param = "\"";
    str_find_param = str_find_param + lot_user->getName() + "\"";
    
    if (str_user_list.find(str_find_param) == string::npos) {
        if (NestAdd_entry_to_list(str_new_user_list,
                                  str_user_list,
                                  str_user_name)
			!= NEST_SUCCESS) {
			return NEST_UNKNOWN_ERROR;
        }
        need_to_modify = true;
    }
    
    string key = LOT_KEY_PREFIX;
    key += lot.lot_handle;

    if(need_to_modify)  {
        string modify_classad = "";
        modify_classad = modify_classad +
			" [ LotUsers = " + str_new_user_list + " ] ";
        modify_classad =
			" [ Updates = " + modify_classad + " ] ";
        ClassAd *mod_classad = parser.ParseClassAd( modify_classad, true );
		nest_debug( NEST_DEBUG_LOTS, "Adding user to lot: '%s'\n",
					modify_classad.c_str() );
        if (!this->lot_catalog->ModifyClassAd( key, mod_classad )) {    
            return NEST_CLASSAD_ERROR;
        }
    }
    
    return NEST_SUCCESS; 

}

NestReplyStatus
nestLotManager::delete_user_from_lot(const NestLotHandle *handle,
                                     const nestUser	 *user, 
                                     const nestUser  *lot_user)
{
    if (!NestEnableLots) {
        return NEST_LOT_DISABLED;
    }

    if(!handle || !user || !lot_user) {
        return NEST_BAD_PARAMETERS;
    }

    NestLotInternal lot;
    NestReplyStatus reply = stat_lot( handle, &lot );
    NEST_REPLY_CHECK(reply);

    if(lot.owner->isUser() ) {
        if(!is_access_allowed_to_lot(handle, user))
            return NEST_NONPRIVELEDGED_USER;
    } else {
        if(!is_access_allowed_to_lot(handle, user, false))
            return NEST_NONPRIVELEDGED_USER;
    }

    string str_user_list(lot.lot_users);
    string str_new_user_list;
    string str_user_name(lot_user->getName() );

    bool need_to_modify = false;
    
    string str_find_param = "\"";
    str_find_param = str_find_param + lot_user->getName() + "\"";
    
    if (str_user_list.find(str_find_param)
             != string::npos) {
        if (NestDelete_entry_from_list(str_new_user_list,
                                  str_user_list,
                                  str_user_name)
                                    != NEST_SUCCESS) {
                        return NEST_UNKNOWN_ERROR;
        }
        need_to_modify = true;
    }
  
    string key = LOT_KEY_PREFIX;
    key += lot.lot_handle;
    
    if(need_to_modify)  {
        string modify_classad = "";
        modify_classad = modify_classad +
                        " [ LotUsers = " + str_new_user_list + " ] ";
        modify_classad = " [ Updates = " + modify_classad + " ] ";
		nest_debug( NEST_DEBUG_LOTS, "Deleting user from lot: '%s'\n",
					modify_classad.c_str() );
        ClassAd *mod_classad = parser.ParseClassAd( modify_classad,
                                                    true );
        if (!this->lot_catalog->ModifyClassAd( key, mod_classad )) {    
            return NEST_CLASSAD_ERROR;
        }
    }
    
    return NEST_SUCCESS; 
}

int
nestLotManager::init()
{

    has_kernel_quota = true;
	const char *str;

    if ( GetConfigValue(Config_MaxLotDuration, &str) && *str != 0) {
        int tmp = atoi( str );
		if ( tmp > 0 ) {
			max_lot_duration = tmp;
		}
    }

    if ( GetConfigValue(Config_DefaultLotDuration, &str) && *str != 0) {
        int tmp = atoi( str );
		if ( tmp > 0 ) {
			default_lot_duration = tmp;
		}
    }

    if( GetConfigValue(Config_DisableKernelQuota, NULL) ) {
        has_kernel_quota = false;        
    }
 
    if(has_kernel_quota) {
        
        char            cwd[NEST_MAXPATHLEN];
        char            dev[NEST_MAXPATHLEN];
 
        strncpy(cwd, this->root_dir, NEST_MAXPATHLEN);
 
        if(!NestPathToDevice(cwd,dev)) {
            nest_die( "No information about the device that holds %s !\n",
                      this->root_dir );
        }
    }
    
    return 0; 
}


NestReplyStatus
nestLotManager::initialize(ClassAdCollection *resources)
{
    char            cwd[NEST_MAXPATHLEN];
    char            dev[NEST_MAXPATHLEN];

    strncpy(cwd, this->root_dir, NEST_MAXPATHLEN);

    nest_debug( NEST_DEBUG_STORAGE,
                "NestDevice is found from cwd %s\n",
                cwd );

    if(!NestPathToDevice(cwd, dev) ) {
        nest_die( "No information about the device that holds %s !\n",
                  this->root_dir );
    }

    nest_debug( NEST_DEBUG_STORAGE, "NestDevice is %s\n", dev );

    NestReplyStatus reply;

	// Initialize the device manager
    this->DeviceMngr = new DeviceManager;
    if ( !this->DeviceMngr ) {
        return NEST_OUT_OF_RESOURCES;
    }
    reply = this->DeviceMngr->initialize( resources, dev );
    NEST_REPLY_CHECK(reply);

	// Total usage that we know about
	const NestDevice *disk = DeviceMngr->get_device_entry( dev );
    nestLotSize  total_user_usage = 0;

	// See if we have quota support..
	reply = QuotaManager::has_kernel_support( disk );
	if(reply == NEST_SUCCESS) {
		has_kernel_quota = true;
	}
	else {
		has_kernel_quota = false;
        nest_warn( __FILE__, __FUNCTION__,
                "Running with out kernel quota support.\n");
	}

	// And, the quota manager
    if( has_kernel_quota ) {
        this->QuotaMngr = new QuotaManager;
        if (!this->QuotaMngr) {
            return NEST_OUT_OF_RESOURCES;
        }

		// Initalize it
        reply = QuotaMngr->initialize( resources, disk, &total_user_usage );
        NEST_REPLY_CHECK(reply);
        if (total_user_usage > 0) {
            nest_warn(
				__FILE__, __FUNCTION__,
				"Device %s mounted at %s seems to have an "
				"already set-up Quota system! (%lld) \n",
				dev,
				cwd,
				total_user_usage );
            nest_warn(
				__FILE__, __FUNCTION__,
				"Unless files owned by valid NeST users, within %s "
				"are deleted, the behavior of NeST Lots are undefined!\n",
				cwd);
        }
    }

	// Update the storage space based on disk capacity & usage
	nestLotSize	capacity = disk->getCapacity( ) + total_user_usage;
    if (  ( storage_space >= 0 ) && ( storage_space < capacity )  ) {
        capacity = storage_space;
    } else {
        storage_space = capacity;
	}

	// Continue..
    this->lot_catalog = resources;
    this->query.Bind(resources);
		
	this->lot_catalog_view = NEST_LOT_CATALOG_VIEW;	

    ViewName    parent_view = "root";

    string view_constraint(""),
           view_rank("other.Size"),
           view_partition_expr("");

    view_constraint = view_constraint +
                      " other.Type == \"" + LOT_KEY_PREFIX + "\" ";
    if (!this->lot_catalog->CreateSubView( this->lot_catalog_view,
                                           parent_view,
                                           view_constraint,
                                           view_rank,
                                           view_partition_expr ) ) {
        nest_debug( NEST_DEBUG_ALWAYS,
                    "nestLotManager Couldn't create view for itself!\n");
        return NEST_CLASSAD_ERROR;
    }

    this->next_rsv_id = 1;
    string str_key = RSV_ID_KEY,
           str_lot_classad = "";

    str_lot_classad = str_lot_classad +  
		"[  Type = \"" + RSV_ID_KEY + "\"; " +
		"   RsvId = 1" +
		" ]";
    
    ClassAd *lot_class_ad = parser.ParseClassAd( str_lot_classad, true);
    if (!lot_class_ad) {
        return NEST_CLASSAD_ERROR;
    }
    if (!this->lot_catalog->AddClassAd( str_key, lot_class_ad )) {
        return NEST_CLASSAD_ERROR;
    }

    return NEST_SUCCESS;
}

NestReplyStatus
nestLotManager::recover(ClassAdCollection *resources)
{
    char            cwd[NEST_MAXPATHLEN];
    char            dev[NEST_MAXPATHLEN];

    strncpy(cwd, this->root_dir, NEST_MAXPATHLEN);

    nest_debug( NEST_DEBUG_STORAGE,
                "NestDevice is found from CWD %s\n",
                cwd );

    if(!NestPathToDevice(cwd,dev)) {
        nest_die( "No information about the device that holds %s !\n",
                  this->root_dir );
    }

    NestReplyStatus reply;

	// Create the device manager
    this->DeviceMngr = new DeviceManager;
    if (!this->DeviceMngr) {
        return NEST_OUT_OF_RESOURCES;
    }
    reply = this->DeviceMngr->recover( resources );
    NEST_REPLY_CHECK(reply);

	// Get the current disk...
	const NestDevice *disk = DeviceMngr->get_device_entry( dev );

	// See if we have quota support..
	reply = QuotaManager::has_kernel_support( disk );
	if(reply == NEST_SUCCESS) {
		has_kernel_quota = true;
        this->QuotaMngr = new QuotaManager;
        if (!this->QuotaMngr) {
            return NEST_OUT_OF_RESOURCES;
        }

        reply = this->QuotaMngr->recover( resources, disk );
        NEST_REPLY_CHECK(reply);
	}
	else {
		has_kernel_quota = false;
        nest_warn( __FILE__, __FUNCTION__,
                "Running with out kernel quota support.\n");
	}

    this->lot_catalog = resources;
    this->query.Bind(resources);
	
	this->lot_catalog_view = NEST_LOT_CATALOG_VIEW;

    string key = RSV_ID_KEY;
    ClassAd *rsv_id_ad = this->lot_catalog->GetClassAd( key );

    if(!rsv_id_ad) {
        return NEST_CLASSAD_ERROR;
    }
    int tmp_rsv_id = 0;
    bool result = rsv_id_ad->EvaluateAttrNumber("RsvId", tmp_rsv_id);
    if(!result) {
        return NEST_CLASSAD_ERROR;
    }
    if(tmp_rsv_id<=0) {
        return NEST_INCONSISTENT_STATE;
    }
    this->next_rsv_id = tmp_rsv_id;
    
	return NEST_SUCCESS;
}

NestReplyStatus
nestLotManager::synchronize()
{
    if(has_kernel_quota) {
        this->QuotaMngr->synchronize();
    }
    this->DeviceMngr->synchronize();
    return NEST_SUCCESS;
}

NestReplyStatus
nestLotManager::shutdown()
{
    if(has_kernel_quota) {
        this->QuotaMngr->shutdown();
    }
    this->DeviceMngr->shutdown();
    return NEST_SUCCESS;
}
