/***************************Copyright-DO-NOT-REMOVE-THIS-LINE**
  *
  * Condor Software Copyright Notice
  * Copyright (C) 1990-2005, Condor Team, Computer Sciences Department,
  * University of Wisconsin-Madison, WI.
  *
  * This source code is covered by the Condor Public License, which can
  * be found in the accompanying LICENSE.TXT file, or online at
  * www.condorproject.org.
  *
  * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
  * AND THE UNIVERSITY OF WISCONSIN-MADISON "AS IS" AND ANY EXPRESS OR
  * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
  * WARRANTIES OF MERCHANTABILITY, OF SATISFACTORY QUALITY, AND FITNESS
  * FOR A PARTICULAR PURPOSE OR USE ARE DISCLAIMED. THE COPYRIGHT
  * HOLDERS AND CONTRIBUTORS AND THE UNIVERSITY OF WISCONSIN-MADISON
  * MAKE NO MAKE NO REPRESENTATION THAT THE SOFTWARE, MODIFICATIONS,
  * ENHANCEMENTS OR DERIVATIVE WORKS THEREOF, WILL NOT INFRINGE ANY
  * PATENT, COPYRIGHT, TRADEMARK, TRADE SECRET OR OTHER PROPRIETARY
  * RIGHT.
  *
  ****************************Copyright-DO-NOT-REMOVE-THIS-LINE**/

#ifndef _NEST_PROTOCOL_H
#define _NEST_PROTOCOL_H

#include "nest_lot.h"
#include "nest_storage_common.h"
#include "dynamic-lard.h"

// forward declaration
class nestRequest;


typedef enum {
       NEST_PROTOCOL,
       HTTP_PROTOCOL,
      PROXY_PROTOCOL,
    STORAGE_PROTOCOL,
  ADVERTISE_PROTOCOL,
} nestProtocolType;

    // TODO: is this kludgey or elegant?
    // don't count DISK_PROTOCOL since we don't actually listen for it
#define PROTOCOL_COUNT 3

class nestAd;
class nestClient;

class nestProtocol {

public:

        // I wish the destructor could be pure virtual
    virtual ~nestProtocol() = 0;

    virtual bool sendErrorMessage( const char *filename,
								   const char *func_name,
								   int error ) = 0;

    virtual bool sendAck(  ) = 0;

    virtual bool sendFilesize( nestFileSize size) = 0;

    virtual int  getType( ) = 0;

    virtual bool hasSocket() = 0;

    virtual int  getSocket() = 0;

        // MUST NOT BLOCK !!!!  This is exec'ed by main server thread.
    virtual nestRequest* receiveRequest( void ) = 0;

    virtual bool sendDirectoryList(  NestFileInfo *list );

    virtual NestReplyStatus sendProxyIndex( DynamicLard *index );
    
    virtual bool sendNestLot( NestLotInternal *rstat);

    virtual bool sendLotHandle( NestLotHandle *handle);

    virtual bool sendAd( nestAd *ad );

    virtual bool sendPwd( const char *pwd );

    virtual bool sendVersion( const char *version );

    virtual bool sendDiskspace( nestLotSize avail, nestLotSize total );

    virtual bool sendFinalAck( int transferType );

    virtual bool sendInitialAck(  );

    virtual bool sendTransferInfo( class nestRequest *req );

    virtual bool sendTimestamp(  );

    virtual bool sendBandwidthStatistics(  );

    virtual NestReplyStatus sendFileinfo( const NestFileInfo *info );

    virtual NestReplyStatus setFileattr( NestFileInfo *info,
										 const char *fname,
										 const void *fattr );
    
    virtual bool sendPhysicalName( const char *physical_name );

    virtual bool sendAcl( const char *user,
                          int is_user,
                          nest_access_rights_t access );
    virtual bool sendAcl( const NestACL *Acl );

    virtual bool sendProxyList( list <char *> *proxy ); 

    virtual bool persistent();

        // we used to use this in concurrency_nob but I think it may not
        // actually be necessary so I've disabled it.  if it makes problems
        // then re-enable
    virtual nestProtocol *clone();

    virtual NestReplyStatus receive_state( int pipe );

    virtual NestReplyStatus send_state( int pipe );


    virtual NestReplyStatus sendFilehandle( NestFileInfo *fi );

        // MUST NOT BLOCK !!!!  This is exec'ed by main server thread.
    virtual NestReplyStatus gssAccept( char **username,
									   char **globus_username );
    
    virtual ssize_t Receive( char *buffer, ssize_t n )= 0;

    virtual ssize_t Receiven( char *buffer, ssize_t n );

    virtual ssize_t Send( const string &buffer );

    virtual ssize_t Send( const char *buffer, ssize_t n ) = 0;

    virtual ssize_t Sendn( const char *buffer, ssize_t n );

    virtual bool endThirdPartyTransfer( int type );

	virtual const char *getName( void ) = 0;

};


#define PROTOCOL_WARN nest_debug( NEST_DEBUG_PROTOCOL, \
        "nest_protocol for type %d: default method used %s: %d\n", \
        getType(), \
        __FILE__, __LINE__ );

#define PROTOCOL_DIE nest_die( \
        "nest_protocol for type %d: default method used %s: %d\n", \
        getType(), \
        __FILE__, __LINE__ );

#include "nest_common.h"

#endif
