/***************************Copyright-DO-NOT-REMOVE-THIS-LINE**
  *
  * Condor Software Copyright Notice
  * Copyright (C) 1990-2005, Condor Team, Computer Sciences Department,
  * University of Wisconsin-Madison, WI.
  *
  * This source code is covered by the Condor Public License, which can
  * be found in the accompanying LICENSE.TXT file, or online at
  * www.condorproject.org.
  *
  * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
  * AND THE UNIVERSITY OF WISCONSIN-MADISON "AS IS" AND ANY EXPRESS OR
  * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
  * WARRANTIES OF MERCHANTABILITY, OF SATISFACTORY QUALITY, AND FITNESS
  * FOR A PARTICULAR PURPOSE OR USE ARE DISCLAIMED. THE COPYRIGHT
  * HOLDERS AND CONTRIBUTORS AND THE UNIVERSITY OF WISCONSIN-MADISON
  * MAKE NO MAKE NO REPRESENTATION THAT THE SOFTWARE, MODIFICATIONS,
  * ENHANCEMENTS OR DERIVATIVE WORKS THEREOF, WILL NOT INFRINGE ANY
  * PATENT, COPYRIGHT, TRADEMARK, TRADE SECRET OR OTHER PROPRIETARY
  * RIGHT.
  *
  ****************************Copyright-DO-NOT-REMOVE-THIS-LINE**/

#include "nest_proxy_client_cache.h"

    
nestProxyClientCache::nestProxyClientCache( ) {
    connections = nest_hash_table_create( 1037, NULL );
}

int destroyPool( const char *key, void *item, void *args ) {
    assert( args == NULL );
    nest_debug(NEST_DEBUG_PROXY, "Tearing down client cache pool for %s\n",key);
    delete( (nestClientPool*)item );
    return 1;
}
    
nestProxyClientCache::~nestProxyClientCache( ) {
    nest_hash_table_iterate( connections, destroyPool, NULL );
    nest_hash_table_delete( connections );
}

int sumPools( const char *key, void *item, void *args ) {
    int this_pool = ((nestClientPool*)item)->size();
    int *sum = (int *)args;
    *sum += this_pool;
    return 1;
}

int
nestProxyClientCache::size( ) {
    int sum = 0;
    nest_hash_table_iterate( connections, sumPools, &sum );
    return sum;
}


NestReplyStatus
nestProxyClientCache::insertProxyClient( const char *location, 
        nestClient *client ) 
{
    nestClientPool *clients = 
        (nestClientPool *)nest_hash_table_lookup( connections, location );
    if ( ! clients ) {
        nest_debug( NEST_DEBUG_PROXY, "Creating new client pool to cache"
                " proxy connections to %s.\n", location );
        clients = new nestClientPool( false );
        assert( clients );
        nest_hash_table_insert( connections, location, (void*)clients );
    }
	nest_debug( NEST_DEBUG_PROXY, "Inserting cached proxy to %s (%d)\n",
			location, client->getId() );
    clients->addClient( client );
    return NEST_SUCCESS;
}
    
nestClient *
nestProxyClientCache::getCachedProxy( const char *location ) 
{
    nestClientPool *clients = 
        (nestClientPool *)nest_hash_table_lookup( connections, location );

    if ( ! clients ) {
        return NULL;
    } else {
		nestClient *client = removeAvailableConnection( clients );
        if ( client != NULL ) {
            nest_debug( NEST_DEBUG_PROXY, "Returning cached proxy to %s (%d)\n",
				location, client->getId() );
        }
		return client;
    }
}

nestClient *
nestProxyClientCache::removeAvailableConnection( nestClientPool *clients ) {

    nestClient *client = NULL;
    client = clients->getFront();
    if ( client != NULL ) assert( ! client->isActive() );
/*
    for( CLIENT_ITERATOR itr=clients->begin(); itr!=clients->end(); *itr++) {
        nestClient *cursor = *itr;
        if ( ! cursor->isActive() ) {
            client = cursor;
            break;
        }
    }
*/
    return client;
}
