/***************************Copyright-DO-NOT-REMOVE-THIS-LINE**
  *
  * Condor Software Copyright Notice
  * Copyright (C) 1990-2005, Condor Team, Computer Sciences Department,
  * University of Wisconsin-Madison, WI.
  *
  * This source code is covered by the Condor Public License, which can
  * be found in the accompanying LICENSE.TXT file, or online at
  * www.condorproject.org.
  *
  * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
  * AND THE UNIVERSITY OF WISCONSIN-MADISON "AS IS" AND ANY EXPRESS OR
  * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
  * WARRANTIES OF MERCHANTABILITY, OF SATISFACTORY QUALITY, AND FITNESS
  * FOR A PARTICULAR PURPOSE OR USE ARE DISCLAIMED. THE COPYRIGHT
  * HOLDERS AND CONTRIBUTORS AND THE UNIVERSITY OF WISCONSIN-MADISON
  * MAKE NO MAKE NO REPRESENTATION THAT THE SOFTWARE, MODIFICATIONS,
  * ENHANCEMENTS OR DERIVATIVE WORKS THEREOF, WILL NOT INFRINGE ANY
  * PATENT, COPYRIGHT, TRADEMARK, TRADE SECRET OR OTHER PROPRIETARY
  * RIGHT.
  *
  ****************************Copyright-DO-NOT-REMOVE-THIS-LINE**/

#ifndef _NEST_PROXY_CLIENT_CACHE_H
#define _NEST_PROXY_CLIENT_CACHE_H

#include "nest_common.h"
#include "nest_client_pool.h"
#include "nest_hash_table.h"
#include "nest_proxy.h"

// this class just maps strings (machine names) to colletions of 
// client objects to that machine.  it assumes that is it used
// synchronously and uses the internal isBusy() flag of a client
// to know whether a client is already being used for a transfer
// it only returns clients that are ! isBusy()
//
// this means this class is definitely not thread safe and is meant to be
// used in the syncronous serialized request handling thread of the dispatcher
// it also looks for not busy clients just by doing a list traversal which
// should be fine for small numbers of clients as in our expected workloads

class
nestProxyClientCache {

public:
    nestProxyClientCache( );

    ~nestProxyClientCache( );

        // if it returns NULL, you need to make one and insert it
        // it removes it from its internal list, when done you need to reinsert
    nestClient * getCachedProxy( const char *location );

    NestReplyStatus insertProxyClient( const char *location, nestClient *client );

    int size();

private:

    struct nest_hash_table *connections;

    nestClient *removeAvailableConnection( nestClientPool *clients );

};

#endif
