/***************************Copyright-DO-NOT-REMOVE-THIS-LINE**
  *
  * Condor Software Copyright Notice
  * Copyright (C) 1990-2005, Condor Team, Computer Sciences Department,
  * University of Wisconsin-Madison, WI.
  *
  * This source code is covered by the Condor Public License, which can
  * be found in the accompanying LICENSE.TXT file, or online at
  * www.condorproject.org.
  *
  * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
  * AND THE UNIVERSITY OF WISCONSIN-MADISON "AS IS" AND ANY EXPRESS OR
  * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
  * WARRANTIES OF MERCHANTABILITY, OF SATISFACTORY QUALITY, AND FITNESS
  * FOR A PARTICULAR PURPOSE OR USE ARE DISCLAIMED. THE COPYRIGHT
  * HOLDERS AND CONTRIBUTORS AND THE UNIVERSITY OF WISCONSIN-MADISON
  * MAKE NO MAKE NO REPRESENTATION THAT THE SOFTWARE, MODIFICATIONS,
  * ENHANCEMENTS OR DERIVATIVE WORKS THEREOF, WILL NOT INFRINGE ANY
  * PATENT, COPYRIGHT, TRADEMARK, TRADE SECRET OR OTHER PROPRIETARY
  * RIGHT.
  *
  ****************************Copyright-DO-NOT-REMOVE-THIS-LINE**/
#include "nest_common.h"
#include "nest_proxy_fetch.h"
#include "nest_client.h"


nestProxyFetch::nestProxyFetch( const char *homefile,
                                const char *proxyfile,
                                const char *shadowfile,
                                long  home_offset,
                                long  home_length,
                                nestClient *client,
                                bool forwarded_request )
{
    assert( homefile   != NULL );
    assert( proxyfile  != NULL );
    assert( shadowfile != NULL );
    assert( client != NULL );

    this->homefile   = NestStrdup( homefile );
    this->proxyfile  = NestStrdup( proxyfile );
    this->shadowfile = NestStrdup( shadowfile ); 

	this->client = client;

    this->home_off = home_offset;
    this->home_len = home_length;

    this->server     = NULL;
    this->attempts   = 0;
    this->forwarded  = forwarded_request;
}

nestProxyFetch::~nestProxyFetch()
{
    if ( homefile )     NestFree( homefile );
    if ( proxyfile )    NestFree( proxyfile );
    if ( shadowfile )   NestFree( shadowfile );
    if ( server )       NestFree( server );
}

void
nestProxyFetch::setServer( const char *server )
{
    if ( this->server != NULL ) NestFree( this->server );
    this->server = NestStrdup( server );
}

bool
nestProxyFetch::wasForwarded()
{
    return forwarded;
}

const char *
nestProxyFetch::getServer()
{
    return server;
}

const char *
nestProxyFetch::getHomeFile()
{
    return homefile;
}

const char *
nestProxyFetch::getProxyFile()
{
    return proxyfile;
}

const char *
nestProxyFetch::getShadowFile()
{
    return shadowfile;
}

const char *
nestProxyFetch::getClientHostName()
{
    return client->getHostName();
}

const char *
nestProxyFetch::getClientHostIdentity()
{
    return client->getHostIdentity();
}

const char *
nestProxyFetch::getClientCwd()
{
    return client->getCwd();
}

const nestClient *
nestProxyFetch::getClient()
{
    return client;
}

const nestUser *
nestProxyFetch::getClientUser()
{
    return client->getUser( );
}

long
nestProxyFetch::getHomeOffset()
{
    return home_off;
}

long
nestProxyFetch::getHomeLength()
{ 
    return home_len;
}

int
nestProxyFetch::getAttempts()
{
    return attempts;
}

void
nestProxyFetch::incrementAttempts()
{
    attempts++;
}
