/***************************Copyright-DO-NOT-REMOVE-THIS-LINE**
  *
  * Condor Software Copyright Notice
  * Copyright (C) 1990-2005, Condor Team, Computer Sciences Department,
  * University of Wisconsin-Madison, WI.
  *
  * This source code is covered by the Condor Public License, which can
  * be found in the accompanying LICENSE.TXT file, or online at
  * www.condorproject.org.
  *
  * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
  * AND THE UNIVERSITY OF WISCONSIN-MADISON "AS IS" AND ANY EXPRESS OR
  * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
  * WARRANTIES OF MERCHANTABILITY, OF SATISFACTORY QUALITY, AND FITNESS
  * FOR A PARTICULAR PURPOSE OR USE ARE DISCLAIMED. THE COPYRIGHT
  * HOLDERS AND CONTRIBUTORS AND THE UNIVERSITY OF WISCONSIN-MADISON
  * MAKE NO MAKE NO REPRESENTATION THAT THE SOFTWARE, MODIFICATIONS,
  * ENHANCEMENTS OR DERIVATIVE WORKS THEREOF, WILL NOT INFRINGE ANY
  * PATENT, COPYRIGHT, TRADEMARK, TRADE SECRET OR OTHER PROPRIETARY
  * RIGHT.
  *
  ****************************Copyright-DO-NOT-REMOVE-THIS-LINE**/

#include "nest_common.h"
#include "nest_pthread_util.h"

int nest_pthread_mutex_init(pthread_mutex_t *mutex, 
                            const pthread_mutexattr_t *attr){
  int ret;
  assert(mutex!=NULL);
  ret = pthread_mutex_init(mutex, attr);
  if(ret != 0){
    nest_perror("pthread_mutex_init");
    assert(0);
  }
  return ret;
}

int nest_pthread_cond_init(pthread_cond_t *cond, 
                           const pthread_condattr_t *attr){
  int ret;
  assert(cond!=NULL);
  ret = pthread_cond_init(cond, attr);
  if(ret != 0){
    nest_perror("pthread_cond_init");
    assert(0);
  }
  return ret;
}

int nest_pthread_cond_destroy(pthread_cond_t *cond){
  int ret;
  assert(cond!=NULL);
  ret = pthread_cond_destroy(cond);
  if(ret!=0){
    nest_perror("pthread_cond_destroy");
    assert(0);
  }
  return 0;
}


int nest_pthread_mutex_lock(pthread_mutex_t *mp){
  int ret;
  assert(mp!=NULL);
  ret = pthread_mutex_lock(mp);
  if(ret!=0){
    nest_perror("pthread_mutex_lock");
    assert(0);
  }
  return ret;
}

int nest_pthread_mutex_unlock(pthread_mutex_t *mp){
  int ret;
  assert(mp!=NULL);
  ret = pthread_mutex_unlock(mp);
  if(ret != 0){
    nest_perror("pthread_mutex_unlock");
    assert(0);
  }
  return ret;
}

int nest_pthread_mutex_destroy(pthread_mutex_t *mp){
  int ret;
  assert(mp!=NULL);
  ret = pthread_mutex_destroy(mp);
  if(ret!=0){
    nest_perror("pthread_mutex_unlock");
    assert(0);
  }
  return ret;
}

int nest_pthread_cond_wait( pthread_cond_t *cvp, 
                            pthread_mutex_t *mp ) {
    int ret;
    assert( cvp != NULL );
    assert( mp  != NULL );

    ret = pthread_cond_wait( cvp, mp );
    if ( ret != 0 ) {
        nest_perror( "pthread_cond_wait" );
        assert( 0 );
    }
    return ret;
}

int nest_pthread_cond_timedwait(pthread_cond_t *cvp, 
                                pthread_mutex_t *mp,
               struct timespec *abstime){
  int ret;
  assert(cvp!=NULL);
  assert(mp!=NULL);
  assert(abstime != NULL);

  ret = pthread_cond_timedwait(cvp, mp, abstime);
  if(ret!=0 && ret != ETIMEDOUT){
    nest_perror("pthread_cond_timedwait");
    assert(0);
  }
  return ret;
}

int nest_pthread_cond_signal(pthread_cond_t *cond){
  int ret;
  assert(cond!=NULL);
  
  ret = pthread_cond_signal(cond);
  if(ret!=0){
    nest_perror("pthread_cond_signal");
    assert(0);
  }
  return ret;
}

int nest_pthread_cond_broadcast(pthread_cond_t *cond){
  int ret;
  assert( cond != NULL );
  ret = pthread_cond_broadcast( cond );
  if(ret != 0) {
    nest_perror("pthread_cond_broadcast");
    assert(0);
  }
  return ret;
}

int nest_pthread_create(pthread_t *new_thread_ID,
           const pthread_attr_t *attr,
           void * (*start_func)(void *), 
           void *arg){
  int ret;
  assert(new_thread_ID != NULL);
  assert(start_func != NULL);
  
  ret = pthread_create(new_thread_ID, attr, start_func, arg);
  if(ret!=0){
    nest_perror("pthread_create");
    assert(0);
  }
  return ret;
}

int nest_pthread_attr_init(pthread_attr_t *attr){
  int ret;
  assert(attr!=NULL);
  
  ret = pthread_attr_init(attr);
  if(ret!=0){
    nest_perror("pthread_attr_init");
    assert(0);
  }
  return ret;
}
int nest_pthread_attr_setscope(pthread_attr_t *attr, 
                               int contentionscope){
  int ret;
  assert(attr!=NULL);
 
  ret = pthread_attr_setscope(attr, contentionscope);
  if(ret!=0){
    nest_perror("pthread_attr_setscope");
    assert(0);
  }
  return ret;
}
int nest_pthread_attr_setschedpolicy(pthread_attr_t *attr, 
                                     int policy){
  int ret;
  assert(attr!=NULL);
  
  ret = pthread_attr_setschedpolicy(attr, policy);
  if(ret!=0 && ret != ENOTSUP){
    nest_perror("pthread_attr_setschedpolicy");
    assert(0);
  }
  return ret;
}




