/***************************Copyright-DO-NOT-REMOVE-THIS-LINE**
  *
  * Condor Software Copyright Notice
  * Copyright (C) 1990-2005, Condor Team, Computer Sciences Department,
  * University of Wisconsin-Madison, WI.
  *
  * This source code is covered by the Condor Public License, which can
  * be found in the accompanying LICENSE.TXT file, or online at
  * www.condorproject.org.
  *
  * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
  * AND THE UNIVERSITY OF WISCONSIN-MADISON "AS IS" AND ANY EXPRESS OR
  * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
  * WARRANTIES OF MERCHANTABILITY, OF SATISFACTORY QUALITY, AND FITNESS
  * FOR A PARTICULAR PURPOSE OR USE ARE DISCLAIMED. THE COPYRIGHT
  * HOLDERS AND CONTRIBUTORS AND THE UNIVERSITY OF WISCONSIN-MADISON
  * MAKE NO MAKE NO REPRESENTATION THAT THE SOFTWARE, MODIFICATIONS,
  * ENHANCEMENTS OR DERIVATIVE WORKS THEREOF, WILL NOT INFRINGE ANY
  * PATENT, COPYRIGHT, TRADEMARK, TRADE SECRET OR OTHER PROPRIETARY
  * RIGHT.
  *
  ****************************Copyright-DO-NOT-REMOVE-THIS-LINE**/

#include "nest_common.h"
#include "nest_pwent.h"

nestPwent::nestPwent( void )
{
	this->pwent.pw_name = NULL;
	this->pwent.pw_passwd = NULL;
	this->pwent.pw_uid = 0;
	this->pwent.pw_gid = 0;
	this->pwent.pw_gecos = NULL;
	this->pwent.pw_dir = NULL;
	this->pwent.pw_shell = NULL;
	this->valid = false;
}

nestPwent::~nestPwent( void )
{
	clear( );
}

void
nestPwent::clear( void )
{
	free( this->pwent.pw_name );
	free( this->pwent.pw_passwd );
	free( this->pwent.pw_gecos );
	free( this->pwent.pw_dir );
	free( this->pwent.pw_shell );
	valid = false;
}

int
nestPwent::findent( user_id_t uid, const char *pwfile )
{
	clear( );

	if ( NULL == pwfile ) {
		pwfile = SystemPwdFile;
	}

	// Open the passwd file
	FILE *fp = fopen( pwfile, "r" );
	if ( ! fp ) {
		return -1;
	}

	// Read it, find the matching entry
    struct passwd *tmppw;
    while ( (tmppw = fgetpwent(fp)) != NULL) {
        if ( (user_id_t) uid == (user_id_t)(tmppw->pw_uid) ) {
			copyent( tmppw );
			break;
        }
    }
	fclose( fp );

	if ( valid ) {
		return 0;
	} else {
		return -1;
	}
}

int
nestPwent::findent( const char *name, const char *pwfile )
{
	clear( );

	if ( NULL == pwfile ) {
		pwfile = SystemPwdFile;
	}

	// Open the passwd file
	FILE *fp = fopen( pwfile, "r" );
	if ( ! fp ) {
		return -1;
	}

	// Read it, find the matching entry
    struct passwd *tmppw;
    while ( (tmppw = fgetpwent(fp)) != NULL) {
        if ( ! strcmp( tmppw->pw_name, name ) ) {
			copyent( tmppw );
			break;
        }
    }
	fclose( fp );

	if ( valid ) {
		return 0;
	} else {
		return -1;
	}
}

const struct passwd *
nestPwent::getent( void ) const
{
	if ( ! valid ) {
		return NULL;
	} else {
		return &pwent;
	}
}

const char *
nestPwent::getent_name( void ) const
{
	if ( ! valid ) {
		return NULL;
	} else {
		return pwent.pw_name;
	}
}

const char *
nestPwent::getent_passwd( void ) const
{
	if ( ! valid ) {
		return NULL;
	} else {
		return pwent.pw_passwd;
	}
}

uid_t
nestPwent::getent_uid( void ) const
{
	if ( ! valid ) {
		return (uid_t) -1;
	} else {
		return pwent.pw_uid;
	}
}

gid_t
nestPwent::getent_gid( void ) const
{
	if ( ! valid ) {
		return (gid_t) -1;
	} else {
		return pwent.pw_gid;
	}
}

const char *
nestPwent::getent_gecos( void ) const
{
	if ( ! valid ) {
		return NULL;
	} else {
		return pwent.pw_gecos;
	}
}

const char *
nestPwent::getent_dir( void ) const
{
	if ( ! valid ) {
		return NULL;
	} else {
		return pwent.pw_dir;
	}
}

const char *
nestPwent::getent_shell( void ) const
{
	if ( ! valid ) {
		return NULL;
	} else {
		return pwent.pw_shell;
	}
}

int
nestPwent::copyent( struct passwd *ent )
{
	this->pwent.pw_name = strdup( ent->pw_name );
	if ( ! this->pwent.pw_name ) {
		return -1;
	}

	this->pwent.pw_passwd = strdup( ent->pw_passwd );
	if ( ! this->pwent.pw_passwd ) {
		return -1;
	}

	this->pwent.pw_uid = ent->pw_uid;

	this->pwent.pw_gid = ent->pw_gid;

	this->pwent.pw_gecos = strdup( ent->pw_gecos );
	if ( ! this->pwent.pw_gecos ) {
		return -1;
	}

	this->pwent.pw_dir = strdup( ent->pw_dir );
	if ( ! this->pwent.pw_dir ) {
		return -1;
	}

	this->pwent.pw_shell = strdup( ent->pw_shell );
	if ( ! this->pwent.pw_shell ) {
		return -1;
	}

	valid = 1;
	return 0;
}
