/***************************Copyright-DO-NOT-REMOVE-THIS-LINE**
  *
  * Condor Software Copyright Notice
  * Copyright (C) 1990-2005, Condor Team, Computer Sciences Department,
  * University of Wisconsin-Madison, WI.
  *
  * This source code is covered by the Condor Public License, which can
  * be found in the accompanying LICENSE.TXT file, or online at
  * www.condorproject.org.
  *
  * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
  * AND THE UNIVERSITY OF WISCONSIN-MADISON "AS IS" AND ANY EXPRESS OR
  * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
  * WARRANTIES OF MERCHANTABILITY, OF SATISFACTORY QUALITY, AND FITNESS
  * FOR A PARTICULAR PURPOSE OR USE ARE DISCLAIMED. THE COPYRIGHT
  * HOLDERS AND CONTRIBUTORS AND THE UNIVERSITY OF WISCONSIN-MADISON
  * MAKE NO MAKE NO REPRESENTATION THAT THE SOFTWARE, MODIFICATIONS,
  * ENHANCEMENTS OR DERIVATIVE WORKS THEREOF, WILL NOT INFRINGE ANY
  * PATENT, COPYRIGHT, TRADEMARK, TRADE SECRET OR OTHER PROPRIETARY
  * RIGHT.
  *
  ****************************Copyright-DO-NOT-REMOVE-THIS-LINE**/

#include "nest_quota_manager.h"

QuotaManager::QuotaManager( void )
{

    quota_inst_map = NULL;

    for(int i=0; i < NUM_QUOTA_INSTANCES; i++) {
        quota_inst[i] = NULL;
    }

    this->quota_catalog = NULL;

    return;
}

QuotaManager::~QuotaManager( void )
{
    return;
}


int
QuotaManager::locate_quota_inst(const char *dev)
{

    for(QuotaInstanceMap *cur = quota_inst_map; cur; cur = cur->next) {
        if (!strcmp(cur->dev, dev)) {
            return cur->qi_index;
        }
    }

    return -1;
}


NestReplyStatus
QuotaManager::add_entity(const nestUserOrGroup *name,
                         const char *dev,
                         nestLotSize size,
                         int inode_count)
{

    if (!dev) {
        return NEST_BAD_PARAMETERS;
    }

    nestLotSize  tmp_size;
    int tmp_inode_count;

    if (view_quota_status(name, dev, &tmp_size, &tmp_inode_count)
            == NEST_SUCCESS) {
        return NEST_USER_EXISTS;
    }

    int qmap = this->locate_quota_inst(dev);

    if (qmap < 0) {
        return NEST_NO_QUOTA_ENFORCEMENT;
    }

    if (this->quota_inst[qmap]->add_entity(name,
                                           dev,
                                           size,
                                           inode_count)
                != NEST_SUCCESS) {
        return NEST_QUOTA_ENFORCEMENT_ERROR;
    }

    string  str_key,
            str_quota_status_classad = "";


    char    str_is_user[NEST_NAMELEN],
            str_size[NEST_NAMELEN],
            str_inode[NEST_NAMELEN];

    NestSnprintf(str_is_user, NEST_NAMELEN, "%s",
				 name->isUser() ? "true" : "false");
    NestSnprintf(str_size, NEST_NAMELEN, "%lld", size);
    NestSnprintf(str_inode, NEST_NAMELEN, "%d", inode_count);

    str_key = QUOTA_KEY_PREFIX;
    str_key = str_key + name->getName() + ":" +
                        str_is_user + ":" +
                        dev;

    str_quota_status_classad =  str_quota_status_classad +
		"[  Type = \"" + QUOTA_KEY_PREFIX + "\"; " +
		"   Name = \"" + name->getName() + "\"; " +
		"   IsUser = " + str_is_user + "; " +
		"   Device = \"" + dev + "\"; " +
		"   DataSize = " + str_size + "; " +
		"   InodeCount = " + str_inode +
		"]";

    ClassAd *quota_class_ad = parser.ParseClassAd( str_quota_status_classad,
                                                   true );

    if (!quota_class_ad) {
        return NEST_UNKNOWN_ERROR;
    }


    if (!this->quota_catalog->AddClassAd( str_key, quota_class_ad )) {
        return NEST_UNKNOWN_ERROR;
    }

    return NEST_SUCCESS;
}


NestReplyStatus
QuotaManager::delete_entity(const nestUserOrGroup *name,
                            const char *dev)
{

    if (!dev) {
        return NEST_BAD_PARAMETERS;
    }

    int qmap = this->locate_quota_inst(dev);

    if (qmap < 0) {
        return NEST_NO_QUOTA_ENFORCEMENT;
    }

    if (this->quota_inst[qmap]->delete_entity(name,
                                              dev)
            != NEST_SUCCESS) {
        return NEST_QUOTA_ENFORCEMENT_ERROR;
    }

    string  str_key;

    char    str_is_user[NEST_NAMELEN];

    NestSnprintf(str_is_user, NEST_NAMELEN, "%s",
				 name->isUser() ? "true" : "false");

    str_key = QUOTA_KEY_PREFIX;
    str_key = str_key + name->getName() + ":" +
                str_is_user + ":" +
                dev;

    if (!this->quota_catalog->RemoveClassAd( str_key )) {
        nest_debug( NEST_DEBUG_LOTS,
                    "Couldn't remove Quota ClassAd from the collection!\n" );
        return NEST_CLASSAD_ERROR;
    }

    return NEST_SUCCESS;
}

NestReplyStatus
QuotaManager::update_entity(const nestUserOrGroup *name,
                            const char *dev,
                            nestLotSize new_size,
                            int new_inode_count)
{

    if (!dev) {
        return NEST_BAD_PARAMETERS;
    }

    int qmap = this->locate_quota_inst(dev);

    if (qmap < 0) {
        return NEST_NO_QUOTA_ENFORCEMENT;
    }

    if (this->quota_inst[qmap]->update_entity(name,
                                              dev,
                                              new_size,
                                              new_inode_count)
            != NEST_SUCCESS) {
        return NEST_QUOTA_ENFORCEMENT_ERROR;
    }

    char    str_is_user[NEST_NAMELEN],
            str_size[NEST_NAMELEN],
            str_inode[NEST_NAMELEN];

    NestSnprintf(str_is_user, NEST_NAMELEN, "%s",
				 name->isUser() ? "true" : "false");
    NestSnprintf(str_size, NEST_NAMELEN, "%lld", new_size);
    NestSnprintf(str_inode, NEST_NAMELEN, "%d", new_inode_count);

    string str_key = QUOTA_KEY_PREFIX;
    str_key = str_key + name->getName() + ":" +
                        str_is_user + ":" +
                        dev;

    string str_modify_classad = "";
    str_modify_classad = str_modify_classad +
		"[  DataSize = " + str_size + "; " +
		"   InodeCount = " + str_inode +
		"]";

    str_modify_classad = " [ Updates = " + str_modify_classad + " ] ";

    ClassAd *modify_classad = parser.ParseClassAd( str_modify_classad, true );

    if (!this->quota_catalog->ModifyClassAd( str_key, modify_classad )) {
        nest_debug( NEST_DEBUG_LOTS,
                    "Modify ClassAd failed in update entity!\n" );
        return NEST_CLASSAD_ERROR;
    }

    return NEST_SUCCESS;
}


NestReplyStatus
QuotaManager::view_quota_status(const nestUserOrGroup *name,
                                const char *dev,
                                nestLotSize *size,
                                int *inode_count)
{

    if (!dev || !size || !inode_count) {
        return NEST_BAD_PARAMETERS;
    }

    char    str_is_user[NEST_NAMELEN];

    NestSnprintf(str_is_user, NEST_NAMELEN, "%s",
				 name->isUser() ? "true" : "false");

    string str_key = QUOTA_KEY_PREFIX;
    str_key = str_key + name->getName() + ":" +
                        str_is_user + ":" +
                        dev;

    ClassAd *quota_class_ad = this->quota_catalog->GetClassAd(str_key);

    if (!quota_class_ad) {
        return NEST_USER_NOT_FOUND;
    }

    bool result;
    nestLotSize tmp_size;
	char tmp[64];
    result = quota_class_ad->EvaluateAttrString( "DataSize", tmp, sizeof(tmp));
	tmp_size = atoll( tmp );
    int tmp_inode_count;
    result &= quota_class_ad->EvaluateAttrInt( "InodeCount", tmp_inode_count );

    if (!result) {
        nest_debug( NEST_DEBUG_LOTS,
                    "Couldn't find req attrs in view_entity!\n" );
        return NEST_CLASSAD_ERROR;
    }

    *size = tmp_size;
    *inode_count = tmp_inode_count;

    return NEST_SUCCESS;
}


NestReplyStatus
QuotaManager::view_quota_usage(const nestUserOrGroup *name,
                               const NestDevice *disk,
                               nestLotSize *size,
                               int *inode_count)
{

    if (!disk || !size || !inode_count) {
        return NEST_BAD_PARAMETERS;
    }

    int qmap = this->locate_quota_inst( disk->getDevName() );

    if (qmap < 0) {
        return NEST_NO_QUOTA_ENFORCEMENT;
    }

    if (this->quota_inst[qmap]->view_entity_usage(name,
												  disk,
												  size,
												  inode_count)
            != NEST_SUCCESS) {
        return NEST_QUOTA_ENFORCEMENT_ERROR;
    }

    return NEST_SUCCESS;
}


NestReplyStatus
QuotaManager::incr_quota(const nestUserOrGroup *name,
                         const char *dev,
                         nestLotSize more_size)
{

    if (!name || !dev) {
        return NEST_BAD_PARAMETERS;
    }

    nestLotSize cur_size;
    int cur_inode_count;
    NestReplyStatus reply;

    reply = this->view_quota_status( name, dev, &cur_size, &cur_inode_count );
    if (reply != NEST_SUCCESS) {
        if (this->add_entity(name,
                             dev,
                             0,
                             0) != NEST_SUCCESS) {
            nest_debug( NEST_DEBUG_LOTS,
                        "Couldn't add entity to the collection!\n" );
            return NEST_CLASSAD_ERROR;
        }
        cur_size = 0;
        cur_inode_count = 0;
    }

    cur_size += more_size;
	nestLotSize kbytes = ( more_size + 1023 ) / 1024;
    cur_inode_count += (int) ( kbytes / DATA_KB_PER_INODE );
    return this->update_entity( name, dev, cur_size, cur_inode_count );
}


NestReplyStatus
QuotaManager::decr_quota(const nestUserOrGroup *name,
                         const char *dev,
                         nestLotSize less_size)
{
    if (!name || !dev) {
        return NEST_BAD_PARAMETERS;
    }

    nestLotSize cur_size;
    int cur_inode_count;

    NestReplyStatus reply = this->view_quota_status(name,
                                                    dev,
                                                    &cur_size,
                                                    &cur_inode_count);
    NEST_REPLY_CHECK(reply);

    cur_size -= less_size;
    cur_inode_count -= (int) ((less_size*1024 + 1023)/DATA_KB_PER_INODE);

    if (cur_size < 0) {
        cur_size = 0;
    }
    if (cur_inode_count < 0) {
        cur_inode_count = 0;
    }

    return(this->update_entity(name,
                               dev,
                               cur_size,
                               cur_inode_count));
}

NestReplyStatus
QuotaManager::initialize(ClassAdCollection *resources,
                         const NestDevice *disk,
                         nestLotSize *user_usage)
{

    if (!resources || !disk || !user_usage) {
        return NEST_BAD_PARAMETERS;
    }

    char        fstype[NEST_MAXPATHLEN];
	const char  *dev;

    if(!NestFSType(dev,fstype)) {
        nest_debug( NEST_DEBUG_ALWAYS,
                   "Device %s not recognized by getfsspec()"
					" in QuotaManager!\n",
                    dev );
        return NEST_UNKNOWN_ERROR;
    }

    if ( !strcmp(fstype, "ext2") || 
		 !strcmp(fstype, "ext3") ) {
        QuotaInstanceMap    *tmp = (QuotaInstanceMap *)
			NestMalloc (sizeof(QuotaInstanceMap));
        strcpy(tmp->dev, dev );
        tmp->qi_index = FS_EXT2;
        tmp->next = quota_inst_map;
        quota_inst_map = tmp;

        if (!quota_inst[FS_EXT2]) {
#ifdef HAS_LINUX_EXT2_FS_H
            quota_inst[FS_EXT2] = new FsExt2QuotaInstance();
#else
            quota_inst[FS_EXT2] = 0;
#endif

            if (!quota_inst[FS_EXT2]) {
                return NEST_OUT_OF_RESOURCES;
            }
            if (quota_inst[FS_EXT2]->initialize( disk ) != NEST_SUCCESS) {
                return NEST_NO_QUOTA_ENFORCEMENT;
            }
        }
    }
    else {
        return NEST_NO_QUOTA_ENFORCEMENT;
    }

    this->quota_catalog = resources;

    this->query.Bind(resources);
    this->quota_catalog_view = NEST_QUOTA_CATALOG_VIEW;

    ViewName    parent_view = "root";

    string view_constraint(""),
           view_rank(""),
           view_partition_expr("");

    view_constraint = view_constraint +
		" other.Type == \"" + QUOTA_KEY_PREFIX + "\" ";

    if (!this->quota_catalog->CreateSubView( this->quota_catalog_view,
											 parent_view,
                                             view_constraint,
                                             view_rank,
                                             view_partition_expr ) ) {
        nest_debug( NEST_DEBUG_ALWAYS,
                    "QuotaManager Couldn't create view for itself!\n");
        nest_debug(NEST_DEBUG_ALWAYS, "CondorErrMsg says: %s",
            CondorErrMsg.c_str());
        return NEST_CLASSAD_ERROR;
    }

    string constraint = "";
    constraint = constraint +
		"other.IsUser == true && " +
		"other.ValidEntry == true";

    ExprTree    *constraint_tree = this->parser.ParseExpression(constraint);
    if (!constraint_tree) {
        return NEST_CLASSAD_ERROR;
    }

    this->query.Query(NEST_IDENTITIES_VIEW, constraint_tree);
    this->query.ToFirst();

    //  delete constraint_tree;

    string  id_key;
    ClassAd *id_ad;
    nestLotSize  usage;
    int     inode_count;

    if (!this->query.Current(id_key)) {
        return NEST_SUCCESS;
    }

    do {
        id_ad = this->quota_catalog->GetClassAd( id_key );
        if (!id_ad) {
            continue;
        }

        string  actual_name;
        if (!id_ad->EvaluateAttrString( "ActualName", actual_name )) {
            continue;
        }
		nestUser	user( &actual_name );

		if ( NestAdminUser && user.isAdmin() ) {
            /*
                NestAdminUser ( AdminUid ) already owns storage_dir
                and meta_dir

                so exclude the admin's quota_status from this
            */
            continue;
        }

        usage = 0;
        inode_count = 0;
        this->view_quota_usage(&user, disk, &usage, &inode_count);

        *user_usage += usage;

    } while (this->query.Next(id_key));

    return NEST_SUCCESS;
}

NestReplyStatus
QuotaManager::recover( ClassAdCollection *resources,
					   const NestDevice *disk )
{

    char        fstype[NEST_MAXPATHLEN];
	const char  *dev = disk->getDevName( );

    if(!NestFSType(dev, fstype)) {
        nest_debug( NEST_DEBUG_ALWAYS,
					"Device %s not recognized by getfsspec()"
					" in QuotaManager!\n",
					dev );
        return NEST_UNKNOWN_ERROR;
    }

    if ( !strcmp(fstype, "ext2") ||
		 !strcmp(fstype, "ext3") ) {
        QuotaInstanceMap    *tmp =
			(QuotaInstanceMap *) NestMalloc( sizeof(QuotaInstanceMap) );

         if (!tmp) {
            return NEST_OUT_OF_RESOURCES;
         }

        strcpy(tmp->dev, dev);
        tmp->qi_index = FS_EXT2;
        tmp->next = quota_inst_map;
        quota_inst_map = tmp;

        if (!quota_inst[FS_EXT2]) {
#ifdef HAS_LINUX_EXT2_FS_H
            quota_inst[FS_EXT2] = new FsExt2QuotaInstance();
#else
            quota_inst[FS_EXT2] = 0;
#endif

            if (!quota_inst[FS_EXT2]) {
                return NEST_OUT_OF_RESOURCES;
            }
            if (quota_inst[FS_EXT2]->initialize(disk) != NEST_SUCCESS) {
                return NEST_NO_QUOTA_ENFORCEMENT;
            }
        }
    }
    else {
        return NEST_NO_QUOTA_ENFORCEMENT;
    }

    this->quota_catalog = resources;
    this->query.Bind(resources);

    this->quota_catalog_view = NEST_QUOTA_CATALOG_VIEW;

    this->synchronize();

    return NEST_SUCCESS;
}

NestReplyStatus
QuotaManager::shutdown()
{
    return NEST_SUCCESS;
}


NestReplyStatus
QuotaManager::synchronize()
{
    bool    is_user;
    nestLotSize  size;
    int     inode_count;
    string  name,
            str_dev;
    bool    result;

    this->query.Query(this->quota_catalog_view, NULL);
    this->query.ToFirst();

    string query_key;

    if (!this->query.Current(query_key)) {
        return NEST_SUCCESS;  //  No device found!
    }

    do {
        ClassAd *quota_class_ad = this->quota_catalog->GetClassAd(query_key);

        if (!quota_class_ad) {
            continue;
        }

		char tmp[64];
        result  = quota_class_ad->EvaluateAttrString( "Name", name );
        result &= quota_class_ad->EvaluateAttrBool( "IsUser", is_user );
        result &= quota_class_ad->EvaluateAttrString( "Device", str_dev );
        result &= quota_class_ad->EvaluateAttrString( "DataSize", tmp,
													  sizeof( tmp) );
		size = atoll( tmp );
        result &= quota_class_ad->EvaluateAttrInt( "InodeCount", inode_count );

        if (!result) {
            nest_debug( NEST_DEBUG_LOTS,
                        "Not all the relevant attributes found" );
            nest_debug( NEST_DEBUG_LOTS,
                        " in QuotaStatus Classad!\n" );
            continue;
        }

        int qmap = this->locate_quota_inst(str_dev.c_str());

        if (qmap < 0) {
            nest_debug( NEST_DEBUG_LOTS,
						"Quota enforcement not found for dev %s "
						"while syncing!\n",
                        str_dev.c_str() );
            continue;
        }

		const nestUserOrGroup	*ug;
		if ( is_user ) {
			ug = new nestUser( &name );
		} else {
			ug = new nestGroup( &name );
		}
		NestReplyStatus status = 
			this->quota_inst[qmap]->update_entity( ug,
												   str_dev.c_str(),
												   size,
												   inode_count );
		delete ug;
		if ( status != NEST_SUCCESS) {
            nest_debug( NEST_DEBUG_LOTS,
                        "Error while syncing quota status!\n" );
            continue;
        }

    } while (this->query.Next(query_key));

    return NEST_SUCCESS;
}

NestReplyStatus
QuotaManager::has_kernel_support( const NestDevice *dev )
{

    if (!dev) {
        return NEST_BAD_PARAMETERS;
    }

    char        fstype[NEST_MAXPATHLEN];

    if(!NestFSType(dev->getDevName() ,fstype)) {
        nest_debug( NEST_DEBUG_ALWAYS,
					"Device %s not recognized by getfsspec() "
					"in QuotaManager!\n",
                    dev->getDevName()  );
        return NEST_UNKNOWN_ERROR;
    }

    if ( !strcmp(fstype, "ext2") || 
		 !strcmp(fstype, "ext3") ) {

#ifdef HAS_LINUX_EXT2_FS_H
        QuotaInstance *qi = new FsExt2QuotaInstance();
#else
        QuotaInstance *qi = 0;
#endif

        if (!qi) {
            return NEST_OUT_OF_RESOURCES;
        }
        if (qi->initialize(dev) != NEST_SUCCESS) {
            return NEST_NO_QUOTA_ENFORCEMENT;
        }
    }
    else {
        return NEST_NO_QUOTA_ENFORCEMENT;
    }

    return NEST_SUCCESS;
}
