/***************************Copyright-DO-NOT-REMOVE-THIS-LINE**
  *
  * Condor Software Copyright Notice
  * Copyright (C) 1990-2005, Condor Team, Computer Sciences Department,
  * University of Wisconsin-Madison, WI.
  *
  * This source code is covered by the Condor Public License, which can
  * be found in the accompanying LICENSE.TXT file, or online at
  * www.condorproject.org.
  *
  * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
  * AND THE UNIVERSITY OF WISCONSIN-MADISON "AS IS" AND ANY EXPRESS OR
  * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
  * WARRANTIES OF MERCHANTABILITY, OF SATISFACTORY QUALITY, AND FITNESS
  * FOR A PARTICULAR PURPOSE OR USE ARE DISCLAIMED. THE COPYRIGHT
  * HOLDERS AND CONTRIBUTORS AND THE UNIVERSITY OF WISCONSIN-MADISON
  * MAKE NO MAKE NO REPRESENTATION THAT THE SOFTWARE, MODIFICATIONS,
  * ENHANCEMENTS OR DERIVATIVE WORKS THEREOF, WILL NOT INFRINGE ANY
  * PATENT, COPYRIGHT, TRADEMARK, TRADE SECRET OR OTHER PROPRIETARY
  * RIGHT.
  *
  ****************************Copyright-DO-NOT-REMOVE-THIS-LINE**/

#include "nest_request.h"
#include <string.h>

// Helper functions
void set_handle( const char *handle_str, NestLotHandle *handle )
{
	if ( handle_str && isdigit( *handle_str ) ) {
		rsv_id_t	tmp_id = (rsv_id_t) -1;

		tmp_id = strtoul( handle_str, NULL, 10 );
		NestAdd_id_to_handle(handle, tmp_id);
	}
}

// Copy constructor
nestRequest::nestRequest( const nestRequest *req )
{
	// TODO
	init( );
}

// different requests pass different semantic data for arg1 and arg2
// quota requests pass the user name for arg1
// third party transfers pass the remotehost as arg1 and remote file as arg2
nestRequest::nestRequest( int type, NestReplyStatus status )
{
	init( );
	this->req_type = type;
	this->status = status;
}

void
nestRequest::init( void )
{
	this->validated = false;

	this->user = NULL;
	this->group = NULL;
	this->file = NULL;
	this->dest_file = NULL;
	this->dir = NULL;
	this->start = -1;
	this->end = -1;
	this->is_sym_link = false;

	this->auth_type = 0;
	this->auth_type_set = false;

	this->remote_file = NULL;
	this->remote_host = NULL;
	this->remote_protocol = -1;

	this->file_attr = NULL;
	this->chunk_size = 0;

	init( lot_handle );
	this->lot_size = -1;
	this->lot_size_set = false;
	this->lot_type = -1;
	this->lot_duration = -1;
	this->lot_duration_set = false;
	init( lot_phandle );
	this->lot_pug = NULL;
	init( lot_src_handle );
	init( lot_dst_handle );
	this->lot_bytes_to_move = -1;

	this->coll_name = NULL;
	this->coll_comment = NULL;
	this->coll_version = NULL;
	this->coll_mtime = 0;

	this->rights = 0;
	this->rights_set = false;

	this->proxy_location = NULL;

	this->debug_string = NULL;

    NestGettimeofday( &this->createtime );
}

nestRequest::~nestRequest( void )
{
	if ( this->user ) 			delete this->user;
	if ( this->group )			delete this->group;
	if ( this->file )			free( (void*) this->file );
	if ( this->dest_file )		free( (void*) this->dest_file );
	if ( this->dir )			free( (void*) this->dir );

	if ( this->remote_file )	free( (void*) this->remote_file );
	if ( this->remote_host )	free( (void*) this->remote_host );

	if ( this->lot_pug )		delete this->lot_pug;

	if ( this->coll_name )		free( (void*) this->coll_name );
	if ( this->coll_comment )	free( (void*) this->coll_comment );
	if ( this->coll_version )	free( (void*) this->coll_version );
	if ( this->proxy_location )	free( (void*) this->proxy_location );

}

bool
nestRequest::validate( void )
{
	int			failed_line = 0;
	const char	*failed_str = NULL;
#define HANDLE_INVALID_REQ(_s_)\
	{ failed_line = __LINE__; failed_str = _s_; goto DONE; }

	// Depends on the request type
	switch( req_type )
	{
	case DIR_LIST_REQUEST:
	case AVAILABLE_SPACE_QUERY:
	case TIME_REQUEST:
	case SERVER_TERMINATE_REQUEST:
	case FORMAT_DISK_REQUEST:
	case PWD_REQUEST:
	case NEST_SYNC:
	case REQUEST_NEST_AD:
	case VERSION_REQUEST:
	case PROXY_QUERY:
	case PROXY_INDEX_PUSH:
	case PROXY_FLUSH:
		break;

    case CLIENT_SEND_FILE:
		if ( ! file ) HANDLE_INVALID_REQ( "file" );
		if ( start < 0 ) HANDLE_INVALID_REQ( "start" );
		if ( end < 0 ) HANDLE_INVALID_REQ( "end" );
		break;

    case CLIENT_RECV_FILE:
		if ( ! file ) HANDLE_INVALID_REQ( "file" );
		if ( start < 0 ) HANDLE_INVALID_REQ( "start" );
		break;

    case TEST_CLIENT_TO_SERVER:
    case TEST_SERVER_TO_CLIENT:
		if ( end < 0 ) HANDLE_INVALID_REQ( "end" );
		break;

    case FILE_DELETE_REQUEST:
    case FILESIZE_REQUEST:
    case PHYSNAME_REQUEST:
    case FILEINFO_REQUEST:
		if ( ! file ) HANDLE_INVALID_REQ( "file" );
		break;

    case RMDIR_REQUEST:
    case CWD_REQUEST:
    case MKDIR_REQUEST:
		if ( ! dir ) HANDLE_INVALID_REQ( "dir" );
		break;

    case FILE_RENAME_REQUEST:
	case COPY_REQUEST:
	case LINK_REQUEST:
		if ( ! file ) HANDLE_INVALID_REQ( "file" );
		if ( ! dest_file ) HANDLE_INVALID_REQ( "dest_file" );
		break;

	case ADD_NEST_USER:
	case ADD_USER_TO_GROUP:
	case DELETE_USER_FROM_GROUP:
		if ( ! user ) HANDLE_INVALID_REQ( "user" );
		if ( ! group ) HANDLE_INVALID_REQ( "group" );
		break;

	case ADD_NEST_GROUP:
	case DELETE_NEST_GROUP:
		if ( ! group ) HANDLE_INVALID_REQ( "group" );
		break;

	case DELETE_NEST_USER:
		if ( ! user ) HANDLE_INVALID_REQ( "user" );
		break;

	case THIRD_PARTY_GET:
	case THIRD_PARTY_PUT:
		if ( ! file ) HANDLE_INVALID_REQ( "file" );
		if ( ! remote_host ) HANDLE_INVALID_REQ( "remote_host" );
		if ( ! remote_file ) HANDLE_INVALID_REQ( "remote_file" );
		if ( start < 0 ) HANDLE_INVALID_REQ( "start" );
		if ( end < 0 ) HANDLE_INVALID_REQ( "end" );
		if ( remote_protocol < 0 ) HANDLE_INVALID_REQ( "remote_protocol" );
		break;

	case CHANGE_USER_REQUEST:
		if ( ! user ) HANDLE_INVALID_REQ( "user" );
		if ( ! auth_type_set ) HANDLE_INVALID_REQ( "auth_type" );
		break;

	case SETATTR_REQUEST:
		if ( ! file_attr ) HANDLE_INVALID_REQ( "file_attr" );
		break;

	case LIST_GROUP_LOTS:
	case LIST_USER_LOTS:
		break;

	case GET_LOT_STAT:
	case TERMINATE_LOT:
		if ( ! isSet( lot_handle) ) HANDLE_INVALID_REQ( "lot_handle" );
		break;

	case UPDATE_LOT:
		if ( ! isSet( lot_handle) ) HANDLE_INVALID_REQ( "lot_handle" );
		if ( ! lot_size_set ) HANDLE_INVALID_REQ( "lot_size" );
		if ( ! lot_duration_set ) HANDLE_INVALID_REQ( "lot_duration" );
		break;

	case CHANGE_LOT_T:
		if ( ! isSet( lot_handle ) ) HANDLE_INVALID_REQ( "lot_handle" );
		if ( lot_type < 0 ) HANDLE_INVALID_REQ( "lot_type" );
		if ( ! lot_duration_set ) HANDLE_INVALID_REQ( "lot_duration" );
		break;

	case ADD_USER_TO_LOT:
	case DELETE_USER_FROM_LOT:
		if ( ! isSet( lot_handle ) ) HANDLE_INVALID_REQ( "lot_handle" );
		if ( user < 0 ) HANDLE_INVALID_REQ( "user" );
		break;

    case ATTACH_LOT_TO_FILE:
    case DETACH_LOT_FROM_FILE:
		if ( ! isSet( lot_handle ) ) HANDLE_INVALID_REQ( "lot_handle" );
		if ( ! file ) HANDLE_INVALID_REQ( "file" );
		break;

	case REQUEST_USER_LOT:
	case REQUEST_GROUP_LOT:
		// User / group not required (default to "client")
		if ( lot_size <= 0 ) HANDLE_INVALID_REQ( "lot_size" );
		if ( lot_type < 0 ) HANDLE_INVALID_REQ( "lot_type" );
		if ( lot_duration < 0 ) HANDLE_INVALID_REQ( "lot_duration" );
		break;

	case LOT_MOVE_FILE:
		if ( ! file ) HANDLE_INVALID_REQ( "file" );
		if ( ! isSet( lot_src_handle ) ) HANDLE_INVALID_REQ("lot_src_handle");
		if ( ! isSet( lot_dst_handle ) ) HANDLE_INVALID_REQ("lot_dst_handle");
		if ( lot_bytes_to_move < 0 ) HANDLE_INVALID_REQ("lot_bytes_to_move");
		break;

	// All must have a name...

    case NAME_COLLECTION:
	case SET_COLLECTION_MTIME:
	case SET_COLLECTION_VERSION:
	case SET_COLLECTION_COMMENT:
	case GET_COLLECTION_MTIME:
	case GET_COLLECTION_VERSION:
	case GET_COLLECTION_COMMENT:
	case UNNAME_COLLECTION:
    case QUERY_COLLECTION:
		if ( ! this->coll_name ) HANDLE_INVALID_REQ( "collection_name" );

		switch( req_type ) {
		case NAME_COLLECTION:
			if ( ! this->file ) HANDLE_INVALID_REQ( "file" );
			break;

		case SET_COLLECTION_MTIME:
			if ( ! this->coll_mtime ) HANDLE_INVALID_REQ("collection_mtime" );
			break;

		case SET_COLLECTION_VERSION:
			if ( ! this->coll_version )
				HANDLE_INVALID_REQ("collection_version" );
			break;

		case SET_COLLECTION_COMMENT:
			if ( ! this->coll_comment )
				HANDLE_INVALID_REQ("collection_comment" );
			break;

		case GET_COLLECTION_MTIME:
		case GET_COLLECTION_VERSION:
		case GET_COLLECTION_COMMENT:
		case UNNAME_COLLECTION:
			break;

		case QUERY_COLLECTION:
			break;
		}
		break;

	case GET_ACL:
		if ( ! user ) HANDLE_INVALID_REQ( "user" );
		break;
		
	case GET_GROUP_ACL:
		if ( ! group ) HANDLE_INVALID_REQ( "group" );
		break;

	case SET_ACL:
		if ( ! user ) HANDLE_INVALID_REQ( "user" );
		if ( ! rights_set ) HANDLE_INVALID_REQ( "rights" );
		break;

	case SET_GROUP_ACL:
		if ( ! group ) HANDLE_INVALID_REQ( "group" );
		if ( ! rights_set ) HANDLE_INVALID_REQ( "rights" );
		break;

    case PROXY_REGISTER:
		if ( ! proxy_location ) HANDLE_INVALID_REQ( "proxy_location" );
		break;

    case DEBUG_REQUEST:
		if ( ! debug_string ) HANDLE_INVALID_REQ( "debug" );
		break;

	default:
		nest_debug( NEST_DEBUG_ALWAYS,
					"I don't know how to vaidate request #%d '%s'\n",
					req_type, toString() );
		assert( 0 );
	}

	// Success??
  DONE:
	if ( failed_line ) {
		nest_debug( NEST_DEBUG_CHIRP,
					"Validation of request type %d failed '%s' @ line %d\n",
					req_type, failed_str, failed_line );
		return false;
	}

	// If we made it here, must be ok
	validated = true;
	return true;
}

// Set the status
void
nestRequest::setStatus( NestReplyStatus status )
{
	this->status = status;
}

// Basic parms -- set user
void
nestRequest::setUser( const nestUser *user, bool copy )
{
	if ( copy ) {
		this->user = new nestUser( user );
	} else {
		this->user = user;
	}
}
void
nestRequest::setUser( const char *user_str )
{
	if ( user_str ) {
		this->user = new nestUser( user_str );
	}
}

// Basic parms -- set group
void
nestRequest::setGroup( const nestGroup *group, bool copy )
{
	if ( copy ) {
		this->group = new nestGroup( group );
	} else {
		this->group = group;
	}
}
void
nestRequest::setGroup( const char *group_str )
{
	if ( group_str ) {
		this->group = new nestGroup( group_str );
	}
}

// Basic parms -- set auth type
void
nestRequest::setAuthType( int type )
{
	this->auth_type = type;
	this->auth_type_set = true;
}
void
nestRequest::setAuthType( const char *auth_str )
{
	if ( auth_str ) {
		setAuthType( atoi( auth_str ) );
	}
}

// Basic parms -- set start
void
nestRequest::setStart( nestFileSize start )
{
	this->start = start;
}
void
nestRequest::setStart( const char *start_str )
{
	if ( start_str ) {
		setStart( atol( start_str ) );
	}
}

// Basic parms -- set end
void
nestRequest::setEnd( nestFileSize end )
{
	this->end = end;
}
void
nestRequest::setEnd( const char *end_str )
{
	if ( end_str ) {
		setEnd( atol( end_str ) );
	}
}

// Basic parms -- set file
void
nestRequest::setFile( const char *file )
{
	if ( file ) {
		this->file = strdup( file );
	}
}

// Basic parms -- set directory
void
nestRequest::setDir( const char *dir )
{
	if ( dir ) {
		this->dir = strdup( dir );
	}
}

// Basic parms -- set "is sym link"
void
nestRequest::setIsSymLink( bool symlink )
{
	this->is_sym_link = symlink;
}
void
nestRequest::setIsSymLink( const char *symlink_str )
{
	if ( symlink_str ) {
		setIsSymLink( atoi( symlink_str ) );
	}
}

// Basic parms -- set dest file (for rename)
void
nestRequest::setDestFile( const char *dest_file )
{
	if ( dest_file ) {
		this->dest_file = strdup( dest_file );
	}
}

// Basic parms -- set remote file (for third party xfer)
void
nestRequest::setRemoteFile( const char *remote_file )
{
	if ( dest_file ) {
		this->dest_file = strdup( dest_file );
	}
}

// Basic parms -- set dest host (for third party xfer)
void
nestRequest::setRemoteHost( const char *remote_host )
{
	if ( remote_host ) {
		this->remote_host = strdup( remote_host );
	}
}

// Basic parms -- set remote protocol
void
nestRequest::setRemoteProtocol( int protocol )
{
	this->remote_protocol = protocol;
}

// Basic parms -- set file attribute
void
nestRequest::setFileAttr( const void *attr )
{
	this->file_attr = attr;
}

// Basic parms -- set protocol
void
nestRequest::setChunkSize( nestFileSize chunk_size  )
{
	this->chunk_size = chunk_size;
}

// Basic Lot parms -- set lot handle
void
nestRequest::setLotHandle( const NestLotHandle *handle )
{
	if ( handle ) {
		memcpy( &this->lot_handle, handle, sizeof( NestLotHandle ) );
	}
}
void
nestRequest::setLotHandle( const char *handle_str )
{
	set_handle( handle_str, &this->lot_handle );
}

// Basic Lot parms -- set size
void
nestRequest::setLotSize( nestLotSize size )
{
	this->lot_size = size;
	this->lot_size_set = true;
}
void
nestRequest::setLotSize( const char *size_str )
{
	if ( size_str ) {
		setLotSize( atoll( size_str ) );
	}
}

// Basic Lot parms -- set lot type
void
nestRequest::setLotType( int lot_type )
{
	this->lot_type = lot_type;
}
void
nestRequest::setLotType( const char *lot_str )
{
	if ( lot_str ) {
		this->lot_type = atoi( lot_str );
	}
}

// Basic Lot parms -- set duration
void
nestRequest::setLotDuration( int duration )
{
	this->lot_duration = duration;
	this->lot_duration_set = true;
}
void
nestRequest::setLotDuration( const char *duration_str )
{
	if ( duration_str ) {
		setLotDuration( atoi( duration_str ) );
	}
}

// Lot create parms -- set parent lot handle
void
nestRequest::setLotParentLot( const NestLotHandle *handle )
{
	if ( handle ) {
		memcpy( &this->lot_phandle, handle, sizeof( NestLotHandle ) );
	}
}
void
nestRequest::setLotParentLot( const char *handle_str )
{
	set_handle( handle_str, &this->lot_phandle );
}

// Lot create parms -- set parent user
void
nestRequest::setLotParentUser( const nestUser *user, bool copy )
{
	if ( lot_pug ) { return; };		// Invalid
	if ( copy ) {
		this->lot_pug = new nestUser( user );
	} else {
		this->lot_pug = user;
	}
}

void
nestRequest::setLotParentUser( const char *user_str )
{
	if( lot_pug ) { return; };
	if ( user_str ) {
		this->lot_pug = new nestUser( user_str );
	}
}

// Lot create parms -- set group user
void
nestRequest::setLotParentGroup( const nestGroup *group, bool copy )
{
	if ( lot_pug ) { return; };		// Invalid
	if ( copy ) {
		this->lot_pug = new nestGroup( group );
	} else {
		this->lot_pug = group;
	}
}

void
nestRequest::setLotParentGroup( const char *group_str )
{
	if( lot_pug ) { return; };
	if ( group_str ) {
		this->lot_pug = new nestGroup( group_str );
	}
}

// Lot move parms -- set source lot handle
void
nestRequest::setLotSourceHandle( const NestLotHandle *handle )
{
	if ( handle ) {
		memcpy( &this->lot_src_handle, handle, sizeof( NestLotHandle ) );
	}
}
void
nestRequest::setLotSourceHandle( const char *handle_str )
{
	set_handle( handle_str, &this->lot_src_handle );
}

// Lot move parms -- set dest lot handle
void
nestRequest::setLotDestHandle( const NestLotHandle *handle )
{
	if ( handle ) {
		memcpy( &this->lot_dst_handle, handle, sizeof( NestLotHandle ) );
	}
}
void
nestRequest::setLotDestHandle( const char *handle_str )
{
	set_handle( handle_str, &this->lot_dst_handle );
}

// Lot move parms -- set "bytes to move"
void
nestRequest::setLotBytesToMove( nestLotSize bytes )
{
	this->lot_bytes_to_move = bytes;
}
void
nestRequest::setLotBytesToMove( const char *bytes_str )
{
	if ( bytes_str ) {
		this->lot_bytes_to_move = atoll( bytes_str );
	}
}

// Basic Collection parms -- set collection name
void
nestRequest::setCollectionName( const char *name )
{
	if ( name ) {
		this->coll_name = strdup( name );
	}
}

// Basic Collection parms -- set comment
void
nestRequest::setCollectionComment( const char *comment )
{
	if( comment ) {
		this->coll_comment = strdup( comment );
	}
}

// Basic Collection parms -- set version
void
nestRequest::setCollectionVersion( const char *version )
{
	if ( version ) {
		this->coll_version = strdup( version );
	}
}

// Basic Collection parms -- set mtime
void
nestRequest::setCollectionMtime( time_t mtime )
{
	this->coll_mtime = mtime;
}
void
nestRequest::setCollectionMtime( const char *mtime_str )
{
	if ( mtime_str && isdigit( *mtime_str ) ) {
		this->coll_mtime = atol( mtime_str );
	}
}

// Basic Perm parms -- set perm type
void
nestRequest::setRights( nest_access_rights_t rights )
{
	this->rights = rights;
	rights_set = true;
}
void
nestRequest::setRights( const char *rights_str )
{
	if ( rights_str ) {
		this->rights = atoi( rights_str );
		rights_set = true;
	}
}

// Proxy parms -- set location
void
nestRequest::setProxyLocation( const char *location )
{
	if ( location ) {
		this->proxy_location = strdup( location );
	}
}

// Debug parms -- set debug string
void
nestRequest::setDebugString( const char *debug_string )
{
	if ( debug_string ) {
		this->debug_string = strdup( debug_string );
	}
}

bool
nestRequest::isSet( const NestLotHandle &lot ) const
{
	if ( ! lot.rsv_id ) {
		return false;
	} else {
		return true;
	}
}

void
nestRequest::init( NestLotHandle &lot )
{
	lot.rsv_id = 0;
}

long
nestRequest::getAge( void ) const
{
    return NestElapsedTime( &createtime );
}

const char *
nestRequest::toString( void ) const
{
    switch( req_type ) {
    case TIME_REQUEST:                       // 0
        return "time";
    case CLIENT_SEND_FILE:
        return "client send";
    case CLIENT_RECV_FILE:
        return "client recv";
    case FILE_DELETE_REQUEST:
        return "rm";
    case BANDWIDTH_STATISTICS_REQUEST:
        return "statistics";
    case DIR_LIST_REQUEST:                   // 5
        return "ls";
    case MKDIR_REQUEST:
        return "mkdir";
    case CWD_REQUEST:
        return "cwd";
    case PWD_REQUEST:
        return "pwd";
    case FILESIZE_REQUEST:
        return "filesize";
    case SERVER_TERMINATE_REQUEST:           // 10
        return "server terminate";
    case CLIENT_TERMINATE_REQUEST:
        return "client terminate";
    case TEST_MESSAGE:
        return "test message";
    case TEST_CLIENT_TO_SERVER:
        return "test client-server";
    case TEST_SERVER_TO_CLIENT:
        return "test server-client";
    case REQUEST_DONE:                       // 15
        return "request done";
    case CHANGE_USER_REQUEST:
        return "change user";
    case AVAILABLE_SPACE_QUERY:
        return "available";
    case FORMAT_DISK_REQUEST:
        return "format";
    case CONNECTION_REQUEST:
        return "open connection";
    case CLOSE_CONNECTION_REQUEST:           // 20
        return "close connection";
    case THIRD_PARTY_GET:
        return "3p get";
    case THIRD_PARTY_PUT:
        return "3p put";
    case DEBUG_REQUEST:
        return "debug";
    case RMDIR_REQUEST:
        return "rmdir";
    case NAME_COLLECTION:                    // 25
        return "name collection";
    case UNNAME_COLLECTION:
        return "unname collection";
    case QUERY_COLLECTION:
        return "query collection";
    case HELP_REQUEST:
        return "help";
    case RHELP_REQUEST:
        return "rhelp";
    case NEST_SYNC:                          // 30
        return "sync";
    case REQUEST_NEST_AD:
        return "ad";
    case LINK_REQUEST:
        return "link";
    case ADD_NEST_USER:
        return "add user";
    case ADD_NEST_GROUP:
        return "add group";
    case DELETE_NEST_USER:                   // 35
        return "delete user";
    case BAD_REQUEST:
        return "bad request";
    case DELETE_NEST_GROUP:
        return "delete group";
    case ADD_USER_TO_GROUP:
        return "add user to group";
    case DELETE_USER_FROM_GROUP:
        return "delete user from group";
    case REQUEST_USER_LOT:                   // 40
        return "request user lot";
    case REQUEST_GROUP_LOT:
        return "request group lot";
    case LIST_USER_LOTS:
        return "list user lots";
    case LIST_GROUP_LOTS:
        return "list group lots";
    case GET_LOT_STAT:
        return "get lot";
    case UPDATE_LOT:                         // 45
        return "update lot";
    case CHANGE_LOT_T:
        return "chage lot type";
    case TERMINATE_LOT:
        return "terminate lot";
    case VERSION_REQUEST:
        return "version";
    case PHYSNAME_REQUEST:
        return "physname";
    case GET_ACL:                            // 50
        return "get acl";
    case SET_ACL:                            
        return "set acl";
    case GET_GROUP_ACL:
        return "get group acl";
    case SET_GROUP_ACL:
        return "set group acl";
    case LOT_MOVE_FILE:
        return "move lot";
    case GET_FILEHANDLE:                     // 55   
        return "filehandle";
    case FILEINFO_REQUEST:                   
        return "fileinfo";
    case PROXY_GET:
        return "proxy get";
    case PROXY_PUT:
        return "proxy put";
    case PROXY_REGISTER:
        return "proxy register";
    case PROXY_QUERY:                        // 60
        return "proxy query";
    case COPY_REQUEST:                        
        return "copy";
    case ADD_USER_TO_LOT:
        return "add user to lot";
    case DELETE_USER_FROM_LOT:
        return "delete user from lot";
    case FILE_RENAME_REQUEST:
        return "mv";
    case ATTACH_LOT_TO_FILE:
        return "attach lot to file";
    case DETACH_LOT_FROM_FILE:
        return "detach lot from file";
	case PROXY_INDEX_PUSH:
		return "proxy index push";
	case PROXY_FLUSH:
		return "proxy flush";
    case SETATTR_REQUEST:
        return "set fileattr";
	case SET_COLLECTION_MTIME:
		return "set collection time";
	case SET_COLLECTION_VERSION:
		return "set collection version";
	case SET_COLLECTION_COMMENT:
		return "set collection comment";
	case GET_COLLECTION_MTIME:
		return "get collection time";
	case GET_COLLECTION_VERSION:
		return "get collection version";
	case GET_COLLECTION_COMMENT:
		return "get collection comment";
    default:
    	return "unknown request type : __FILE__ __FUNCTION__ __LINE__"; 
    }
}
