/***************************Copyright-DO-NOT-REMOVE-THIS-LINE**
  *
  * Condor Software Copyright Notice
  * Copyright (C) 1990-2005, Condor Team, Computer Sciences Department,
  * University of Wisconsin-Madison, WI.
  *
  * This source code is covered by the Condor Public License, which can
  * be found in the accompanying LICENSE.TXT file, or online at
  * www.condorproject.org.
  *
  * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
  * AND THE UNIVERSITY OF WISCONSIN-MADISON "AS IS" AND ANY EXPRESS OR
  * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
  * WARRANTIES OF MERCHANTABILITY, OF SATISFACTORY QUALITY, AND FITNESS
  * FOR A PARTICULAR PURPOSE OR USE ARE DISCLAIMED. THE COPYRIGHT
  * HOLDERS AND CONTRIBUTORS AND THE UNIVERSITY OF WISCONSIN-MADISON
  * MAKE NO MAKE NO REPRESENTATION THAT THE SOFTWARE, MODIFICATIONS,
  * ENHANCEMENTS OR DERIVATIVE WORKS THEREOF, WILL NOT INFRINGE ANY
  * PATENT, COPYRIGHT, TRADEMARK, TRADE SECRET OR OTHER PROPRIETARY
  * RIGHT.
  *
  ****************************Copyright-DO-NOT-REMOVE-THIS-LINE**/

#ifndef _NEST_REQUEST_H
#define _NEST_REQUEST_H

#include "nest_common.h"
// #include "nest_request_parms.h"

    // requests
enum
{
    TIME_REQUEST,                       // 0
    CLIENT_SEND_FILE,
    CLIENT_RECV_FILE,
    FILE_DELETE_REQUEST,
    BANDWIDTH_STATISTICS_REQUEST,

    DIR_LIST_REQUEST,                   // 5
    MKDIR_REQUEST,
    CWD_REQUEST,
    PWD_REQUEST,
    FILESIZE_REQUEST,

    SERVER_TERMINATE_REQUEST,           // 10
    CLIENT_TERMINATE_REQUEST,
    TEST_MESSAGE,
    TEST_CLIENT_TO_SERVER,
    TEST_SERVER_TO_CLIENT,

    REQUEST_DONE,                       // 15
    CHANGE_USER_REQUEST,
    AVAILABLE_SPACE_QUERY,
    FORMAT_DISK_REQUEST,
    CONNECTION_REQUEST,

    CLOSE_CONNECTION_REQUEST,           // 20
    THIRD_PARTY_GET,
    THIRD_PARTY_PUT,
    DEBUG_REQUEST,
    RMDIR_REQUEST,

    NAME_COLLECTION,                    // 25
    UNNAME_COLLECTION,
    QUERY_COLLECTION,
    HELP_REQUEST,
    RHELP_REQUEST,

    NEST_SYNC,                          // 30
    REQUEST_NEST_AD,
    LINK_REQUEST,
    ADD_NEST_USER,
    ADD_NEST_GROUP,

    DELETE_NEST_USER,                   // 35
    BAD_REQUEST,
    DELETE_NEST_GROUP,
    ADD_USER_TO_GROUP,
    DELETE_USER_FROM_GROUP,

    REQUEST_USER_LOT,                   // 40
    REQUEST_GROUP_LOT,
    LIST_USER_LOTS,
    LIST_GROUP_LOTS,
    GET_LOT_STAT,

    UPDATE_LOT,                         // 45
    CHANGE_LOT_T,
    TERMINATE_LOT,
    VERSION_REQUEST,
    PHYSNAME_REQUEST,

    GET_ACL,                            // 50
    SET_ACL,                            
    GET_GROUP_ACL,
    SET_GROUP_ACL,
    LOT_MOVE_FILE,

    GET_FILEHANDLE,                     // 55
    FILEINFO_REQUEST,
    PROXY_GET,
    PROXY_PUT,
    PROXY_REGISTER,

    PROXY_QUERY,                        // 60
    COPY_REQUEST,                       
    ADD_USER_TO_LOT,
    DELETE_USER_FROM_LOT,
    FILE_RENAME_REQUEST,

    ATTACH_LOT_TO_FILE,                 // 65
    DETACH_LOT_FROM_FILE,
    PROXY_INDEX_PUSH,                   
    PROXY_FLUSH,
    SETATTR_REQUEST,

	SET_COLLECTION_MTIME,				// 70
	SET_COLLECTION_VERSION,
	SET_COLLECTION_COMMENT,
	GET_COLLECTION_MTIME,
	GET_COLLECTION_VERSION,

	GET_COLLECTION_COMMENT				// 75
};

class nestRequest
{

  public:
    nestRequest( int type, NestReplyStatus status = NEST_SUCCESS );
    nestRequest( const nestRequest *);
    ~nestRequest( );

	// Validate the whole thing
	bool validate( void );
	bool isValid( void ) const
		{ return validated; };

	// Access meta info
	int getType( void ) const
		{ return req_type; };
	int getReqType( void ) const
		{ return req_type; };
	bool isWrite( int type ) const
		{ return ( type == CLIENT_SEND_FILE ); };
	bool isWrite( void ) const
		{ return( isWrite( req_type ) ); };
	bool isThirdParty( void ) const
		{ return ( req_type == THIRD_PARTY_GET ||
				   req_type == THIRD_PARTY_PUT ); };
	NestReplyStatus getStatus( void ) const
		{ return ( status ); };
    const char *toString( void ) const;
	long getAge( void ) const;

	// Access methods: User / group
	const nestUser *getUser( void ) const
		{ return user; };
	const nestGroup *getGroup( void ) const
		{ return group; };

	// Access authentication parms
	int getAuthType( void ) const
		{ return auth_type; };

	// Access methods: File
	const char *getFile( void ) const
		{ return file; };
	const char *getNestFile( void ) const
		{ return file; };
	const char *getDestFile( void ) const
		{ return dest_file; };
	const char *getDir( void ) const
		{ return dir; };
	nestFileSize getStart( void ) const
		{ return start; };
	nestFileSize getEnd( void ) const
		{ return end; };
    nestFileSize getLength( void ) const
		{ return ( end - start ); };
	bool getIsSymLink( void ) const
		{ return is_sym_link; };

	// Access third party xfer parameters
	const char *getRemoteFile( void ) const
		{ return remote_file; };
	const char *getRemoteHost( void ) const
		{ return remote_host; };
	int getRemoteProtocol( void ) const
		{ return remote_protocol; };

	// Access lot parameters
	const NestLotHandle *getLotHandle( void ) const
		{ return &lot_handle; };
	nestLotSize getLotSize( void ) const
		{ return lot_size; };
	int getLotType( void ) const
		{ return lot_type; };
	int getLotDuration( void ) const
		{ return lot_duration; };

	// Access lot creation parameters
	const nestUserOrGroup *getLotParentUG( void ) const
		{ return lot_pug; };
	const NestLotHandle *getLotParentLot( void ) const
		{ return &lot_phandle; };

	// Access lot move parameters
	const NestLotHandle *getLotSourceHandle( void ) const
		{ return &lot_src_handle; };
	const NestLotHandle *getLotDestHandle( void ) const
		{ return &lot_dst_handle; };
	nestLotSize getLotBytesToMove( void ) const
		{ return lot_bytes_to_move; };

	// Access collection parameters
	const char *getCollectionName( void ) const
		{ return coll_name; };
	const char *getCollectionComment( void ) const
		{ return coll_comment; };
	const char *getCollectionVersion( void ) const
		{ return coll_version; };
	time_t getCollectionMtime( void ) const
		{ return coll_mtime; };

	// Access proxy parms
	const char *getProxyLocation( void ) const
		{ return proxy_location; };

	// Access permissions
	nest_access_rights_t getRights( void ) const
		{ return rights; };

	// Access debug parms
	const char *getDebugString( void ) const
		{ return debug_string; };

	// Access: Misc
	const void *getFileAttr( void ) const
		{ return file_attr; };
	nestFileSize getChunkSize( void ) const
		{ return chunk_size; };

	// Set the status
	void setStatus( NestReplyStatus status );

	// Set user/group parameters
	void setUser( const nestUser *, bool copy = true );
	void setUser( const char * );
	void setGroup( const nestGroup *, bool copy = true );
	void setGroup( const char * );

	// Set authentication parms
	void setAuthType( int );
	void setAuthType( const char * );

	// Set file parameters
	void setFile( const char * );
	void setDestHost( const char * );
	void setDestFile( const char * );
	void setDir( const char * );
	void setStart( nestFileSize );
	void setStart( const char * );
	void setEnd( nestFileSize );
	void setEnd( const char * );
	void setIsSymLink( bool );
	void setIsSymLink( const char * );

	// Set third party xfer parameters
	void setRemoteHost( const char * );
	void setRemoteFile( const char * );
	void setRemoteProtocol( const char * );
	void setRemoteProtocol( int );

	// Set Lot parameters
	void setLotHandle( const NestLotHandle * );
	void setLotHandle( const char * );
	void setLotSize( nestLotSize );
	void setLotSize( const char * );
	void setLotType( int );
	void setLotType( const char * );
	void setLotDuration( int );
	void setLotDuration( const char * );

	// Set Lot creation parameters
	void setLotParentUser( const char * );
	void setLotParentUser( const nestUser *, bool copy = true );
	void setLotParentGroup( const char * );
	void setLotParentGroup( const nestGroup *, bool copy = true );
	void setLotParentLot( const NestLotHandle * );
	void setLotParentLot( const char * );

	// Set lot move parameters
	void setLotSourceHandle( const NestLotHandle * );
	void setLotSourceHandle( const char * );
	void setLotDestHandle( const NestLotHandle * );
	void setLotDestHandle( const char * );
	void setLotBytesToMove( nestLotSize );
	void setLotBytesToMove( const char * );

	// Set collection parameters
	void setCollectionName( const char * );
	void setCollectionComment( const char * );
	void setCollectionVersion( const char * );
	void setCollectionMtime( time_t mtime );
	void setCollectionMtime( const char * );

	// Set Permissions parms
	void setRights( nest_access_rights_t );
	void setRights( const char * );

	// Set proxy parms
	void setProxyLocation( const char * );

	// Set debug parms
	void setDebugString( const char * );

	// Specialized
	void setFileAttr( const void * );
	void setChunkSize( nestFileSize );


  private:
	void init( void );
	void init( NestLotHandle & );
	bool isSet( const NestLotHandle & ) const;

	// Meta data
	bool			validated;
	NestReplyStatus status;
	int				req_type;
    struct timeval	createtime, starttime, endtime;

	// User / group parameters
	const nestUser	*user;
	const nestGroup	*group;

	// Auth parameters
	int auth_type;
	bool auth_type_set;

	// File parameters
	const char		*file;
	const char		*dest_file;
	const char		*dest_host;
	const char		*dir;
	nestFileSize	start;
	nestFileSize	end;
	bool			is_sym_link;

	// Third party xfer parms
	const char		*remote_file;
	const char		*remote_host;
	int				remote_protocol;

	// Lot parameters
	NestLotHandle	lot_handle;
	int				lot_type;
	nestLotSize		lot_size;
	bool			lot_size_set;
	int   			lot_duration;
	bool			lot_duration_set;

	// Lot creation parameters
	const nestUserOrGroup	*lot_pug;
	NestLotHandle			lot_phandle;

	// Lot move parameters
	NestLotHandle	lot_src_handle;
	NestLotHandle	lot_dst_handle;
	nestLotSize		lot_bytes_to_move;

	// Collection parameters
	const char		*coll_name;
	const char		*coll_comment;
	const char		*coll_version;
	time_t			coll_mtime;

	// Permissions
	nest_access_rights_t	rights;
	bool					rights_set;

	// Proxy parms
	const char		*proxy_location;

	// Debug parms
	const char		*debug_string;

	// Misc
	const void		*file_attr;
	nestFileSize	chunk_size;

};

#endif
