/***************************Copyright-DO-NOT-REMOVE-THIS-LINE**
  *
  * Condor Software Copyright Notice
  * Copyright (C) 1990-2005, Condor Team, Computer Sciences Department,
  * University of Wisconsin-Madison, WI.
  *
  * This source code is covered by the Condor Public License, which can
  * be found in the accompanying LICENSE.TXT file, or online at
  * www.condorproject.org.
  *
  * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
  * AND THE UNIVERSITY OF WISCONSIN-MADISON "AS IS" AND ANY EXPRESS OR
  * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
  * WARRANTIES OF MERCHANTABILITY, OF SATISFACTORY QUALITY, AND FITNESS
  * FOR A PARTICULAR PURPOSE OR USE ARE DISCLAIMED. THE COPYRIGHT
  * HOLDERS AND CONTRIBUTORS AND THE UNIVERSITY OF WISCONSIN-MADISON
  * MAKE NO MAKE NO REPRESENTATION THAT THE SOFTWARE, MODIFICATIONS,
  * ENHANCEMENTS OR DERIVATIVE WORKS THEREOF, WILL NOT INFRINGE ANY
  * PATENT, COPYRIGHT, TRADEMARK, TRADE SECRET OR OTHER PROPRIETARY
  * RIGHT.
  *
  ****************************Copyright-DO-NOT-REMOVE-THIS-LINE**/
#ifndef _NEST_STORAGE_MANAGER_H
#define _NEST_STORAGE_MANAGER_H

#include "nest_arch_specific.h"

#include "nest_common.h"
#include "nest_protocol_storage.h"
#include "nest_user.h"
#include "nest_client.h"

#include "nest_identity_manager.h"
#include "nest_lot_manager.h"
#include "nest_file_catalog.h"
#include "nest_collection_manager.h"

#include "nest_hash_table.h"

class
nestStorageManager
{

public:

	// Constructors & destructors
	nestStorageManager( void );
	virtual ~nestStorageManager( void );

    //  Initialize  Storage Manager
    int  init( void );

    NestReplyStatus  initialize();

    void writeAd( nestProtocol *);

    NestReplyStatus  sync();

    NestReplyStatus  maintenance();
    NestReplyStatus  shutdown();

    //  Meta-file operations
    nestProtocol*    openMetaFile(char *file,
								  int flags );

    NestReplyStatus  renameMetaFile(char *oldpath,
									char *newpath );

    NestReplyStatus  statMetaFile(char *file,
								  nest_stat *st );

    NestReplyStatus  unlinkMetaFile( char *file );

    char*            metaFilepath(char *file );

    nestFileSize     metaFilelength( char *file );

    //  File operations
    NestReplyStatus  open(nestProtocol **result,
						  const nestUser *user,
						  const char *file,
						  int flags,
						  off_t offset,
						  ssize_t filesize,
						  const char *cwd,
						  int link_count = 0 );

    char *           filehandleToPath( NestFilehandle * );
    
    NestReplyStatus  pathToFilehandle(const nestUser *user,
									  const char *file,
									  const char *cwd,
									  NestFilehandle   *inode_no,
									  int   link_count = 0);



    NestReplyStatus  close(nestProtocol *connection,
						   nestClient *client = NULL);

    NestReplyStatus  deleteLink(const char *filename);

    NestReplyStatus  deleteDirContents(const nestUser *user,
									   const char *filename);

    NestReplyStatus  deleteFile(const nestUser *user,
								const char *filename,
								const char *cwd,
								bool update_lot = true);

    NestReplyStatus  renameFile( const nestUser *user,
								 const char *oldpath,
								 const char *newpath,
								 const char *cwd,
								 bool update_lot = true );

    NestReplyStatus  get_type( nest_obj_type_t   *result,
							   const char *filename,
							   const nestUser *user,
							   const char *cwd );

    NestReplyStatus  get_physical_filename(char **physical,
										   const char *filename,
										   const nestUser *user,
										   const char *cwd,
										   int link_count = 0);

    NestReplyStatus  Link(const nestUser *user,
						  const char *oldpath,
						  const char *newpath,
						  int val,
						  const char *cwd);

    NestReplyStatus  sync(const nestUser *user);


    //  Directory operations

    NestReplyStatus frame_full_path( string &result,
									 const nestUser *user,
									 const char *file );

    NestReplyStatus expand_path( string &result,
								 const nestUser *user,
								 const char *cwd,
								 const char *file );


    NestReplyStatus get_dest_dir( char *dest,
								  const char *cwd,
								  const char *dir,
								  const nestUser *user );

    NestReplyStatus chDir(char **new_cwd,
						  const nestUser *user,
						  const char *cwd,
						  const char *dir);

    NestReplyStatus mkDir(const nestUser *user,
						  const char *cwd,
						  const char *dir,
						  user_id_t uid );

    NestReplyStatus rmDir(const nestUser *user,
						  const char *cwd,
						  const char *dir,
						  bool recursive = false );

    char*            getRootDir( );

    NestFileInfo*    directoryContents(const nestUser *user,
									   const char *dir );


    //  User/Group Management operations
#if 0
    NestReplyStatus  add_user(const char *username,
							  const char *groupname,
							  const nestUser *cur_user,
							  const char *identity);
#endif

    NestReplyStatus  add_user(const nestUser *username,
							  const nestUser *cur_user);

    NestReplyStatus  delete_user(const nestUser *cur_user,
								 const nestUser *user_to_delete);

    NestReplyStatus  add_group(const nestUser *cur_user,
							   const nestGroup *group );

    NestReplyStatus  delete_group(const nestUser *cur_user,
								  const nestGroup *group );

    NestReplyStatus  add_user_to_group(const nestUser *cur_user,
									   const nestUser *user,
									   const nestGroup *group );

    NestReplyStatus  delete_user_from_group(const nestUser *cur_user,
											const nestUser *user,
											const nestGroup *group );
											
    NestReplyStatus  add_home_dir(const nestUser *user );

    //  Access control operations
    NestReplyStatus  get_acl(NestACL **result,
							 const char *file,
							 const nestUser *user,
							 const char *cwd);

    NestReplyStatus  get_perm(nest_access_rights_t *result,
							  const nestUserOrGroup *for_name,
							  const char *file,
							  const nestUser *user,
							  const char *cwd);

    NestReplyStatus  set_perm(nest_access_rights_t arg,
							  const nestUserOrGroup *to_name,
							  const char *file,
							  const nestUser *user,
							  const char *cwd);

    NestReplyStatus modify_lot_usage(const nestUser *user,
									 const char *filename,
									 const char *cwd,
									 const NestLotHandle  *from_lot,
									 nestLotSize from_share,
									 const NestLotHandle  *to_lot);

    NestReplyStatus attach_lot_to_file(const NestLotHandle *handle,
									   const char *file,  
									   const nestUser *user,
									   const char *cwd);

    NestReplyStatus detach_lot_from_file(const NestLotHandle *handle,
										 const char *file,  
										 const nestUser *user,
										 const char *cwd);

    NestReplyStatus increaseAllocation(nestClient *client);
    NestReplyStatus resetAllocation(const char *file,
									const nestUser *user,
									const char *cwd,
									nestFileSize offset); 
    
    NestReplyStatus cleanupGftpTransfer(nestClient *client);
                                                
    //  Other operations
    NestReplyStatus  bytesAvailable( int is_user,
									 nestLotSize *free,
									 nestLotSize *total );

    NestReplyStatus  filelength(  nestFileSize *filesize,  
								  const char *filename,
								  const nestUser *user,
								  const char *cwd );

    NestReplyStatus  fileusage(  nestFileSize *size,
								 const char *file,
								 const nestUser *user, 
								 const char *cwd );

    NestReplyStatus  fileinfo(  NestFileInfo *info,
								const char *filename,
								const nestUser *user,
								const char *cwd,
								int link_count = 0 );

    char *           getInitialClientDir( );
    
    bool             validate_user( nestUser *user );

    NestReplyStatus  formatDisk(
		const nestUser *user,
		const nestGroup *group
		);

    NestReplyStatus  rememberLength( 
		const nestUser *user,
		const char *cwd,
		const char *file,
		nestFileSize *length );

    char *           type();

    bool             is_lots_enabled();

  private:

	nestLotSize             storage_space;
	char *                  userlist;
	char *                  denylist;

	mode_t                    file_create_mode;


	struct nest_hash_table	*fh_hash;

	// the flat directory space where the server stores its files
	char*            		rootDir;

    // ClassAdCollectionServer *Resources;
	ClassAdCollection       *Resources;
	ClassAdParser           parser;
	ClassAdUnParser         unparser;
	LocalCollectionQuery    query;

    int chRootDir( );
    int chMetaDir( );
    NestReplyStatus  synchronize();
    NestReplyStatus  get_home_dir(string &result,
								  const nestUser *user);
    NestReplyStatus  delete_home_dir(const nestUser *user);

	void cache_uniquifier (  string &logical,
							 NestFilehandle *uniq,
							 nest_obj_type_t type );

	NestReplyStatus
		closeLotFile( nestClient *client,
					  const nestRequest *req,
					  const nestUser *owner,
					  const char * l_fname,
					  nestFileSize open_filesize_bytes,
					  nestFileSize close_filesize_bytes,
					  bool &delete_file );

};

extern nestStorageManager	*storage_manager;

#endif
