/***************************Copyright-DO-NOT-REMOVE-THIS-LINE**
  *
  * Condor Software Copyright Notice
  * Copyright (C) 1990-2005, Condor Team, Computer Sciences Department,
  * University of Wisconsin-Madison, WI.
  *
  * This source code is covered by the Condor Public License, which can
  * be found in the accompanying LICENSE.TXT file, or online at
  * www.condorproject.org.
  *
  * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
  * AND THE UNIVERSITY OF WISCONSIN-MADISON "AS IS" AND ANY EXPRESS OR
  * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
  * WARRANTIES OF MERCHANTABILITY, OF SATISFACTORY QUALITY, AND FITNESS
  * FOR A PARTICULAR PURPOSE OR USE ARE DISCLAIMED. THE COPYRIGHT
  * HOLDERS AND CONTRIBUTORS AND THE UNIVERSITY OF WISCONSIN-MADISON
  * MAKE NO MAKE NO REPRESENTATION THAT THE SOFTWARE, MODIFICATIONS,
  * ENHANCEMENTS OR DERIVATIVE WORKS THEREOF, WILL NOT INFRINGE ANY
  * PATENT, COPYRIGHT, TRADEMARK, TRADE SECRET OR OTHER PROPRIETARY
  * RIGHT.
  *
  ****************************Copyright-DO-NOT-REMOVE-THIS-LINE**/

#include "nest_config.h"
#include "nest_pwent.h"
#include "nest_user.h"


// Base class
nestUserOrGroup::nestUserOrGroup( void )
{
	actualName = NULL;
}

nestUserOrGroup::~nestUserOrGroup( void )
{
	if ( this->actualName ) {
		NestFree( (void*) this->actualName );
	}
}

bool
nestUserOrGroup::match( const nestUserOrGroup *other ) const
{
	if ( this->isUser() != other->isUser() ) {
		return false;
	}
	if ( this->getName() ) {
		return false;
	}
	if ( !other->getName() ) {
		return false;
	}

	if ( strcmp( this->getName(), other->getName() ) ) {
		return false;
	} else {
		return true;
	}
}

const char *
nestUserOrGroup::setActual( const char *name ) const
{
	if ( this->actualName ) {
		NestFree( (void*) this->actualName );
	}
	this->actualName = strdup( name );
	return this->actualName;
}

const char *
nestUserOrGroup::setActual( const string *name ) const
{
	return setActual( name->c_str() );
}

const char *
nestUserOrGroup::getActual( void ) const
{
	return this->actualName;
}

// here is how a nest user object is used
// A user object is created, and it follows the request

// Basic constructor
nestUser::nestUser( void )
{
	init( );
}

// Copy constructor
nestUser::nestUser( const nestUser *user )
{
	init( );
	copyUser( user );
}    

nestUser::nestUser( const string	*identity_str )
{
	init( );

	// Pull the "C" strings out of the strings, call the
	// char * constructor....
	const char	*tmp_identity = NULL;
	if ( identity_str ) {
		tmp_identity = identity_str->c_str( );
	}
	setIdentity( tmp_identity );
}
nestUser::nestUser( const char		*nestIdentity )
{
	init( );
	setIdentity( nestIdentity );
}

// Destructor
nestUser::~nestUser( void )
{
	freeAll( );		// Free up the memory
}

// "Reset" a user to a "clean" state
int
nestUser::reset( void )
{
	// First, free up all memory
	this->freeAll( );

	// Then, re-initialize it
	return init( );
}

// Copy the contents of another user object
int
nestUser::copyUser( const nestUser *from )
{
	if ( from->osName ) {
		this->osName = strdup( from->osName );
	}
	if ( from->localName ) {
		this->localName = strdup( from->localName );
	}
	if ( from->group ) {
		this->group = new nestGroup( from->group );
	}
	if ( from->identity ) {
		this->identity = strdup ( from->identity );
	}
	if ( from->identityClean ) {
		this->identityClean = strdup( from->identityClean );
	}
	this->userId = from->userId;
	this->dirty = from->dirty;

	// Flags
	this->anonymous = from->anonymous;
	this->admin = from->admin;
	this->everybody = from->everybody;
	return 0;
}

// Simple initializer which zero's everything
int
nestUser::init( void )
{
	this->osName	        = NULL;		// the O/S (UNIX) name
	this->localName         = NULL;		// Local name
	this->identity          = NULL;		// the global name
	this->identityClean		= NULL;		// Without spaces, etc.
	this->group				= NULL;
	this->userId            = USERID_INVALID;
	this->dirty				= USERDIRTY_NONE;

	// Reset the flags
	this->anonymous.Reset( );
	this->admin.Reset( );
	this->everybody.Reset( );
	this->system.Reset( );

	return 0;
}

// Fill in the O/S info if we can..
int
nestUser::fillOsName( void ) const
{
	if ( !isDirty( USERDIRTY_OSNAME ) ) {
		return 0;
	}
	nestUser *m_this = const_cast<nestUser*>( this );
	m_this->unsetDirty( USERDIRTY_OSNAME );
	if ( osName ) {
		return 0;
	}

	// No need to check the identity or UID for dirty status....
	// - Identity can only be filled in from O/S name
	// - UID can only be filled from O/S name

	// If we have an identity, use it
	if ( identity ) {
		const char *tmp = NestGetUserFromIdentity( this->identity );
		if ( tmp ) {
			m_this->osName = strdup( tmp );
			return 0;
		}
	}

	// If we have a valid o/s UID, us it
	if ( USERID_INVALID != userId ) {
		nestPwent	pwent;
		pwent.findent( userId );
		const char *name = pwent.getent_name( );
		if ( name ) {
			m_this->osName = strdup( name );
		}
	}

	return 0;
}

// Fill in the UID, if we can
int
nestUser::fillUid( void ) const
{
	if ( !isDirty( USERDIRTY_UID ) ) {
		return 0;
	}
	nestUser *m_this = const_cast<nestUser*>( this );
	m_this->unsetDirty( USERDIRTY_UID );
	if ( USERID_INVALID != userId ) {
		return 0;
	}

	// Dependant on "O/S Name"
	if ( isDirty( USERDIRTY_OSNAME ) ) {
		fillOsName( );
	}

	if ( osName ) {
		nestPwent	pwent;
		if ( pwent.findent( osName ) == 0 )
		{
			m_this->userId = (user_id_t) pwent.getent_uid( );
		}
	}

	return 0;
}

// Fill in the identity if we can..
int
nestUser::fillIdentity( void ) const
{
	if ( !isDirty( USERDIRTY_IDENTITY ) ) {
		return 0;
	}
	nestUser *m_this = const_cast<nestUser*>( this );

	// Clear my flags to avoid recursion
	m_this->unsetDirty( USERDIRTY_IDENTITY );
	if ( identity ) {
		return 0;
	}

	// Try to find the user name
	if ( isDirty( USERDIRTY_OSNAME ) ) {
		fillOsName( );
	}

	// Finally, set the identity
	const char *identity = NestGetIdentityFromUser( osName );
	if ( identity ) {
		m_this->setIdentity( identity );
	}

	return 0;
}

// Fill in the "Clean" identity if we can..
int
nestUser::fillIdentityClean( void ) const
{
	if ( !isDirty( USERDIRTY_IDENTITY_CLEAN ) ) {
		return 0;
	}
	nestUser *m_this = const_cast<nestUser*>( this );
	m_this->unsetDirty( USERDIRTY_IDENTITY_CLEAN );
	if ( identityClean ) {
		return 0;
	}

	// Fill the identity
	if ( isDirty( USERDIRTY_IDENTITY ) ) {
		fillIdentity( );
	}

	// Store the modified / "cleaned" identity
	if ( identity ) {
		m_this->identityClean = NestGetModifiedIdentity( identity );
	}

	return 0;
}

// Fill in just the flags
int
nestUser::fillSpecial( void ) const
{
	if ( ! isDirty(USERDIRTY_SPECIAL) ) {
		return 0;
	}
	nestUser *m_this = const_cast<nestUser*>( this );
	m_this->unsetDirty( USERDIRTY_SPECIAL );

	// Fill in the identity if we can
	if ( isDirty(USERDIRTY_IDENTITY) ) {
		fillIdentity( );
	}

	// Fill the flags now
	if ( identity ) {

		// Set the admin flag
		if ( ! admin.IsSet() ) {
			m_this->admin.Set( isAdmin( identity ) );
		}

		// Set the anonymous flag
		if ( ! anonymous.IsSet() ) {
			m_this->anonymous.Set( isAnonymous( identity ) );
		}

		// Set the everybody flag
		if ( ! everybody.IsSet() ) {
			m_this->everybody.Set( isEverybody( identity ) );
		}

		// Set the system flag
		if ( ! system.IsSet() ) {
			m_this->system.Set( isSystem( identity ) );
		}
	}

	return 0;
}

// Fill in the pieces that we don't know from what we do (as best we can)
int
nestUser::fillAll( void ) const
{
	// If nothing is dirty, nothing to do!
	if ( ! dirty ) {
		return 0;
	}

	// Modifyable this
	nestUser *m_this = const_cast<nestUser*>( this );

	// Fill in everything we can from the O/S name
	fillOsName( );
	fillUid( );
	fillIdentity( );
	fillSpecial( );

	// Finally, note that it's all clean
	m_this->dirty = 0;

	return 0;
}

// Clean this object (free up all memory)
void
nestUser::freeAll( void )
{
    if ( this->osName ) {
		NestFree( (void*) this->osName );
	}
    if ( this->localName ) {
		NestFree( (void*) this->localName );
	}
    if ( this->group ) {
		delete this->group;
	}
    if ( this->identity ) {
		NestFree( (void*) this->identity );
	}
    if ( this->identityClean ) {
		NestFree( (void*) this->identityClean );
	}
}

// Set UID, return old value
user_id_t
nestUser::setUid( user_id_t uid )
{
	user_id_t	oldUid = userId;
	userId = uid;

	setDirty ( USERDIRTY_IDENTITY |
			   USERDIRTY_IDENTITY_CLEAN |
			   USERDIRTY_SPECIAL |
			   USERDIRTY_OSNAME );
	unsetDirty( USERDIRTY_UID );

	return oldUid;
}

// Reset the identity, optionally store a new one
int
nestUser::resetIdentity( const char *identity )
{
	// Delete existing identity....
    if ( this->identity != NULL) {
        NestFree( (void*)this->identity );
		this->identity = NULL;
    }

	// Reset the flags
	this->anonymous.Reset( );
	this->admin.Reset( );
	this->everybody.Reset( );
	this->system.Reset( );

	// Everything could be dirty...
	setDirty ( USERDIRTY_IDENTITY |
			   USERDIRTY_IDENTITY_CLEAN |
			   USERDIRTY_SPECIAL );
	return 0;
}

int
nestUser::setIdentity( const char *identity )
{
	resetIdentity( identity );

	// If no identity passed in, we're done
	if ( ! identity ) {
		return NEST_BAD_PARAMETERS;
	}

	// Copy the new one in
	if ( identity ) {
		this->identity = NestStrdup( identity );
	}
	unsetDirty( USERDIRTY_IDENTITY );
	setDirty ( USERDIRTY_IDENTITY_CLEAN |
			   USERDIRTY_OSNAME |
			   USERDIRTY_UID |
			   USERDIRTY_SPECIAL );

	// Done
	return NEST_SUCCESS;
}

int
nestUser::setIdentity( const string *ident_str )
{
	if ( !ident_str ) {
		return NEST_SUCCESS;
	}
	return setIdentity( ident_str->c_str( ) );
}

int
nestUser::resetOsName( void )
{
	// Clear things up...
    if ( osName != NULL ) {
        NestFree( (void*) osName );
		osName = NULL;
    }

	userId = USERID_INVALID;
	setDirty( USERDIRTY_OSNAME );
	return 0;
}

int
nestUser::setOsName( const char *name )
{
	// It's dirty
	dirty = true;

	// Reset first
	resetOsName( );

	// No string provided?  Done
	if ( ! name ) {
		return NEST_SUCCESS;
	}

	// Copy the name
	osName = NestStrdup( name );
	if ( ! osName ) {
		return NEST_OUT_OF_RESOURCES;
	}

	setDirty ( USERDIRTY_IDENTITY |
			   USERDIRTY_IDENTITY_CLEAN |
			   USERDIRTY_SPECIAL |
			   USERDIRTY_UID );
	unsetDirty( USERDIRTY_OSNAME );

	// Done
	return NEST_SUCCESS;
}

int
nestUser::setOsName( const string *username )
{
	if ( !username ) {
		return NEST_BAD_PARAMETERS;
	}
	return setOsName( username->c_str() );
}

int
nestUser::setLocalName( const char *name )
{
	// Reset first
	resetLocalName( );

	// No string provided?  Done
	if ( ! name ) {
		return NEST_SUCCESS;
	}

	// Copy the name
	localName = NestStrdup( name );
	if ( ! localName ) {
		return NEST_OUT_OF_RESOURCES;
	}

	// Done
	return NEST_SUCCESS;
}

int
nestUser::resetLocalName( void )
{
	// Clear things up...
    if ( localName != NULL ) {
        NestFree( (void*) localName );
		localName = NULL;
    }

	return 0;
}

int
nestUser::setLocalName( const string *localname )
{
	if ( !localname ) {
		return NEST_BAD_PARAMETERS;
	}
	return setLocalName( localname->c_str() );
}

const char *
nestUser::getLocalName( void ) const
{
	if ( localName ) {
		return localName;
	} else if ( osName ) {
		return osName;
	} else {
		return identity;
	}
}

// Tries to find a match for string as an identity or O/S name
int
nestUser::setName( const char *name )
{
	if ( ! name ) {
		return NEST_BAD_PARAMETERS;
	}

	// "System" user?
	if ( isSpecial( name ) ) {
		setIdentity( name );
		return NEST_SUCCESS;
	}

	// First, try to match the identity...
	grid_mapfile_entry	entry;
	bool matched = NestGetMapfileEntry( name, &entry );
	if ( matched ) {		// match found!
		setIdentity( entry.identity );
		setOsName( entry.name );
		return NEST_SUCCESS;
	}

	// If no match, force it to the "local name"
	return setLocalName( name );
}

int
nestUser::setName( const string *name )
{
	if ( !name ) {
		return NEST_BAD_PARAMETERS;
	}

	return setName( name->c_str( ) );
}

int
nestUser::setGroup( const nestGroup *group )
{
    if ( this->group != NULL ) {
		delete this->group;
		this->group = NULL;
    }

	if ( group ) {
		this->group = new nestGroup( group );
	}

    return NEST_SUCCESS;
}

const char *
nestUser::getGroupName( void ) const
{
	if ( ! group ) {
		return "";
	} else {
		return group->getName( );
	}
}

bool
nestUser::match( const nestUser *other ) const
{
	// If they both have identities & they match, good.
	if ( this->identity && other->identity ) {
		if ( ! strcmp( this->identity, other->identity ) ) {
			return true;
		}
	}

	// Ditto for name strings
	if ( this->osName && other->osName ) {
		if ( ! strcmp( this->osName, other->osName ) ) {
			return true;
		}
	}

	// Otherwise, they don't match
	return false;
}

bool
nestUser::match( const nestUserOrGroup *other ) const
{
	// If it's not a nestUser, it doesn't match!!
	if( ! other->isUser() ) {
		return false;
	}

	return match( dynamic_cast <const nestUser *> ( other ) );
}

bool
nestUser::match( const char *identity ) const
{

	// If no identity, the answer is no
	if ( ! this->identity ) {
		return false;
	}

	// If they both have identities & they match, good.
	if ( ! strcmp( this->identity, identity ) ) {
		return true;
	}

	// Otherwise, they don't match
	return false;
}

bool
nestUser::match( const string *other ) const
{
	if ( !other ) {
		return false;
	}
		
	return match( other->c_str() );
}

bool
nestUser::matchAny( const char *name ) const
{

	// First, try to match identity
	if ( this->identity ) {
		if ( ! strcmp( this->identity, name ) ) {
			return true;
		}
	}

	// Next, try to match the local name
	if ( this->localName ) {
		if ( ! strcmp( this->localName, name ) ) {
			return true;
		}
	}

	// Finally, try to match the O/S user name
	if ( this->osName ) {
		if ( ! strcmp( this->osName, name ) ) {
			return true;
		}
	}

	// No mtach
	return false;
}

bool
nestUser::matchAny( const string *name ) const
{
	if ( !name ) {
		return false;
	}
		
	return matchAny( name->c_str() );
}

bool
nestUser::isSpecial( const char *name ) const
{
	// System user?
	if ( isSystem( name ) ) {
		return true;
	}
	if ( isAnonymous( name ) ) {
		return true;
	}
	if ( isEverybody( name ) ) {
		return true;
	}
	if ( isAdmin( name ) ) {
		return true;
	}
	return false;
}

bool
nestUser::isAdmin( const char *name ) const
{
	if ( ! name ) {
		return false;
	}

	if ( NestAdminUser && !strcmp( name, NestAdminUser ) ) {
		return true;
	}

	return false;
}

bool
nestUser::isAnonymous( const char *name ) const
{
	if ( ! name ) {
		return false;
	}
	if ( NestAnonymousUser && !strcmp( name, NestAnonymousUser ) ) {
		return true;
	}

	return false;
}

bool
nestUser::isEverybody( const char *name ) const
{
	if ( ! name ) {
		return false;
	}
	if ( !strcmp( name, NEST_EVERYBODY ) ) {
		return true;
	}

	return false;
}

bool
nestUser::isSystem( const char *name ) const
{
	if ( ! name ) {
		return false;
	}
	if ( NestSystemUser && !strcmp( name, NestSystemUser ) ) {
		return true;
	}
	return false;
}


// ************************************************
//  nestGroup methods
// ************************************************

nestGroup::nestGroup( void )
{
	init( );
}

nestGroup::nestGroup( const char *name)
{
	init( );
	setName( name );
}

nestGroup::nestGroup( const string *name)
{
	init( );
	setName( name );
}

nestGroup::nestGroup( const char *name, user_id_t groupId )
{
	init( );
	setName( name );
}

nestGroup::nestGroup( const string *name, user_id_t groupId )
{
	init( );
	setName( name );
}

nestGroup::nestGroup( const nestGroup *group)
{
	init( );
	setName( group->name, group->groupId );
}

nestGroup::~nestGroup( void )
{
	if( this->name ) {
		free( (void*) this->name );
	}
}

int
nestGroup::init( void )
{
	this->name = NULL;
	this->groupId = USERID_INVALID;
	return 0;
}

int
nestGroup::setName( const char *name )
{
	if( this->name ) {
		free( (void*) this->name );
	}
	this->name = strdup( name );

	group_ent   ge;
	if (Nestgetgrpnam( name, &ge, SystemGrpFile) < 0) {
		return 0;
	}
	groupId = (user_id_t) ge.grp_gid;
	return 0;
}

int
nestGroup::setName( const char *name, user_id_t groupId )
{
	if( this->name ) {
		free( (void*) this->name );
	}
	this->name = strdup( name );
	this->groupId = groupId;
	return 0;
}

int
nestGroup::setName( const string *name)
{
	setName( name->c_str() );
	return 0;
}

int
nestGroup::setName( const string *name, user_id_t groupId )
{
	setName( name->c_str(), groupId );
	return 0;
}

bool
nestGroup::match( const char *name ) const
{

	// If no name, the answer is no
	if ( ! this->name ) {
		return false;
	}

	// If they both have identities & they match, good.
	if ( ! strcmp( this->name, name ) ) {
		return true;
	}

	// Otherwise, they don't match
	return false;
}

bool
nestGroup::match( const string *other ) const
{
	if ( !other) {
		return false;
	}
	return match( other->c_str() );
}

bool
nestGroup::match( const nestGroup *other ) const
{
	// If they both have identities & they match, good.
	if ( !this->name || other->name ) {
		return false;
	}

	if ( ! strcmp( this->name, other->name ) ) {
		return true;
	}

	// Otherwise, they don't match
	return false;
}

bool
nestGroup::match( const nestUserOrGroup *other ) const
{
	// If it's not a nestGroup, it doesn't match!!
	if( ! other->isGroup() ) {
		return false;
	}

	return match( dynamic_cast <const nestGroup *> ( other ) );
}

bool
nestGroup::isNoGroup( void ) const
{
	if ( noGroup.IsSet() ) {
		return noGroup.Value( );
	}

	// Need a non-const to cache the resutl
	nestGroup *m_this = const_cast<nestGroup*>( this );

	// Reset anonymous ONLY IF it's a valid non-anon user
	if ( ! strcmp( name, NEST_NO_GROUP ) ) {
		return m_this->noGroup.Set( true );
	}

	return m_this->noGroup.Set( false );
}
